@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2016-2017 Intel, Ltd.
  * Copyright (C) 2016-2021 Canonical, Ltd.
+ * Copyright (C) 2021-2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,12 +19,17 @@
  *
  */
 #include "stress-ng.h"
+#include "core-arch.h"
+
+#if defined(HAVE_SYS_AUXV_H)
+#include <sys/auxv.h>
+#endif
 
 #if defined(__linux__)
 
 typedef struct {
-	const char	*name;		/* cache type name */
-	const uint32_t	value;		/* cache type ID */
+	const char	*name;			/* cache type name */
+	const stress_cache_type_t value;	/* cache type ID */
 } stress_generic_map_t;
 
 typedef enum {
@@ -90,14 +96,15 @@ static int stress_get_string_from_file(
  * - kibibytes ('K' - aka KiB).
  * - mebibytes ('M' - aka MiB).
  * - gibibytes ('G' - aka GiB).
+ * - tebibutes ('T' - aka TiB).
  *
  * Returns: size in bytes, or 0 on error.
  */
 static uint64_t stress_size_to_bytes(const char *str)
 {
-	uint64_t            bytes;
-	int                 ret;
-	char                sz;
+	uint64_t bytes;
+	int	 ret;
+	char	 sz;
 
 	if (!str) {
 		pr_dbg("%s: empty string specified\n", __func__);
@@ -124,6 +131,9 @@ static uint64_t stress_size_to_bytes(const char *str)
 	case 'G':
 		bytes *= GB;
 		break;
+	case 'T':
+		bytes *= TB;
+		break;
 	default:
 		pr_err("unable to convert '%c' size to bytes\n", sz);
 		bytes = 0;
@@ -168,11 +178,9 @@ out:
  */
 static int stress_add_cpu_cache_detail(stress_cpu_cache_t *cache, const char *index_path)
 {
-	const size_t index_posn = index_path ? strlen(index_path) : 0;
-	const size_t path_len = index_posn + 32;
 	int ret = EXIT_FAILURE;
 	char tmp[2048];
-	char path[path_len];
+	char path[PATH_MAX];
 
 	(void)memset(path, 0, sizeof(path));
 	if (!cache)
@@ -302,13 +310,13 @@ stress_cpu_cache_t *stress_get_cpu_cache(const stress_cpus_t *cpus, const uint16
 	return stress_get_cache_by_cpu(cpu, cache_level);
 }
 
+#if defined(STRESS_ARCH_SPARC)
 static int stress_get_cpu_cache_value(
 	const char *cpu_path,
 	const char *file,
 	uint64_t *value)
 {
-	const size_t cpu_path_len = strlen(cpu_path);
-	char path[cpu_path_len + 128];
+	char path[PATH_MAX];
 	char tmp[128];
 
 	(void)stress_mk_filename(path, sizeof(path), cpu_path, file);
@@ -318,6 +326,7 @@ static int stress_get_cpu_cache_value(
 	}
 	return -1;
 }
+#endif
 
 /*
  *  stress_get_cpu_cache_auxval()
@@ -407,6 +416,99 @@ static int stress_get_cpu_cache_auxval(stress_cpu_t *cpu)
 #endif
 }
 
+#if defined(STRESS_ARCH_ALPHA)
+/*
+ *  stress_get_cpu_cache_alpha()
+ *	find cache information as provided by linux Alpha from
+ *	/proc/cpu. Assume cache layout for 1st CPU is same for
+ *	all CPUs.
+ */
+static int stress_get_cpu_cache_alpha(
+	stress_cpu_t *cpu,
+	const char *cpu_path)
+{
+	FILE *fp;
+	const size_t count = 4;
+	size_t idx = 0;
+
+	(void)cpu_path;
+
+	/*
+	 * parse /proc/cpu info in the form:
+	 * L1 Icache		: 64K, 2-way, 64b line
+	 * L1 Dcache		: 64K, 2-way, 64b line
+	 * L2 cache		: n/a
+	 * L3 cache		: n/a
+	 */
+	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			count * sizeof(*(cpu->caches)));
+		return 0;
+	}
+
+	fp = fopen("/proc/cpuinfo", "r");
+	if (fp) {
+		char buffer[4096];
+
+		while ((idx < count) && fgets(buffer, sizeof(buffer), fp)) {
+			stress_cache_type_t cache_type = CACHE_TYPE_UNKNOWN;
+			uint16_t cache_level = 0;
+			char *ptr;
+			int cache_size, cache_ways, cache_line_size, n;
+
+			if (!strncmp("L1 Icache", buffer, 9)) {
+				cache_type = CACHE_TYPE_INSTRUCTION;
+				cache_level = 1;
+			} else if (!strncmp("L1 Dcache", buffer, 9))  {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 1;
+			}
+			else if (!strncmp("L2 cache", buffer, 8)) {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 2;
+			}
+			else if (!strncmp("L3 cache", buffer, 8)) {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 3;
+			}
+			else
+				continue;
+			ptr = strchr(buffer, ':');
+			if (!ptr)
+				continue;
+			ptr++;
+			cache_size = 0;
+			cache_ways = 0;
+			cache_line_size = 0;
+			n = sscanf(ptr, "%dK, %d-way, %db line",
+				&cache_size, &cache_ways, &cache_line_size);
+			if (n != 3)
+				continue;
+			cpu->caches[idx].type = cache_type;
+			cpu->caches[idx].level = cache_level;
+			cpu->caches[idx].size = cache_size * 1024;
+			cpu->caches[idx].ways = cache_ways;
+			cpu->caches[idx].line_size = cache_line_size;
+			idx++;
+		}
+		(void)fclose(fp);
+	}
+
+	if (idx == 0) {
+		free(cpu->caches);
+		cpu->caches = NULL;
+		cpu->cache_count = 0;
+
+		return 0;
+	}
+	cpu->cache_count = idx;
+
+	return idx;
+}
+#endif
+
+#if defined(STRESS_ARCH_SPARC)
 /*
  *  stress_get_cpu_cache_sparc64()
  *	find cache information as provided by linux SPARC64
@@ -482,6 +584,7 @@ static int stress_get_cpu_cache_sparc64(
 
 	return count;
 }
+#endif
 
 /*
  *  index_filter()
@@ -537,8 +640,7 @@ static int stress_get_cpu_cache_index(
 	struct dirent **namelist = NULL;
 	int n;
 	uint32_t i;
-	const size_t cpu_path_len = strlen(cpu_path);
-	char path[cpu_path_len + 128];
+	char path[PATH_MAX];
 
 	(void)stress_mk_filename(path, sizeof(path), cpu_path, SYS_CPU_CACHE_DIR);
 	n = scandir(path, &namelist, index_filter, index_sort);
@@ -561,7 +663,7 @@ static int stress_get_cpu_cache_index(
 
 	for (i = 0; i < cpu->cache_count; i++) {
 		const char *name = namelist[i]->d_name;
-		char fullpath[strlen(path) + strlen(name) + 2];
+		char fullpath[PATH_MAX];
 
 		(void)memset(fullpath, 0, sizeof(fullpath));
 		(void)stress_mk_filename(fullpath, sizeof(fullpath), path, name);
@@ -607,9 +709,16 @@ static void stress_get_cpu_cache_details(stress_cpu_t *cpu, const char *cpu_path
 	if (stress_get_cpu_cache_auxval(cpu) > 0)
 		return;
 
+#if defined(STRESS_ARCH_SPARC)
 	/* Try cache info for sparc CPUs */
 	if (stress_get_cpu_cache_sparc64(cpu, cpu_path) > 0)
 		return;
+#endif
+
+#if defined(STRESS_ARCH_ALPHA)
+	if (stress_get_cpu_cache_alpha(cpu, cpu_path) > 0)
+		return;
+#endif
 
 	return;
 }
@@ -645,8 +754,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 
 	for (i = 0; i < cpu_count; i++) {
 		const char *name = namelist[i]->d_name;
-		const size_t fullpath_len = strlen(SYS_CPU_PREFIX) + strlen(name) + 2;
-		char fullpath[fullpath_len];
+		char fullpath[PATH_MAX];
 		stress_cpu_t *const cpu = &cpus->cpus[i];
 
 		(void)memset(fullpath, 0, sizeof(fullpath));
@@ -656,7 +764,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 			/* 1st CPU cannot be taken offline */
 			cpu->online = 1;
 		} else {
-			char onlinepath[fullpath_len + 8];
+			char onlinepath[PATH_MAX + 8];
 			char tmp[2048];
 
 			(void)memset(onlinepath, 0, sizeof(onlinepath));
