@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2014-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,64 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 #include "git-commit-id.h"
+#include "core-builtin.h"
+#include "core-capabilities.h"
+#include "core-hash.h"
+
+#if defined(HAVE_LINUX_FIEMAP_H)
+#include <linux/fiemap.h>
+#endif
+
+#if defined(HAVE_SYS_AUXV_H)
+#include <sys/auxv.h>
+#endif
+
+#if defined(HAVE_SYS_CAPABILITY_H)
+#include <sys/capability.h>
+#endif
+
+#if defined(HAVE_SYS_LOADAVG_H)
+#include <sys/loadavg.h>
+#endif
+
+#if defined(HAVE_SYS_MOUNT_H)
+#include <sys/mount.h>
+#endif
+
+#if defined(HAVE_SYS_PRCTL_H)
+#include <sys/prctl.h>
+#endif
+
+#if defined(HAVE_SYS_STATVFS_H)
+#include <sys/statvfs.h>
+#endif
+
+#if defined(HAVE_SYS_UTSNAME_H)
+#include <sys/utsname.h>
+#endif
+
+#if defined(HAVE_LINUX_FS_H)
+#include <linux/fs.h>
+#endif
+
+#if defined(HAVE_SYS_VFS_H)
+#include <sys/vfs.h>
+#endif
+
+#if defined(HAVE_LINUX_MAGIC_H)
+#include <linux/magic.h>
+#endif
+
+/* prctl(2) timer slack support */
+#if defined(HAVE_SYS_PRCTL_H) && \
+    defined(HAVE_PRCTL) && \
+    defined(PR_SET_TIMERSLACK) && \
+    defined(PR_GET_TIMERSLACK)
+#define HAVE_PRCTL_TIMER_SLACK
+#endif
 
 #if !defined(PR_SET_DISABLE)
 #define SUID_DUMP_DISABLE	(0)       /* No setuid dumping */
@@ -42,8 +92,231 @@
 int __dso_handle;
 #endif
 
+#define STRESS_ABS_MIN_STACK_SIZE	(64 * 1024)
+
 static bool stress_stack_check_flag;
 
+typedef struct {
+	const unsigned long	fs_magic;
+	const char *		fs_name;
+} stress_fs_name_t;
+
+#if defined(HAVE_LINUX_MAGIC_H) &&	\
+    defined(HAVE_SYS_STATFS_H)
+static const stress_fs_name_t stress_fs_names[] = {
+#if defined(ADFS_SUPER_MAGIC)
+	{ ADFS_SUPER_MAGIC,	"adfs" },
+#endif
+#if defined(AFFS_SUPER_MAGIC)
+	{ AFFS_SUPER_MAGIC,	"affs" },
+#endif
+#if defined(AFS_SUPER_MAGIC)
+	{ AFS_SUPER_MAGIC,	"afs" },
+#endif
+#if defined(AUTOFS_SUPER_MAGIC)
+	{ AUTOFS_SUPER_MAGIC,	"autofs" },
+#endif
+#if defined(CEPH_SUPER_MAGIC)
+	{ CEPH_SUPER_MAGIC,	"ceph" },
+#endif
+#if defined(CODA_SUPER_MAGIC)
+	{ CODA_SUPER_MAGIC,	"coda" },
+#endif
+#if defined(CRAMFS_MAGIC)
+	{ CRAMFS_MAGIC,		"cramfs" },
+#endif
+#if defined(CRAMFS_MAGIC_WEND)
+	{ CRAMFS_MAGIC,		"cramfs" },
+#endif
+#if defined(DEBUGFS_MAGIC)
+	{ DEBUGFS_MAGIC,	"debugfs" },
+#endif
+#if defined(SECURITYFS_MAGIC)
+	{ SECURITYFS_MAGIC,	"securityfs" },
+#endif
+#if defined(SELINUX_MAGIC)
+	{ SELINUX_MAGIC,	"selinux" },
+#endif
+#if defined(SMACK_MAGIC)
+	{ SMACK_MAGIC,		"smack" },
+#endif
+#if defined(RAMFS_MAGIC)
+	{ RAMFS_MAGIC,		"ramfs" },
+#endif
+#if defined(TMPFS_MAGIC)
+	{ TMPFS_MAGIC,		"tmpfs" },
+#endif
+#if defined(HUGETLBFS_MAGIC)
+	{ HUGETLBFS_MAGIC,	"hugetlbfs" },
+#endif
+#if defined(SQUASHFS_MAGIC)
+	{ SQUASHFS_MAGIC,	"squashfs" },
+#endif
+#if defined(ECRYPTFS_SUPER_MAGIC)
+	{ ECRYPTFS_SUPER_MAGIC,	"ecryptfs" },
+#endif
+#if defined(EFS_SUPER_MAGIC)
+	{ EFS_SUPER_MAGIC,	"efs" },
+#endif
+#if defined(EROFS_SUPER_MAGIC_V1)
+	{ EROFS_SUPER_MAGIC_V1,	"erofs" },
+#endif
+#if defined(EXT2_SUPER_MAGIC)
+	{ EXT2_SUPER_MAGIC,	"ext2" },
+#endif
+#if defined(EXT3_SUPER_MAGIC)
+	{ EXT3_SUPER_MAGIC,	"ext3" },
+#endif
+#if defined(XENFS_SUPER_MAGIC)
+	{ XENFS_SUPER_MAGIC,	"xenfs" },
+#endif
+#if defined(EXT4_SUPER_MAGIC)
+	{ EXT4_SUPER_MAGIC,	"ext4" },
+#endif
+#if defined(BTRFS_SUPER_MAGIC)
+	{ BTRFS_SUPER_MAGIC,	"btrfs" },
+#endif
+#if defined(NILFS_SUPER_MAGIC)
+	{ NILFS_SUPER_MAGIC,	"nilfs" },
+#endif
+#if defined(F2FS_SUPER_MAGIC)
+	{ F2FS_SUPER_MAGIC,	"f2fs" },
+#endif
+#if defined(HPFS_SUPER_MAGIC)
+	{ HPFS_SUPER_MAGIC,	"hpfs" },
+#endif
+#if defined(ISOFS_SUPER_MAGIC)
+	{ ISOFS_SUPER_MAGIC,	"isofs" },
+#endif
+#if defined(JFFS2_SUPER_MAGIC)
+	{ JFFS2_SUPER_MAGIC,	"jffs2" },
+#endif
+#if defined(XFS_SUPER_MAGIC)
+	{ XFS_SUPER_MAGIC,	"xfs" },
+#endif
+#if defined(PSTOREFS_MAGIC)
+	{ PSTOREFS_MAGIC,	"pstorefs" },
+#endif
+#if defined(EFIVARFS_MAGIC)
+	{ EFIVARFS_MAGIC,	"efivars" },
+#endif
+#if defined(HOSTFS_SUPER_MAGIC)
+	{ HOSTFS_SUPER_MAGIC,	"hostfs" },
+#endif
+#if defined(OVERLAYFS_SUPER_MAGIC)
+	{ OVERLAYFS_SUPER_MAGIC, "overlayfs" },
+#endif
+#if defined(FUSE_SUPER_MAGIC)
+	{ FUSE_SUPER_MAGIC,	"fuse" },
+#endif
+#if defined(MINIX_SUPER_MAGIC)
+	{ MINIX_SUPER_MAGIC,	"minix" },
+#endif
+#if defined(MINIX_SUPER_MAGIC2)
+	{ MINIX_SUPER_MAGIC2,	"minix" },
+#endif
+#if defined(MINIX2_SUPER_MAGIC)
+	{ MINIX2_SUPER_MAGIC,	"minix2" },
+#endif
+#if defined(MINIX2_SUPER_MAGIC2)
+	{ MINIX2_SUPER_MAGIC2,	"minix2" },
+#endif
+#if defined(MINIX3_SUPER_MAGIC)
+	{ MINIX3_SUPER_MAGIC,	"minix3" },
+#endif
+#if defined(MSDOS_SUPER_MAGIC)
+	{ MSDOS_SUPER_MAGIC,	"msdos" },
+#endif
+#if defined(EXFAT_SUPER_MAGIC)
+	{ EXFAT_SUPER_MAGIC,	"exfat" },
+#endif
+#if defined(NCP_SUPER_MAGIC)
+	{ NCP_SUPER_MAGIC,	"ncp" },
+#endif
+#if defined(NFS_SUPER_MAGIC)
+	{ NFS_SUPER_MAGIC,	"nfs" },
+#endif
+#if defined(OCFS2_SUPER_MAGIC)
+	{ OCFS2_SUPER_MAGIC,	"ocfs2" },
+#endif
+#if defined(OPENPROM_SUPER_MAGIC)
+	{ OPENPROM_SUPER_MAGIC,	"openprom" },
+#endif
+#if defined(QNX4_SUPER_MAGIC)
+	{ QNX4_SUPER_MAGIC,	"qnx4" },
+#endif
+#if defined(QNX6_SUPER_MAGIC)
+	{ QNX6_SUPER_MAGIC,	"qnx6" },
+#endif
+#if defined(AFS_FS_MAGIC)
+	{ AFS_FS_MAGIC,		"afs" },
+#endif
+#if defined(REISERFS_SUPER_MAGIC)
+	{ REISERFS_SUPER_MAGIC,	"reiserfs" },
+#endif
+#if defined(SMB_SUPER_MAGIC)
+	{ SMB_SUPER_MAGIC,	"smb" },
+#endif
+#if defined(CIFS_SUPER_MAGIC)
+	{ CIFS_SUPER_MAGIC,	"cifs" },
+#endif
+#if defined(SMB2_SUPER_MAGIC)
+	{ SMB2_SUPER_MAGIC,	"smb2" },
+#endif
+#if defined(CGROUP_SUPER_MAGIC)
+	{ CGROUP_SUPER_MAGIC,	"cgroup" },
+#endif
+#if defined(CGROUP2_SUPER_MAGIC)
+	{ CGROUP2_SUPER_MAGIC,	"cgroup2" },
+#endif
+#if defined(RDTGROUP_SUPER_MAGIC)
+	{ RDTGROUP_SUPER_MAGIC,	"rdtgroup" },
+#endif
+#if defined(DAXFS_MAGIC)
+	{ DAXFS_MAGIC,		"daxfs" },
+#endif
+#if defined(BINFMTFS_MAGIC)
+	{ BINFMTFS_MAGIC,	"binfmtfs" },
+#endif
+#if defined(DEVPTS_SUPER_MAGIC)
+	{ DEVPTS_SUPER_MAGIC,	"devpts" },
+#endif
+#if defined(BINDERFS_SUPER_MAGIC)
+	{ BINDERFS_SUPER_MAGIC,	"binderfs" },
+#endif
+#if defined(FUTEXFS_SUPER_MAGIC)
+	{ FUTEXFS_SUPER_MAGIC,	"futexfs" },
+#endif
+#if defined(PIPEFS_MAGIC)
+	{ PIPEFS_MAGIC,		"pipefs" },
+#endif
+#if defined(PROC_SUPER_MAGIC)
+	{ PROC_SUPER_MAGIC,	"proc" },
+#endif
+#if defined(SOCKFS_MAGIC)
+	{ SOCKFS_MAGIC,		"sockfs" },
+#endif
+#if defined(SYSFS_MAGIC)
+	{ SYSFS_MAGIC,		"sysfs" },
+#endif
+#if defined(NSFS_MAGIC)
+	{ NSFS_MAGIC,		"nsfs" },
+#endif
+#if defined(BPF_FS_MAGIC)
+	{ BPF_FS_MAGIC,		"bpf_fs" },
+#endif
+#if defined(AAFS_MAGIC)
+	{ AAFS_MAGIC,		"aafs" },
+#endif
+#if defined(ZONEFS_MAGIC)
+	{ ZONEFS_MAGIC,		"zonefs" },
+#endif
+#if defined(UDF_SUPER_MAGIC)
+	{ UDF_SUPER_MAGIC,	"udf" },
+#endif
+};
+#endif
+
 typedef struct {
 	const int  signum;
 	const char *name;
@@ -192,15 +465,10 @@ int stress_set_temp_path(const char *path)
 	stress_temp_path_free();
 
 	stress_temp_path = stress_const_optdup(path);
-	if (!stress_temp_path)
-		return -1;
-
-	if (access(path, R_OK | W_OK) < 0) {
-		(void)fprintf(stderr, "aborting: temp-path '%s' must be readable "
-			"and writeable\n", path);
+	if (!stress_temp_path) {
+		(void)fprintf(stderr, "aborting: cannot allocate memory for '%s'\n", path);
 		return -1;
 	}
-
 	return 0;
 }
 
@@ -215,6 +483,22 @@ const char *stress_get_temp_path(void)
 	return stress_temp_path;
 }
 
+/*
+ *  stress_check_temp_path()
+ *	check if temp path is accessible
+ */
+int stress_check_temp_path(void)
+{
+	const char *path = stress_get_temp_path();
+
+	if (access(path, R_OK | W_OK) < 0) {
+		(void)fprintf(stderr, "aborting: temp-path '%s' must be readable "
+			"and writeable\n", path);
+		return -1;
+	}
+	return 0;
+}
+
 /*
  *  stress_mk_filename()
  *	generate a full file name from a path and filename
@@ -236,10 +520,10 @@ size_t stress_mk_filename(
 }
 
 /*
- *  stress_get_pagesize()
- *	get pagesize
+ *  stress_get_page_size()
+ *	get page_size
  */
-size_t stress_get_pagesize(void)
+size_t stress_get_page_size(void)
 {
 	static size_t page_size = 0;
 
@@ -256,6 +540,8 @@ size_t stress_get_pagesize(void)
 			return page_size;
 		}
 	}
+#else
+	UNEXPECTED
 #endif
 #if defined(HAVE_GETPAGESIZE)
 	{
@@ -289,6 +575,7 @@ int32_t stress_get_processors_online(void)
 		processors_online = 1;
 #else
 	processors_online = 1;
+	UNEXPECTED
 #endif
 	return processors_online;
 }
@@ -310,6 +597,7 @@ int32_t stress_get_processors_configured(void)
 		processors_configured = stress_get_processors_online();
 #else
 	processors_configured = 1;
+	UNEXPECTED
 #endif
 	return processors_configured;
 }
@@ -329,6 +617,7 @@ int32_t stress_get_ticks_per_second(void)
 	ticks_per_second = (int32_t)sysconf(_SC_CLK_TCK);
 	return ticks_per_second;
 #else
+	UNEXPECTED
 	return -1;
 #endif
 }
@@ -373,6 +662,8 @@ void stress_get_memlimits(
 		return;
 	}
 	(void)fclose(fp);
+#else
+	UNEXPECTED
 #endif
 }
 
@@ -390,7 +681,7 @@ uint64_t stress_get_phys_mem_size(void)
 {
 #if defined(STRESS_SC_PAGES)
 	uint64_t phys_pages = 0;
-	const size_t page_size = stress_get_pagesize();
+	const size_t page_size = stress_get_page_size();
 	const uint64_t max_pages = ~0ULL / page_size;
 
 	phys_pages = (uint64_t)sysconf(STRESS_SC_PAGES);
@@ -399,6 +690,7 @@ uint64_t stress_get_phys_mem_size(void)
 		phys_pages = max_pages;
 	return phys_pages * page_size;
 #else
+	UNEXPECTED
 	return 0ULL;
 #endif
 }
@@ -433,6 +725,7 @@ uint64_t stress_get_filesystem_size(void)
 
 	return (uint64_t)buf.f_bsize * blocks;
 #else
+	UNEXPECTED
 	return 0ULL;
 #endif
 }
@@ -459,6 +752,7 @@ uint64_t stress_get_filesystem_available_inodes(void)
 
 	return (uint64_t)buf.f_favail;
 #else
+	UNEXPECTED
 	return 0ULL;
 #endif
 }
@@ -476,6 +770,7 @@ int stress_set_nonblock(const int fd)
 		flags = 0;
 	return fcntl(fd, F_SETFL, O_NONBLOCK | flags);
 #else
+	UNEXPECTED
 	flags = 1;
 	return ioctl(fd, FIOBIO, &flags);
 #endif
@@ -490,7 +785,7 @@ int stress_get_load_avg(
 	double *min5,
 	double *min15)
 {
-#if NEED_GLIBC(2, 2, 0) &&	\
+#if defined(HAVE_GETLOADAVG) &&	\
     !defined(__UCLIBC__)
 	int rc;
 	double loadavg[3];
@@ -538,6 +833,8 @@ void stress_parent_died_alarm(void)
     defined(HAVE_SYS_PRCTL_H) &&	\
     defined(PR_SET_PDEATHSIG)
 	(void)prctl(PR_SET_PDEATHSIG, SIGALRM);
+#else
+	UNEXPECTED
 #endif
 }
 
@@ -565,6 +862,8 @@ int stress_process_dumpable(const bool dumpable)
 		lim.rlim_max = 0;
 		(void)setrlimit(RLIMIT_CORE, &lim);
 	}
+#else
+	UNEXPECTED
 #endif
 
 	/*
@@ -603,6 +902,7 @@ int stress_set_timer_slack_ns(const char *opt)
 	timer_slack = stress_get_uint32(opt);
 	(void)stress_set_setting("timer-slack", TYPE_ID_UINT32, &timer_slack);
 #else
+	UNEXPECTED
 	(void)opt;
 #endif
 	return 0;
@@ -621,6 +921,8 @@ void stress_set_timer_slack(void)
 
 	if (stress_get_setting("timer-slack", &timer_slack))
 		(void)prctl(PR_SET_TIMERSLACK, timer_slack);
+#else
+	UNEXPECTED
 #endif
 }
 
@@ -637,6 +939,7 @@ void stress_set_proc_name_init(int argc, char *argv[], char *envp[])
 	(void)argc;
 	(void)argv;
 	(void)envp;
+	UNEXPECTED
 #endif
 }
 
@@ -664,6 +967,25 @@ void stress_set_proc_name(const char *name)
 #endif
 }
 
+/*
+ *  stress_set_proc_state_str
+ *	set process name based on run state string, see
+ *	macros STRESS_STATE_*
+ */
+void stress_set_proc_state_str(const char *name, const char *str)
+{
+	if (g_opt_flags & OPT_FLAGS_KEEP_NAME)
+		return;
+
+#if defined(HAVE_BSD_UNISTD_H) &&	\
+    defined(HAVE_SETPROCTITLE)
+	setproctitle("-%s [%s]", name, str);
+#else
+	(void)name;
+	(void)str;
+#endif
+}
+
 /*
  *  stress_set_proc_state
  *	set process name based on run state, see
@@ -681,18 +1003,10 @@ void stress_set_proc_state(const char *name, const int state)
 		"wait"
 	};
 
-	(void)name;
-
-	if (g_opt_flags & OPT_FLAGS_KEEP_NAME)
-		return;
-
 	if ((state < 0) || (state >= (int)SIZEOF_ARRAY(stress_states)))
 		return;
 
-#if defined(HAVE_BSD_UNISTD_H) &&	\
-    defined(HAVE_SETPROCTITLE)
-	setproctitle("-%s [%s]", name, stress_states[state]);
-#endif
+	stress_set_proc_state_str(name, stress_states[state]);
 }
 
 /*
@@ -776,15 +1090,15 @@ uint64_t stress_uint64_zero(void)
  */
 static void stress_base36_encode_uint64(char dst[14], uint64_t val)
 {
-        static const char b36[] = "abcdefghijklmnopqrstuvwxyz0123456789";
-        const int b = 36;
-        char *ptr = dst;
+	static const char b36[] = "abcdefghijklmnopqrstuvwxyz0123456789";
+	const int b = 36;
+	char *ptr = dst;
 
-        while (val) {
-                *ptr++ = b36[val % b];
-                val /= b;
-        }
-        *ptr = '\0';
+	while (val) {
+		*ptr++ = b36[val % b];
+		val /= b;
+	}
+	*ptr = '\0';
 }
 
 /*
@@ -913,7 +1227,7 @@ int stress_temp_dir_mk(
 		ret = -errno;
 		pr_fail("%s: mkdir '%s' failed, errno=%d (%s)\n",
 			name, tmp, errno, strerror(errno));
-		(void)unlink(tmp);
+		(void)shim_rmdir(tmp);
 	}
 
 	return ret;
@@ -941,7 +1255,7 @@ int stress_temp_dir_rm(
 	char tmp[PATH_MAX + 1];
 
 	stress_temp_dir(tmp, sizeof(tmp), name, pid, instance);
-	ret = rmdir(tmp);
+	ret = shim_rmdir(tmp);
 	if (ret < 0) {
 		ret = -errno;
 		pr_fail("%s: rmdir '%s' failed, errno=%d (%s)\n",
@@ -1025,6 +1339,49 @@ void stress_strnrnd(char *str, const size_t len)
 	*str = '\0';
 }
 
+/*
+ *  stress_little_endian()
+ *	returns true if CPU is little endian
+ */
+bool stress_little_endian(void)
+{
+	const uint32_t x = 0x12345678;
+	const uint8_t *y = (const uint8_t *)&x;
+
+	return *y == 0x78;
+}
+
+static inline uint32_t OPTIMIZE3 stress_swap32(uint32_t val)
+{
+	register uint32_t swap;
+
+	swap = ((val >> 24) & 0x000000ff) |
+	       ((val << 8)  & 0x00ff0000) |
+	       ((val >> 8)  & 0x0000ff00) |
+	       ((val << 24) & 0xff000000);
+
+	return swap;
+}
+
+/*
+ *  stress_uint8rnd4()
+ *	fill a uint8_t buffer full of random data
+ *	buffer *must* be multiple of 4 bytes in size
+ */
+void OPTIMIZE3 stress_uint8rnd4(uint8_t *data, const size_t len)
+{
+	register uint32_t *ptr32 = (uint32_t *)data;
+	register uint32_t *ptr32end = (uint32_t *)(data + len);
+
+	if (stress_little_endian()) {
+		while (ptr32 < ptr32end)
+			*ptr32++ = stress_mwc32();
+	} else {
+		while (ptr32 < ptr32end)
+			*ptr32++ = stress_swap32(stress_mwc32());
+	}
+}
+
 /*
  *  pr_run_info()
  *	short info about the system we are running stress-ng on
@@ -1115,7 +1472,7 @@ void pr_yaml_runinfo(FILE *yaml)
 		pr_yaml(yaml, "      sysname: %s\n", uts.sysname);
 		pr_yaml(yaml, "      nodename: %s\n", uts.nodename);
 		pr_yaml(yaml, "      release: %s\n", uts.release);
-		pr_yaml(yaml, "      version: %s\n", uts.version);
+		pr_yaml(yaml, "      version: '%s'\n", uts.version);
 		pr_yaml(yaml, "      machine: %s\n", uts.machine);
 	}
 #endif
@@ -1132,7 +1489,7 @@ void pr_yaml_runinfo(FILE *yaml)
 		pr_yaml(yaml, "      freeswap: %lu\n", info.freeswap);
 	}
 #endif
-	pr_yaml(yaml, "      pagesize: %zd\n", stress_get_pagesize());
+	pr_yaml(yaml, "      pagesize: %zd\n", stress_get_page_size());
 	pr_yaml(yaml, "      cpus: %" PRId32 "\n", stress_get_processors_configured());
 	pr_yaml(yaml, "      cpus-online: %" PRId32 "\n", stress_get_processors_online());
 	pr_yaml(yaml, "      ticks-per-second: %" PRId32 "\n", stress_get_ticks_per_second());
@@ -1223,6 +1580,19 @@ init_done:
 			name, errno, strerror(errno));
 		return -1;
 	}
+
+	g_shared->cacheline_size = (size_t)STRESS_PROCS_MAX * sizeof(uint8_t) * 2;
+	g_shared->cacheline =
+		(uint8_t *)mmap(NULL, g_shared->cacheline_size,
+				PROT_READ | PROT_WRITE,
+				MAP_SHARED | MAP_ANONYMOUS, -1, 0);
+
+	if (g_shared->cacheline == MAP_FAILED) {
+		g_shared->cacheline = NULL;
+		pr_err("%s: failed to mmap cacheline buffer, errno=%d (%s)\n",
+			name, errno, strerror(errno));
+		return -1;
+	}
 	if (stress_warn_once())
 		pr_dbg("%s: shared cache buffer size: %" PRIu64 "K\n",
 			name, g_shared->mem_cache_size / 1024);
@@ -1238,6 +1608,8 @@ void stress_cache_free(void)
 {
 	if (g_shared->mem_cache)
 		(void)munmap((void *)g_shared->mem_cache, g_shared->mem_cache_size);
+	if (g_shared->cacheline)
+		(void)munmap((void *)g_shared->cacheline, g_shared->cacheline_size);
 }
 
 /*
@@ -1308,7 +1680,7 @@ bool stress_is_prime64(const uint64_t n)
 		return n >= 2;
 	if ((n % 2 == 0) || (n % 3 == 0))
 		return false;
-	max_d = 1.0 + sqrt((double)n);
+	max_d = 1.0 + shim_sqrt((double)n);
 	max = (uint64_t)max_d;
 	for (i = 5; i < max; i+= 6)
 		if ((n % i == 0) || (n % (i + 2) == 0))
@@ -1319,14 +1691,18 @@ bool stress_is_prime64(const uint64_t n)
 /*
  *  stress_get_prime64()
  *	find a prime that is not a multiple of n,
- *	used for file name striding
+ *	used for file name striding. Minimum is 1009,
+ *	max is unbounded. Return a prime > n, each
+ *	call will return the next prime to keep the
+ *	primes different each call.
  */
 uint64_t stress_get_prime64(const uint64_t n)
 {
-	static uint p = 1009;
+	static uint64_t p = 1009;
+	uint64_t odd_n = (n & ~1UL) + 1;
 
-	if (n != p)
-		return p;
+	if (p < odd_n)
+		p = odd_n;
 
 	/* Search for next prime.. */
 	for (;;) {
@@ -1363,11 +1739,43 @@ size_t stress_get_max_file_limit(void)
 	}
 #else
 	max_sysconf = SIZE_MAX;
+	UNEXPECTED
 #endif
 	/* return the lowest of these two */
 	return STRESS_MINIMUM(max_rlim, max_sysconf);
 }
 
+/*
+ *  stress_get_open_count(void)
+ *	get number of open file descriptors
+ */
+static inline size_t stress_get_open_count(void)
+{
+#if defined(__linux__)
+	DIR *dir;
+	struct dirent *d;
+	size_t n = 0;
+
+	dir = opendir("/proc/self/fd");
+	if (!dir)
+		return (size_t)-1;
+
+	while ((d = readdir(dir)) != NULL) {
+		if (isdigit((int)d->d_name[0]))
+			n++;
+	}
+	(void)closedir(dir);
+
+	/*
+	 * opendir used one extra fd that is now
+	 * closed, so take that off the total
+	 */
+	return (n > 1) ? (n - 1) : n;
+#else
+	return 0;
+#endif
+}
+
 /*
  *  stress_get_file_limit()
  *	get max number of files that the current
@@ -1377,28 +1785,33 @@ size_t stress_get_max_file_limit(void)
 size_t stress_get_file_limit(void)
 {
 	struct rlimit rlim;
-	size_t i, last_opened, opened = 0, max = 65536;	/* initial guess */
+	size_t last_opened, opened, max = 65536;	/* initial guess */
 
 	if (!getrlimit(RLIMIT_NOFILE, &rlim))
 		max = (size_t)rlim.rlim_cur;
 
 	last_opened = 0;
 
-	/* Determine max number of free file descriptors we have */
-	for (i = 0; i < max; i++) {
-		if (fcntl((int)i, F_GETFL) > -1) {
-			opened++;
-			last_opened = i;
-		} else {
-			/*
-			 *  Hack: Over 250 contiguously closed files
-			 *  most probably indicates we're at the point
-			 *  were no more opened file descriptors are
-			 *  going to be found, so bail out rather then
-			 *  scanning for any more opened files
-			 */
-			if (i - last_opened > 250)
-				break;
+	opened = stress_get_open_count();
+	if (opened == 0) {
+		size_t i;
+
+		/* Determine max number of free file descriptors we have */
+		for (i = 0; i < max; i++) {
+			if (fcntl((int)i, F_GETFL) > -1) {
+				opened++;
+				last_opened = i;
+			} else {
+				/*
+				 *  Hack: Over 250 contiguously closed files
+				 *  most probably indicates we're at the point
+				 *  were no more opened file descriptors are
+				 *  going to be found, so bail out rather then
+				 *  scanning for any more opened files
+				 */
+				if (i - last_opened > 250)
+					break;
+			}
 		}
 	}
 	return max - opened;
@@ -1431,10 +1844,36 @@ int stress_get_bad_fd(void)
 		if (fcntl((int)i, F_GETFL) == -1)
 			return i;
 	}
+#else
+	UNEXPECTED
 #endif
 	return -1;
 }
 
+/*
+ *  stress_sigaltstack_no_check()
+ *	attempt to set up an alternative signal stack with no
+ *	minimum size check on stack
+ *	  stack - must be at least MINSIGSTKSZ
+ *	  size  - size of stack (- STACK_ALIGNMENT)
+ */
+int stress_sigaltstack_no_check(void *stack, const size_t size)
+{
+#if defined(HAVE_SIGALTSTACK)
+	stack_t ss;
+
+	ss.ss_sp = (void *)stack;
+	ss.ss_size = size;
+	ss.ss_flags = 0;
+	return sigaltstack(&ss, NULL);
+#else
+	UNEXPECTED
+	(void)stack;
+	(void)size;
+	return 0;
+#endif
+}
+
 /*
  *  stress_sigaltstack()
  *	attempt to set up an alternative signal stack
@@ -1444,22 +1883,19 @@ int stress_get_bad_fd(void)
 int stress_sigaltstack(void *stack, const size_t size)
 {
 #if defined(HAVE_SIGALTSTACK)
-	stack_t ss;
-
 	if (size < (size_t)STRESS_MINSIGSTKSZ) {
 		pr_err("sigaltstack stack size %zu must be more than %zuK\n",
 			size, (size_t)STRESS_MINSIGSTKSZ / 1024);
 		return -1;
 	}
-	ss.ss_sp = (void *)stack;
-	ss.ss_size = size;
-	ss.ss_flags = 0;
-	if (sigaltstack(&ss, NULL) < 0) {
+
+	if (stress_sigaltstack_no_check(stack, size) < 0) {
 		pr_fail("sigaltstack failed: errno=%d (%s)\n",
 			errno, strerror(errno));
 		return -1;
 	}
 #else
+	UNEXPECTED
 	(void)stack;
 	(void)size;
 #endif
@@ -1593,10 +2029,19 @@ unsigned int stress_get_cpu(void)
  */
 const char *stress_get_compiler(void)
 {
-#if defined(__clang_major__) &&	\
-    defined(__clang_minor__)
+#if   defined(__ICC) && 		\
+      defined(__INTEL_COMPILER) &&	\
+      defined(__INTEL_COMPILER_UPDATE)
+	static const char cc[] = "icc " XSTRINGIFY(__INTEL_COMPILER) "." XSTRINGIFY(__INTEL_COMPILER_UPDATE) "";
+#elif defined(__TINYC__)
+	static const char cc[] = "tcc " XSTRINGIFY(__TINYC__) "";
+#elif defined(__PCC__) &&			\
+       defined(__PCC_MINOR__)
+	static const char cc[] = "pcc " XSTRINGIFY(__PCC__) "." XSTRINGIFY(__PCC_MINOR__) "." XSTRINGIFY(__PCC_MINORMINOR__) "";
+#elif defined(__clang_major__) &&	\
+      defined(__clang_minor__)
 	static const char cc[] = "clang " XSTRINGIFY(__clang_major__) "." XSTRINGIFY(__clang_minor__) "";
-#elif defined(__GNUC__) &&	\
+#elif defined(__GNUC__) &&		\
       defined(__GNUC_MINOR__)
 	static const char cc[] = "gcc " XSTRINGIFY(__GNUC__) "." XSTRINGIFY(__GNUC_MINOR__) "";
 #else
@@ -1623,6 +2068,8 @@ const char *stress_get_uname_info(void)
 		(void)snprintf(str, sizeof(str), "%s %s %s", buf.machine, buf.sysname, buf.release);
 		return str;
 	}
+#else
+	UNEXPECTED
 #endif
 	return "unknown";
 }
@@ -1699,7 +2146,7 @@ size_t stress_probe_max_pipe_size(void)
 		return max_pipe_size;
 
 #if defined(F_SETPIPE_SZ)
-	page_size = stress_get_pagesize();
+	page_size = stress_get_page_size();
 
 	/*
 	 *  Try and find maximum pipe size directly
@@ -1732,7 +2179,7 @@ size_t stress_probe_max_pipe_size(void)
 ret:
 	max_pipe_size = sz;
 #else
-	max_pipe_size = stress_get_pagesize();
+	max_pipe_size = stress_get_page_size();
 #endif
 	return max_pipe_size;
 }
@@ -1805,7 +2252,7 @@ char *stress_uint64_to_str(char *str, size_t len, const uint64_t val)
  *  stress_check_root()
  *	returns true if root
  */
-static bool stress_check_root(void)
+static inline bool stress_check_root(void)
 {
 	return (geteuid() == 0);
 }
@@ -1952,30 +2399,30 @@ size_t stress_text_addr(char **start, char **end)
 	extern char __executable_start;
 	intptr_t text_start = (intptr_t)&__executable_start;
 #elif defined(__APPLE__)
-        extern char _mh_execute_header;
-        intptr_t text_start = (intptr_t)&_mh_execute_header;
+	extern char _mh_execute_header;
+	intptr_t text_start = (intptr_t)&_mh_execute_header;
 #elif defined(__OpenBSD__)
-        extern char _start[];
-        intptr_t text_start = (intptr_t)&_start[0];
+	extern char _start[];
+	intptr_t text_start = (intptr_t)&_start[0];
 #elif defined(__TINYC__)
-        extern char _start;
-        intptr_t text_start = (intptr_t)&_start;
+	extern char _start;
+	intptr_t text_start = (intptr_t)&_start;
 #else
-        extern char _start;
-        intptr_t text_start = (intptr_t)&_start;
+	extern char _start;
+	intptr_t text_start = (intptr_t)&_start;
 #endif
 
 #if defined(__APPLE__)
-        extern void *get_etext(void);
-        intptr_t text_end = (intptr_t)get_etext();
+	extern void *get_etext(void);
+	intptr_t text_end = (intptr_t)get_etext();
 #elif defined(__TINYC__)
-        extern char _etext;
-        intptr_t text_end = (intptr_t)&_etext;
+	extern char _etext;
+	intptr_t text_end = (intptr_t)&_etext;
 #else
-        extern char etext;
-        intptr_t text_end = (intptr_t)&etext;
+	extern char etext;
+	intptr_t text_end = (intptr_t)&etext;
 #endif
-        const size_t text_len = (size_t)(text_end - text_start);
+	const size_t text_len = (size_t)(text_end - text_start);
 
 	if ((start == NULL) || (end == NULL) || (text_start >= text_end))
 		return 0;
@@ -2000,6 +2447,7 @@ bool stress_is_dev_tty(const int fd)
 		return true;
 	return !strncmp("/dev/tty", name, 8);
 #else
+	UNEXPECTED
 	(void)fd;
 
 	/* Assume it is */
@@ -2058,15 +2506,10 @@ bool stress_warn_once_hash(const char *filename, const int line)
 {
 	uint32_t free_slot, i, j, h = (stress_hash_pjw(filename) + (uint32_t)line);
 	bool not_warned_yet = true;
-#if defined(HAVE_LIB_PTHREAD)
-        int ret;
-#endif
 	if (!g_shared)
 		return true;
 
-#if defined(HAVE_LIB_PTHREAD)
-        ret = shim_pthread_spin_lock(&g_shared->warn_once.lock);
-#endif
+	stress_lock_acquire(g_shared->warn_once.lock);
 	free_slot = STRESS_WARN_HASH_MAX;
 
 	/*
@@ -2092,11 +2535,9 @@ bool stress_warn_once_hash(const char *filename, const int line)
 		g_shared->warn_once.hash[free_slot] = h;
 	}
 unlock:
-#if defined(HAVE_LIB_PTHREAD)
-        if (!ret)
-                shim_pthread_spin_unlock(&g_shared->warn_once.lock);
-#endif
-        return not_warned_yet;
+	stress_lock_release(g_shared->warn_once.lock);
+
+	return not_warned_yet;
 }
 
 /*
@@ -2242,8 +2683,6 @@ ignore_eintr:
 			rbytes += ret;
 	} while (ret > 0 && (rbytes != size));
 
-	pr_dbg_v("stress_read_buffer: size=%ld read=%ld sz2=%ld\n", size, rbytes, ret);
-
 	return (ret <= 0)? ret : rbytes;
 }
 
@@ -2268,8 +2707,6 @@ ignore_eintr:
 			wbytes += ret;
 	} while (ret > 0 && (wbytes != size));
 
-	pr_dbg_v("stress_write_buffer: size=%ld written=%ld sz2=%ld\n", size, wbytes, ret);
-
 	return (ret <= 0)? ret : wbytes;
 }
 
@@ -2301,6 +2738,7 @@ int stress_get_kernel_release(void)
 
 	return stress_kernel_release(major, minor, patchlevel);
 #else
+	UNEXPECTED
 	return -1;
 #endif
 }
@@ -2382,9 +2820,9 @@ int stress_read_fdinfo(const pid_t pid, const int fd)
 	char buf[4096];
 
 	(void)snprintf(path, sizeof(path), "/proc/%d/fdinfo/%d",
-                (int)pid, fd);
+		(int)pid, fd);
 
-        return (int)system_read(path, buf, sizeof(buf));
+	return (int)system_read(path, buf, sizeof(buf));
 #else
 	(void)pid;
 	(void)fd;
@@ -2419,12 +2857,15 @@ size_t stress_hostname_length(void)
  */
 static inline long stress_min_aux_sig_stack_size(void)
 {
-#if defined(HAVE_GETAUXVAL) &&	\
+#if defined(HAVE_SYS_AUXV_H) && \
+    defined(HAVE_GETAUXVAL) &&	\
     defined(AT_MINSIGSTKSZ)
-	long sz = getauxval(AT_MINSIGSTKSZ);
+	long sz = (long)getauxval(AT_MINSIGSTKSZ);
 
 	if (sz > 0)
 		return sz;
+#else
+	UNEXPECTED
 #endif
 	return -1;
 }
@@ -2440,21 +2881,22 @@ size_t stress_sig_stack_size(void)
 
 	/* return cached copy */
 	if (sz > 0)
-		return sz;
+		return (size_t)sz;
 
 	min = stress_min_aux_sig_stack_size();
 #if defined(_SC_SIGSTKSZ)
 	sz = sysconf(_SC_SIGSTKSZ);
-	if (sz > min)
+	if (min < sz)
 		min = sz;
 #endif
 #if defined(SIGSTKSZ)
-	if (SIGSTKSZ > min)
+	if (SIGSTKSZ > min) {
+		/* SIGSTKSZ may be sysconf(_SC_SIGSTKSZ) */
 		min = SIGSTKSZ;
-#else
-	if (8192 > min)
-		min = 8192;
+	}
 #endif
+	if (min < STRESS_ABS_MIN_STACK_SIZE)
+		min = STRESS_ABS_MIN_STACK_SIZE;
 	sz = min;
 
 	return (size_t)sz;
@@ -2470,21 +2912,22 @@ size_t stress_min_sig_stack_size(void)
 
 	/* return cached copy */
 	if (sz > 0)
-		return sz;
+		return (size_t)sz;
 
 	min = stress_min_aux_sig_stack_size();
 #if defined(_SC_MINSIGSTKSZ)
 	sz = sysconf(_SC_MINSIGSTKSZ);
-	if (sz > min)
+	if (min < sz)
 		min = sz;
 #endif
 #if defined(SIGSTKSZ)
-	if (SIGSTKSZ > min)
+	if (SIGSTKSZ > min) {
+		/* SIGSTKSZ may be sysconf(_SC_SIGSTKSZ) */
 		min = SIGSTKSZ;
-#else
-	if (8192 > min)
-		min = 8192;
+	}
 #endif
+	if (min < STRESS_ABS_MIN_STACK_SIZE)
+		min = STRESS_ABS_MIN_STACK_SIZE;
 	sz = min;
 
 	return (size_t)sz;
@@ -2500,7 +2943,7 @@ size_t stress_min_pthread_stack_size(void)
 
 	/* return cached copy */
 	if (sz > 0)
-		return sz;
+		return (size_t)sz;
 
 	min = stress_min_aux_sig_stack_size();
 #if defined(__SC_THREAD_STACK_MIN_VALUE)
@@ -2548,8 +2991,8 @@ extern void __stack_chk_fail(void);
 NORETURN WEAK void __stack_chk_fail(void)
 {
 	if (stress_stack_check_flag) {
-		fprintf(stderr, "Stack overflow detected! Aborting stress-ng.\n");
-		fflush(stderr);
+		(void)fprintf(stderr, "Stack overflow detected! Aborting stress-ng.\n");
+		(void)fflush(stderr);
 		abort();
 	}
 	/* silently exit */
@@ -2582,6 +3025,7 @@ int stress_tty_width(void)
 		return max_width;
 	return ret;
 #else
+	UNEXPECTED
 	return max_width;
 #endif
 }
@@ -2605,6 +3049,7 @@ size_t stress_get_extents(const int fd)
 
 	return fiemap.fm_mapped_extents;
 #else
+	UNEXPECTED
 	(void)fd;
 
 	return 0;
@@ -2627,3 +3072,117 @@ bool stress_redo_fork(const int err)
 	}
 	return false;
 }
+
+/*
+ *  stress_sighandler_nop()
+ *	no-operation signal handler
+ */
+void stress_sighandler_nop(int sig)
+{
+	(void)sig;
+}
+
+/*
+ *  stress_clear_warn_once()
+ *	clear the linux warn once warnings flag, kernel warn once
+ *	messages can be re-issued
+ */
+void stress_clear_warn_once(void)
+{
+#if defined(__linux__)
+	if (stress_check_capability(SHIM_CAP_IS_ROOT))
+		(void)system_write("/sys/kernel/debug/clear_warn_once", "1", 1);
+#endif
+}
+
+/*
+ *  stress_flag_permutation()
+ *	given flag mask in flags, generate all possible permutations
+ *	of bit flags. e.g.
+ *		flags = 0x81;
+ *			-> b00000000
+ *			   b00000001
+ *			   b10000000
+ *			   b10000001
+ */
+size_t stress_flag_permutation(const int flags, int **permutations)
+{
+	unsigned int flag_bits;
+	unsigned int n_bits;
+	register unsigned int j, n_flags;
+	int *perms;
+
+	*permutations = NULL;
+
+	for (n_bits = 0, flag_bits = (unsigned int)flags; flag_bits; flag_bits >>= 1U)
+		n_bits += (flag_bits & 1U);
+
+	n_flags = 1U << n_bits;
+	perms = calloc((size_t)n_flags, sizeof(*perms));
+	if (!perms)
+		return 0;
+
+	/*
+	 *  Generate all the possible flag settings in order
+	 */
+	for (j = 0; j < n_flags; j++) {
+		register int i;
+		register unsigned int j_mask = 1U;
+
+		for (i = 0; i < 32; i++) {
+			const int i_mask = (int)(1U << i);
+
+			if (flags & i_mask) {
+				if (j & j_mask)
+					perms[j] |= i_mask;
+				j_mask <<= 1U;
+			}
+		}
+	}
+	*permutations = perms;
+	return (size_t)n_flags;
+}
+
+/*
+ *  stress_fs_magic_to_name()
+ *	return the human readable file system type based on fs type magic
+ */
+const char *stress_fs_magic_to_name(const unsigned long fs_magic)
+{
+	static char unknown[32];
+#if defined(HAVE_LINUX_MAGIC_H) &&	\
+    defined(HAVE_SYS_STATFS_H)
+	size_t i;
+
+	for (i = 0; i < SIZEOF_ARRAY(stress_fs_names); i++) {
+		if (stress_fs_names[i].fs_magic == fs_magic)
+			return stress_fs_names[i].fs_name;
+	}
+#endif
+	(void)snprintf(unknown, sizeof(unknown), "unknown 0x%lx", fs_magic);
+
+	return unknown;
+}
+
+/*
+ *  stress_fs_type()
+ *	return the file system type that the given filename is in
+ */
+const char *stress_fs_type(const char *filename)
+{
+#if defined(HAVE_LINUX_MAGIC_H) &&	\
+    defined(HAVE_SYS_STATFS_H)
+	struct statfs buf;
+	static char tmp[64];
+
+	if (statfs(filename, &buf) != 0) {
+		return "";
+	}
+	(void)snprintf(tmp, sizeof(tmp), ", filesystem type: %s", stress_fs_magic_to_name((unsigned long)buf.f_type));
+	return tmp;
+#else
+	(void)filename;
+
+	return "";
+#endif
+}
