@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,17 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
+#if defined(HAVE_LINUX_IF_ALG_H)
+#include <linux/if_alg.h>
+#endif
+
+#if defined(HAVE_LINUX_SOCKET_H)
+#include <linux/socket.h>
+#endif
+
 #define ALLOC_SLOP	(64)
 
 static const stress_help_t help[] = {
@@ -129,8 +132,8 @@ static void stress_af_alg_add_crypto_defconfigs(void);
  */
 static stress_crypto_type_t name_to_type(const char *buffer)
 {
-	char *ptr = strchr(buffer, ':');
-        const char *end = ptr + strlen(buffer);
+	const char *ptr = strchr(buffer, ':');
+	const char *end = ptr + strlen(buffer);
 	size_t i;
 
 	if (!ptr)
@@ -547,7 +550,7 @@ retry_bind:
 			goto err_close;
 		} else {
 			if (memcmp(input, output, DATA_LEN)) {
-				pr_err("%s: decrypted data "
+				pr_fail("%s: decrypted data "
 					"different from original data "
 					"using kernel crypto engine %s "
 					"(possible kernel bug)\n",
@@ -726,26 +729,26 @@ static void stress_af_alg_dump_crypto_list(void)
 	for (ci = crypto_info_list; ci; ci = ci->next) {
 		if (ci->internal)
 			continue;
-		fprintf(stdout, "{ .crypto_type = %s, .type = \"%s\", .name = \"%s\"",
+		(void)fprintf(stdout, "{ .crypto_type = %s, .type = \"%s\", .name = \"%s\"",
 			type_to_type_string(ci->crypto_type), ci->type, ci->name);
 		if (ci->block_size)
-			fprintf(stdout, ",\t.block_size = %d",
+			(void)fprintf(stdout, ",\t.block_size = %d",
 				ci->block_size);
 		if (ci->max_key_size)
-			fprintf(stdout, ",\t.max_key_size = %d",
+			(void)fprintf(stdout, ",\t.max_key_size = %d",
 				ci->max_key_size);
 		if (ci->max_auth_size)
-			fprintf(stdout, ",\t.max_auth_size = %d",
+			(void)fprintf(stdout, ",\t.max_auth_size = %d",
 				ci->max_auth_size);
 		if (ci->iv_size)
-			fprintf(stdout, ",\t.iv_size = %d",
+			(void)fprintf(stdout, ",\t.iv_size = %d",
 				ci->iv_size);
 		if (ci->digest_size)
-			fprintf(stdout, ",\t.digest_size = %d",
+			(void)fprintf(stdout, ",\t.digest_size = %d",
 				ci->digest_size);
-		fprintf(stdout, " },\n");
+		(void)fprintf(stdout, " },\n");
 	}
-	fflush(stdout);
+	(void)fflush(stdout);
 }
 
 /*
@@ -866,7 +869,7 @@ static int stress_af_alg(const stress_args_t *args)
  */
 static char *dup_field(const char *buffer)
 {
-	char *ptr = strchr(buffer, ':');
+	const char *ptr = strchr(buffer, ':');
 	char *eol = strchr(buffer, '\n');
 
 	if (!ptr)
@@ -921,16 +924,23 @@ static bool stress_af_alg_add_crypto(stress_crypto_info_t *info)
 	if ((info->name == NULL) || (info->type == NULL))
 		return false;
 
+	/*
+	 * Deprecated in Linux 5.9
+	 * see commit 9ace6771831017ce75a2bdf03c284b686dd39dba
+         */
+	if (strcmp(info->name, "ecb(arc4)") == 0)
+		return false;
+
 	/* Scan for duplications */
 	for (ci = crypto_info_list; ci; ci = ci->next) {
-		if (strcmp(ci->name, info->name) == 0 &&
-		    strcmp(ci->type, info->type) == 0 &&
-		    ci->block_size == info->block_size &&
-		    ci->max_key_size == info->max_key_size &&
-		    ci->max_auth_size == info->max_auth_size &&
-		    ci->iv_size == info->iv_size &&
-		    ci->digest_size == info->digest_size &&
-		    ci->internal == info->internal)
+		if ((strcmp(ci->name, info->name) == 0) &&
+		    (strcmp(ci->type, info->type) == 0) &&
+		    (ci->block_size == info->block_size) &&
+		    (ci->max_key_size == info->max_key_size) &&
+		    (ci->max_auth_size == info->max_auth_size) &&
+		    (ci->iv_size == info->iv_size) &&
+		    (ci->digest_size == info->digest_size) &&
+		    (ci->internal == info->internal))
 			return false;
 	}
 	/*
@@ -1060,6 +1070,7 @@ stressor_info_t stress_af_alg_info = {
 	.deinit = stress_af_alg_deinit,
 	.class = CLASS_CPU | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
 
