@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,27 +16,36 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
+#include "core-capabilities.h"
 
-#if defined(__linux__) && defined(HAVE_APPARMOR)
-
-#include <sys/select.h>
+#if defined(HAVE_SYS_APPARMOR_H)
 #include <sys/apparmor.h>
+#endif
+
+#if defined(HAVE_SYS_SELECT_H)
+#include <sys/select.h>
+#endif
+
+static const stress_help_t help[] = {
+	{ NULL,	"apparmor",	  "start N workers exercising AppArmor interfaces" },
+	{ NULL,	"apparmor-ops N", "stop after N bogo AppArmor worker bogo operations" },
+	{ NULL,	NULL,		  NULL }
+};
+
+#if defined(HAVE_APPARMOR) &&		\
+    defined(HAVE_SYS_APPARMOR_H) && 	\
+    defined(HAVE_SYS_SELECT_H)
 
 #define APPARMOR_BUF_SZ	(4096)
 
-typedef int (*apparmor_func)(const char *name,
-			     const uint64_t max_ops, uint64_t *counter);
+typedef int (*stress_apparmor_func)(const stress_args_t *args);
 
 static volatile bool apparmor_run = true;
 static char *apparmor_path = NULL;
+static void *counter_lock;
+static char *data_copy, *data_prev;
 
 extern char g_apparmor_data[];
 extern const size_t g_apparmor_data_len;
@@ -44,38 +54,45 @@ extern const size_t g_apparmor_data_len;
  *  stress_apparmor_supported()
  *      check if AppArmor is supported
  */
-int stress_apparmor_supported(void)
+static int stress_apparmor_supported(const char *name)
 {
 	int fd;
 	char path[PATH_MAX];
 
+	if (!stress_check_capability(SHIM_CAP_MAC_ADMIN)) {
+		pr_inf_skip("%s stressor will be skipped, "
+			"need to be running with CAP_SYS_ADMIN "
+			"rights for this stressor\n", name);
+		return -1;
+	}
+
 	/* Initial sanity checks for AppArmor */
 	if (!aa_is_enabled()) {
-		pr_inf("apparmor stressor will be skipped, "
+		pr_inf_skip("apparmor stressor will be skipped, "
 			"AppArmor is not enabled\n");
 		return -1;
 	}
 	if (aa_find_mountpoint(&apparmor_path) < 0) {
-		pr_inf("apparmor stressor will be skipped, "
+		pr_inf_skip("apparmor stressor will be skipped, "
 			"cannot get AppArmor path, errno=%d (%s)\n",
 			errno, strerror(errno));
 		return -1;
 	}
 	/* ..and see if profiles are accessible */
-	(void)snprintf(path, sizeof(path), "%s/%s", apparmor_path, "profiles");
+	(void)stress_mk_filename(path, sizeof(path), apparmor_path, "profiles");
 	if ((fd = open(path, O_RDONLY)) < 0) {
 		switch (errno) {
 		case EACCES:
-			pr_inf("apparmor stressor will be skipped, "
-				"stress-ng needs to be run with root "
+			pr_inf_skip("apparmor stressor will be skipped, "
+				"stress-ng needs CAP_MAC_ADMIN "
 				"privilege to access AppArmor /sys files.\n");
 			break;
 		case ENOENT:
-			pr_inf("apparmor stressor will be skipped, "
+			pr_inf_skip("apparmor stressor will be skipped, "
 				"AppArmor /sys files do not exist\n");
 			break;
 		default:
-			pr_inf("apparmor stressor will be skipped, "
+			pr_inf_skip("apparmor stressor will be skipped, "
 				"cannot access AppArmor /sys files: "
 				"errno=%d (%s)\n", errno, strerror(errno));
 			break;
@@ -89,23 +106,38 @@ int stress_apparmor_supported(void)
 	return 0;
 }
 
+static bool stress_apparmor_keep_stressing_inc(const stress_args_t *args, bool inc)
+{
+	/* fast check */
+	if (!apparmor_run)
+		return false;
+
+	return inc_counter_lock(args, counter_lock, inc);
+}
 
 /*
  *  stress_apparmor_handler()
  *      signal handler
  */
-static void MLOCKED stress_apparmor_handler(int dummy)
+static void MLOCKED_TEXT stress_apparmor_alrm_handler(int signum)
 {
-	(void)dummy;
+	(void)signum;
 
 	apparmor_run = false;
 }
 
+static void MLOCKED_TEXT stress_apparmor_usr1_handler(int signum)
+{
+	(void)signum;
+}
+
 /*
  *  stress_apparmor_read()
  *	read a proc file
  */
-static void stress_apparmor_read(const char *path)
+static void stress_apparmor_read(
+	const stress_args_t *args,
+	const char *path)
 {
 	int fd;
 	ssize_t i = 0;
@@ -116,12 +148,12 @@ static void stress_apparmor_read(const char *path)
 	/*
 	 *  Multiple randomly sized reads
 	 */
-	while (apparmor_run && (i < (4096 * APPARMOR_BUF_SZ))) {
-		ssize_t ret, sz = 1 + (mwc32() % sizeof(buffer));
+	while (i < (4096 * APPARMOR_BUF_SZ)) {
+		ssize_t ret, sz = 1 + (stress_mwc32() % sizeof(buffer));
 redo:
-		if (!g_keep_stressing_flag)
+		if (!stress_apparmor_keep_stressing_inc(args, false))
 			break;
-		ret = read(fd, buffer, sz);
+		ret = read(fd, buffer, (size_t)sz);
 		if (ret < 0) {
 			if ((errno == EAGAIN) || (errno == EINTR))
 				goto redo;
@@ -139,6 +171,7 @@ redo:
  *	read directory
  */
 static void stress_apparmor_dir(
+	const stress_args_t *args,
 	const char *path,
 	const bool recurse,
 	const int depth)
@@ -146,7 +179,7 @@ static void stress_apparmor_dir(
 	DIR *dp;
 	struct dirent *d;
 
-	if (!g_keep_stressing_flag)
+	if (!stress_apparmor_keep_stressing_inc(args, false))
 		return;
 
 	/* Don't want to go too deep */
@@ -160,23 +193,20 @@ static void stress_apparmor_dir(
 	while ((d = readdir(dp)) != NULL) {
 		char name[PATH_MAX];
 
-		if (!g_keep_stressing_flag)
+		if (!stress_apparmor_keep_stressing_inc(args, false))
 			break;
-		if (!strcmp(d->d_name, ".") ||
-		    !strcmp(d->d_name, ".."))
+		if (stress_is_dot_filename(d->d_name))
 			continue;
 		switch (d->d_type) {
 		case DT_DIR:
 			if (recurse) {
-				(void)snprintf(name, sizeof(name),
-					"%s/%s", path, d->d_name);
-				stress_apparmor_dir(name, recurse, depth + 1);
+				(void)stress_mk_filename(name, sizeof(name), path, d->d_name);
+				stress_apparmor_dir(args, name, recurse, depth + 1);
 			}
 			break;
 		case DT_REG:
-			(void)snprintf(name, sizeof(name),
-				"%s/%s", path, d->d_name);
-			stress_apparmor_read(name);
+			(void)stress_mk_filename(name, sizeof(name), path, d->d_name);
+			stress_apparmor_read(args, name);
 			break;
 		default:
 			break;
@@ -185,62 +215,62 @@ static void stress_apparmor_dir(
 	(void)closedir(dp);
 }
 
-
 /*
  *  apparmor_spawn()
  *	spawn a process
  */
 static pid_t apparmor_spawn(
-	const char *name,
-	const uint64_t max_ops,
-	uint64_t *counter,
-	apparmor_func func)
+	const stress_args_t *args,
+	stress_apparmor_func func)
 {
 	pid_t pid;
 
 again:
 	pid = fork();
 	if (pid < 0) {
-		if (g_keep_stressing_flag && (errno == EAGAIN))
+		if (stress_redo_fork(errno))
 			goto again;
 		return -1;
 	}
 	if (pid == 0) {
-		int ret;
+		int ret = EXIT_SUCCESS;
+
+		if (!stress_apparmor_keep_stressing_inc(args, false))
+			goto abort;
+
+		if (stress_sighandler(args->name, SIGALRM,
+				      stress_apparmor_alrm_handler, NULL) < 0) {
+			_exit(EXIT_FAILURE);
+		}
 
-		if (stress_sighandler(name, SIGALRM,
-				stress_apparmor_handler, NULL) < 0)
-			exit(EXIT_FAILURE);
 		(void)setpgid(0, g_pgrp);
+		(void)sched_settings_apply(true);
 		stress_parent_died_alarm();
-
-		ret = func(name, max_ops, counter);
+		if (!stress_apparmor_keep_stressing_inc(args, false))
+			goto abort;
+		ret = func(args);
+abort:
 		free(apparmor_path);
-		exit(ret);
+		(void)kill(args->pid, SIGUSR1);
+		_exit(ret);
 	}
 	(void)setpgid(pid, g_pgrp);
 	return pid;
 }
 
-
 /*
  *  apparmor_stress_profiles()
  *	hammer profile reading
  */
-static int apparmor_stress_profiles(
-	const char *name,
-	const uint64_t max_ops,
-	uint64_t *counter)
+static int apparmor_stress_profiles(const stress_args_t *args)
 {
 	char path[PATH_MAX];
 
-	(void)name;
-	(void)snprintf(path, sizeof(path), "%s/%s", apparmor_path, "profiles");
+	(void)stress_mk_filename(path, sizeof(path), apparmor_path, "profiles");
 
 	do {
-		stress_apparmor_read(path);
-		(*counter)++;
-	} while (apparmor_run && (!max_ops || *counter < max_ops));
+		stress_apparmor_read(args, path);
+	} while (stress_apparmor_keep_stressing_inc(args, true));
 
 	return EXIT_SUCCESS;
 }
@@ -249,20 +279,15 @@ static int apparmor_stress_profiles(
  *  apparmor_stress_features()
  *	hammer features reading
  */
-static int apparmor_stress_features(
-	const char *name,
-	const uint64_t max_ops,
-	uint64_t *counter)
+static int apparmor_stress_features(const stress_args_t *args)
 {
 	char path[PATH_MAX];
 
-	(void)name;
-	(void)snprintf(path, sizeof(path), "%s/%s", apparmor_path, "features");
+	(void)stress_mk_filename(path, sizeof(path), apparmor_path, "features");
 
 	do {
-		stress_apparmor_dir(path, true, 0);
-		(*counter)++;
-	} while (apparmor_run && (!max_ops || *counter < max_ops));
+		stress_apparmor_dir(args, path, true, 0);
+	} while (stress_apparmor_keep_stressing_inc(args, true));
 
 	return EXIT_SUCCESS;
 }
@@ -271,10 +296,7 @@ static int apparmor_stress_features(
  *  apparmor_stress_kernel_interface()
  *	load/replace/unload stressing
  */
-static int apparmor_stress_kernel_interface(
-	const char *name,
-	const uint64_t max_ops,
-	uint64_t *counter)
+static int apparmor_stress_kernel_interface(const stress_args_t *args)
 {
 	int rc = EXIT_SUCCESS;
 	aa_kernel_interface *kern_if;
@@ -285,9 +307,8 @@ static int apparmor_stress_kernel_interface(
 	do {
 		int ret = aa_kernel_interface_new(&kern_if, NULL, NULL);
 		if (ret < 0) {
-			pr_fail("%s: aa_kernel_interface_new() failed, "
-				"errno=%d (%s)\n", name,
-				errno, strerror(errno));
+			pr_fail("%s: aa_kernel_interface_new() failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			rc = EXIT_FAILURE;
 			break;
 		}
@@ -300,8 +321,8 @@ static int apparmor_stress_kernel_interface(
 			g_apparmor_data, g_apparmor_data_len);
 		if (ret < 0) {
 			if (errno != EEXIST) {
-				pr_fail("%s: aa_kernel_interface_load_policy() failed, "
-					"errno=%d (%s)\n", name, errno, strerror(errno));
+				pr_fail("%s: aa_kernel_interface_load_policy() failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				rc = EXIT_FAILURE;
 			}
 		}
@@ -313,15 +334,9 @@ static int apparmor_stress_kernel_interface(
 		 */
 		ret = aa_kernel_interface_replace_policy(kern_if,
 			g_apparmor_data, g_apparmor_data_len);
-		if (ret < 0) {
+		if (ret < 0)
 			aa_kernel_interface_unref(kern_if);
 
-			pr_fail("%s: aa_kernel_interface_replace_policy() failed, "
-				"errno=%d (%s)\n", name, errno,
-				strerror(errno));
-			rc = EXIT_FAILURE;
-		}
-
 		/*
 		 *  Removal may fail if another stressor has already removed the
 		 *  policy, so we may get ENOENT
@@ -330,16 +345,17 @@ static int apparmor_stress_kernel_interface(
 			"/usr/bin/pulseaudio-eg");
 		if (ret < 0) {
 			if (errno != ENOENT) {
-				pr_fail("%s: aa_kernel_interface_remove_policy() failed, "
-					"errno=%d (%s)\n", name, errno,
-					strerror(errno));
+				aa_kernel_interface_unref(kern_if);
+
+				pr_fail("%s: aa_kernel_interface_remove_policy() failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				rc = EXIT_FAILURE;
+				break;
 			}
 		}
 		aa_kernel_interface_unref(kern_if);
 
-		(*counter)++;
-	} while (g_keep_stressing_flag && apparmor_run && (!max_ops || *counter < max_ops));
+	} while (stress_apparmor_keep_stressing_inc(args, true));
 
 	return rc;
 }
@@ -352,10 +368,10 @@ static inline void apparmor_corrupt_flip_bits_random(
 	char *copy, const size_t len)
 {
 	uint32_t i;
-	const uint32_t n = mwc32() % 17;
+	const uint32_t n = stress_mwc32() % 17;
 
 	for (i = 0; i < n; i++) {
-		uint32_t rnd = mwc32();
+		uint32_t rnd = stress_mwc32();
 
 		copy[rnd % len] ^= (1 << ((rnd >> 16) & 7));
 	}
@@ -370,10 +386,10 @@ static inline void apparmor_corrupt_flip_seq(
 {
 	static size_t p = 0;
 
-	if (p > (len * sizeof(char)))
+	if (p > len)
 		p = 0;
 
-	copy[p / sizeof(char)] ^= (1 << (p & 7));
+	copy[p] ^= (1 << (p & 7));
 	p++;
 }
 
@@ -386,10 +402,10 @@ static inline void apparmor_corrupt_clr_seq(
 {
 	static size_t p = 0;
 
-	if (p > (len * sizeof(char)))
+	if (p > len)
 		p = 0;
 
-	copy[p / sizeof(char)] &= ~(1 << (p & 7));
+	copy[p] &= ~(1 << (p & 7));
 	p++;
 }
 
@@ -402,22 +418,21 @@ static inline void apparmor_corrupt_set_seq(
 {
 	static size_t p = 0;
 
-	if (p > (len * sizeof(char)))
+	if (p > len)
 		p = 0;
 
-	copy[p / sizeof(char)] |= (1 << (p & 7));
+	copy[p] |= (1 << (p & 7));
 	p++;
 }
 
-
 /*
  *  apparmor_corrupt_flip_byte_random()
- *	ramndomly flip entire bytes
+ *	randomly flip entire bytes
  */
 static inline void apparmor_corrupt_flip_byte_random(
 	char *copy, const size_t len)
 {
-	copy[mwc32() % len] ^= 0xff;
+	copy[stress_mwc32() % len] ^= 0xff;
 }
 
 /*
@@ -428,10 +443,10 @@ static inline void apparmor_corrupt_clr_bits_random(
 	char *copy, const size_t len)
 {
 	uint32_t i;
-	const uint32_t n = mwc32() % 17;
+	const uint32_t n = stress_mwc32() % 17;
 
 	for (i = 0; i < n; i++) {
-		uint32_t rnd = mwc32();
+		uint32_t rnd = stress_mwc32();
 
 		copy[rnd % len] &= ~(1 << ((rnd >> 16) & 7));
 	}
@@ -445,10 +460,10 @@ static inline void apparmor_corrupt_set_bits_random(
 	char *copy, const size_t len)
 {
 	uint32_t i;
-	const uint32_t n = mwc32() % 17;
+	const uint32_t n = stress_mwc32() % 17;
 
 	for (i = 0; i < n; i++) {
-		uint32_t rnd = mwc32();
+		uint32_t rnd = stress_mwc32();
 
 		copy[rnd % len] |= (1 << ((rnd >> 16) & 7));
 	}
@@ -461,7 +476,7 @@ static inline void apparmor_corrupt_set_bits_random(
 static inline void apparmor_corrupt_clr_byte_random(
 	char *copy, const size_t len)
 {
-	copy[mwc32() % len] = 0;
+	copy[stress_mwc32() % len] = 0;
 }
 
 /*
@@ -471,7 +486,7 @@ static inline void apparmor_corrupt_clr_byte_random(
 static inline void apparmor_corrupt_set_byte_random(
 	char *copy, const size_t len)
 {
-	copy[mwc32() % len] = 0xff;
+	copy[stress_mwc32() % len] = (char)0xff;
 }
 
 /*
@@ -482,188 +497,235 @@ static inline void apparmor_corrupt_flip_bits_random_burst(
 	char *copy, const size_t len)
 {
 	uint32_t i;
-	size_t p = (size_t)mwc32() % (len * sizeof(char));
+	size_t p = (size_t)stress_mwc32() % (len * sizeof(*copy));
 
 	for (i = 0; i < 32; i++) {
-		if (p > (len * sizeof(char)))
+		if (p > len)
 			p = 0;
 
-		copy[p / sizeof(char)] ^= (1 << (p & 7));
+		copy[p] ^= (1 << (p & 7));
 		p++;
 	}
 }
 
+/*
+ *  apparmor_corrupt_flip_one_bit_random()
+ *	randomly flip 1 bit
+ */
+static inline void apparmor_corrupt_flip_one_bit_random(
+	char *copy, const size_t len)
+{
+	uint32_t rnd = stress_mwc32();
+
+	copy[rnd % len] ^= (1 << ((stress_mwc8() & 7)));
+}
 
 /*
  *  apparmor_stress_corruption()
  *	corrupt data and see if we can oops the loader
  *	parser.
  */
-static int apparmor_stress_corruption(
-	const char *name,
-	const uint64_t max_ops,
-	uint64_t *counter)
+static int apparmor_stress_corruption(const stress_args_t *args)
 {
-	char copy[g_apparmor_data_len];
-
-	int rc = EXIT_SUCCESS;
+	int rc = EXIT_SUCCESS, i = (int)args->instance, ret = -1;
+	int j = 0;
 	aa_kernel_interface *kern_if;
 
-
 	/*
 	 *  Lets feed AppArmor with some bit corrupted data...
 	 */
 	do {
-		int ret;
-
-		(void)memcpy(copy, g_apparmor_data, g_apparmor_data_len);
+		if (ret < 0 || j > 1024) {
+			(void)memcpy(data_copy, g_apparmor_data, g_apparmor_data_len);
+			j = 0;
+		}
 		/*
 		 *  Apply various corruption methods
 		 */
-		switch ((*counter) % 10) {
+		switch (i) {
 		case 0:
-			apparmor_corrupt_flip_seq(copy, g_apparmor_data_len);
+			apparmor_corrupt_flip_seq(data_copy, g_apparmor_data_len);
 			break;
 		case 1:
-			apparmor_corrupt_clr_seq(copy, g_apparmor_data_len);
+			apparmor_corrupt_clr_seq(data_copy, g_apparmor_data_len);
 			break;
 		case 2:
-			apparmor_corrupt_set_seq(copy, g_apparmor_data_len);
+			apparmor_corrupt_set_seq(data_copy, g_apparmor_data_len);
 			break;
 		case 3:
-			apparmor_corrupt_flip_bits_random(copy,
+			apparmor_corrupt_flip_bits_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 4:
-			apparmor_corrupt_flip_byte_random(copy,
+			apparmor_corrupt_flip_byte_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 5:
-			apparmor_corrupt_clr_bits_random(copy,
+			apparmor_corrupt_clr_bits_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 6:
-			apparmor_corrupt_set_bits_random(copy,
+			apparmor_corrupt_set_bits_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 7:
-			apparmor_corrupt_clr_byte_random(copy,
+			apparmor_corrupt_clr_byte_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 8:
-			apparmor_corrupt_set_byte_random(copy,
+			apparmor_corrupt_set_byte_random(data_copy,
 				g_apparmor_data_len);
 			break;
 		case 9:
-			apparmor_corrupt_flip_bits_random_burst(copy,
+			apparmor_corrupt_flip_bits_random_burst(data_copy,
 				g_apparmor_data_len);
 			break;
+		case 10:
+			apparmor_corrupt_flip_one_bit_random(data_copy,
+				g_apparmor_data_len);
+			i = 0;
+			break;
 		default:
 			/* Should not happen */
+			i = 0;
 			break;
 		}
 
 		ret = aa_kernel_interface_new(&kern_if, NULL, NULL);
 		if (ret < 0) {
-			pr_fail("%s: aa_kernel_interface_new() failed, "
-				"errno=%d (%s)\n", name, errno,
-				strerror(errno));
+			pr_fail("%s: aa_kernel_interface_new() failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return EXIT_FAILURE;
 		}
 		/*
 		 *  Expect EPROTO failures
 		 */
 		ret = aa_kernel_interface_replace_policy(kern_if,
-			copy, g_apparmor_data_len);
+			data_copy, g_apparmor_data_len);
 		if (ret < 0) {
+			j--;
 			if ((errno != EPROTO) &&
-			    (errno != EPROTONOSUPPORT)) {
-				pr_fail("%s: aa_kernel_interface_replace_policy() failed, "
-					"errno=%d (%s)\n", name, errno,
+			    (errno != EPROTONOSUPPORT) &&
+			     errno != ENOENT) {
+				pr_inf("%s: aa_kernel_interface_replace_policy() failed, "
+					"errno=%d (%s)\n", args->name, errno,
 					strerror(errno));
-				rc = EXIT_FAILURE;
 			}
+			(void)memcpy(data_copy, data_prev, g_apparmor_data_len);
+		} else {
+			(void)memcpy(data_prev, data_copy, g_apparmor_data_len);
 		}
 		aa_kernel_interface_unref(kern_if);
-		(*counter)++;
-	} while (g_keep_stressing_flag && apparmor_run &&
-		 (!max_ops || *counter < max_ops));
+		i++;
+	} while (stress_apparmor_keep_stressing_inc(args, true));
 
 	return rc;
 }
 
-
-static const apparmor_func apparmor_funcs[] = {
+static const stress_apparmor_func apparmor_funcs[] = {
 	apparmor_stress_profiles,
 	apparmor_stress_features,
 	apparmor_stress_kernel_interface,
 	apparmor_stress_corruption,
 };
 
-
 /*
  *  stress_apparmor()
  *	stress AppArmor
  */
-int stress_apparmor(const args_t *args)
+static int stress_apparmor(const stress_args_t *args)
 {
 	const size_t n = SIZEOF_ARRAY(apparmor_funcs);
-	const size_t counters_sz = n * sizeof(uint64_t);
 	pid_t pids[n];
 	size_t i;
-	uint64_t *counters, tmp_counter = 0, ops;
+	int rc = EXIT_NO_RESOURCE;
 
-	counters = mmap(NULL, counters_sz, PROT_READ | PROT_WRITE,
-		MAP_SHARED | MAP_ANONYMOUS, -1, 0);
-	if (counters == MAP_FAILED) {
-		pr_err("%s: mmap failed: errno=%d (%s)\n",
-			args->name, errno, strerror(errno));
-		exit(EXIT_FAILURE);
+	if (stress_sighandler(args->name, SIGUSR1, stress_apparmor_usr1_handler, NULL) < 0)
+		return EXIT_FAILURE;
+
+	data_copy = malloc(g_apparmor_data_len);
+	if (!data_copy) {
+		pr_inf("%s: failed to allocate apparmor data copy buffer, skipping stressor\n", args->name);
+		return rc;
+	}
+	data_prev = malloc(g_apparmor_data_len);
+	if (!data_prev) {
+		pr_inf("%s: failed to allocate apparmor data prev buffer, skipping stressor\n", args->name);
+		goto err_free_data_copy;
 	}
-	(void)memset(counters, 0, counters_sz);
 
-	if ((args->max_ops > 0) && (args->max_ops < n))
-		ops = n;
-	else
-		ops = args->max_ops / n;
+	counter_lock = stress_lock_create();
+	if (!counter_lock) {
+		pr_inf("%s: failed to create counter lock. skipping stressor\n", args->name);
+		goto err_free_data_prev;
+	}
 
 	for (i = 0; i < n; i++) {
-		pids[i] = apparmor_spawn(args->name, ops,
-			&counters[i], apparmor_funcs[i]);
+		pids[i] = apparmor_spawn(args, apparmor_funcs[i]);
 	}
-	do {
-		(void)select(0, NULL, NULL, NULL, NULL);
-		for (i = 0; i < n; i++)
-			tmp_counter += counters[i];
-	} while (keep_stressing());
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
+	while (stress_apparmor_keep_stressing_inc(args, false)) {
+#if defined(HAVE_SELECT)
+		(void)select(0, NULL, NULL, NULL, NULL);
+#else
+		(void)pause();
+#endif
+	}
+
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
+	/* Wakeup, time to die */
+	for (i = 0; i < n; i++) {
+		if (pids[i] >= 0)
+			(void)kill(pids[i], SIGALRM);
+	}
+	/* Now apply death grip */
 	for (i = 0; i < n; i++) {
 		int status;
 
 		if (pids[i] >= 0) {
-			(void)kill(pids[i], SIGALRM);
-			(void)waitpid(pids[i], &status, 0);
-			*(args->counter) += counters[i];
+			(void)kill(pids[i], SIGKILL);
+			(void)shim_waitpid(pids[i], &status, 0);
 		}
 	}
-	(void)munmap(counters, counters_sz);
 
 	free(apparmor_path);
 	apparmor_path = NULL;
+	(void)stress_lock_destroy(counter_lock);
 
-	return EXIT_SUCCESS;
+	rc = EXIT_SUCCESS;
+
+err_free_data_prev:
+	free(data_prev);
+err_free_data_copy:
+	free(data_copy);
+
+	return rc;
 }
+
+stressor_info_t stress_apparmor_info = {
+	.stressor = stress_apparmor,
+	.supported = stress_apparmor_supported,
+	.class = CLASS_OS | CLASS_SECURITY,
+	.help = help
+};
+
 #else
-int stress_apparmor_supported(void)
+
+static int stress_apparmor_supported(const char *name)
 {
-	pr_inf("apparmor stressor will be skipped, "
-		"AppArmor is not available\n");
+	pr_inf_skip("%s: stressor will be skipped, "
+		"AppArmor is not available\n", name);
 	return -1;
 }
 
-int stress_apparmor(const args_t *args)
-{
-	return stress_not_implemented(args);
-
-}
+stressor_info_t stress_apparmor_info = {
+	.stressor = stress_not_implemented,
+	.supported = stress_apparmor_supported,
+	.class = CLASS_OS | CLASS_SECURITY,
+	.verify = VERIFY_ALWAYS,
+	.help = help
+};
 #endif
