@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,12 +16,6 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
@@ -41,7 +36,7 @@ static int do_fchown(
 	const uid_t uid,
 	const gid_t gid)
 {
-	int ret, tmp;
+	int tmp;
 
 	if (fchown(fd, uid, gid) < 0)
 		return -errno;
@@ -70,15 +65,13 @@ static int do_fchown(
 	/*
 	 *  Exercise fchown with invalid fd
 	 */
-	ret = fchown(bad_fd, uid, gid);
-	(void)ret;
+	VOID_RET(int, fchown(bad_fd, uid, gid));
 
 	return 0;
 
 restore:
 	tmp = errno;
-	ret = fchown(fd, uid, gid);
-	(void)ret;
+	VOID_RET(int, fchown(fd, uid, gid));
 
 	return -tmp;
 }
@@ -94,7 +87,7 @@ static int do_chown(
 	const uid_t uid,
 	const gid_t gid)
 {
-	int ret, tmp;
+	int tmp;
 
 	if (chown_func(filename, uid, gid) < 0)
 		return -errno;
@@ -124,8 +117,7 @@ static int do_chown(
 
 restore:
 	tmp = errno;
-	ret = chown_func(filename, uid, gid);
-	(void)ret;
+	VOID_RET(int, chown_func(filename, uid, gid));
 
 	return -tmp;
 }
@@ -190,8 +182,9 @@ static int stress_chown(const stress_args_t *args)
 			if (++retries >= 100) {
 				pr_err("%s: chown: file %s took %d "
 					"retries to open and gave up "
-					"(instance %" PRIu32 ")\n",
-					args->name, filename, retries, args->instance);
+					"(instance %" PRIu32 ")%s\n",
+					args->name, filename, retries, args->instance,
+					stress_fs_type(filename));
 				goto tidy;
 			}
 		}
@@ -204,8 +197,9 @@ static int stress_chown(const stress_args_t *args)
 
 		ret = do_fchown(fd, bad_fd, cap_chown, uid, gid);
 		if ((ret < 0) && (ret != -EPERM))
-			pr_fail("%s: fchown failed, errno=%d (%s)\n",
-				args->name, errno, strerror(errno));
+			pr_fail("%s: fchown failed, errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
 
 		ret = do_chown(chown, filename, cap_chown, uid, gid);
 		if (ret < 0) {
@@ -218,8 +212,9 @@ static int stress_chown(const stress_args_t *args)
 				goto tidy;
 			}
 			if (ret != -EPERM)
-				pr_fail("%s: chown %s failed, errno=%d (%s)\n",
-					args->name, filename, errno, strerror(errno));
+				pr_fail("%s: chown %s failed, errno=%d (%s)%s\n",
+					args->name, filename, errno, strerror(errno),
+					stress_fs_type(filename));
 		}
 		ret = do_chown(lchown, filename, cap_chown, uid, gid);
 		if (ret < 0) {
@@ -232,9 +227,11 @@ static int stress_chown(const stress_args_t *args)
 				goto tidy;
 			}
 			if (ret != -EPERM)
-				pr_fail("%s: chown %s failed, errno=%d (%s)\n",
-					args->name, filename, errno, strerror(errno));
+				pr_fail("%s: lchown %s failed, errno=%d (%s)%s\n",
+					args->name, filename, errno, strerror(errno),
+					stress_fs_type(filename));
 		}
+		(void)shim_fsync(fd);
 		inc_counter(args);
 	} while (keep_stressing(args));
 
@@ -244,8 +241,8 @@ tidy:
 
 	if (fd >= 0)
 		(void)close(fd);
-	(void)unlink(filename);
-	(void)rmdir(pathname);
+	(void)shim_unlink(filename);
+	(void)shim_rmdir(pathname);
 
 	return rc;
 }
@@ -253,5 +250,6 @@ tidy:
 stressor_info_t stress_chown_info = {
 	.stressor = stress_chown,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
