@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2016-2017 Canonical, Ltd.
+ * Copyright (C) 2016-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,40 +16,48 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
+#include "core-net.h"
 
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#if defined(AF_INET6)
-#include <netinet/in.h>
-#endif
-#if defined(AF_UNIX)
+#if defined(HAVE_SYS_UN_H)
 #include <sys/un.h>
 #endif
 
-#if defined(__linux__) && defined(__NR_sendmmsg) && NEED_GLIBC(2,14,0)
-#define HAVE_SENDMMSG
+#if defined(HAVE_LINUX_SOCKIOS_H)
+#include <linux/sockios.h>
 #endif
 
-#define DCCP_OPT_SEND		0x01
-#define DCCP_OPT_SENDMSG	0x02
-#define DCCP_OPT_SENDMMSG	0x03
+#include <netinet/in.h>
+
+#define MIN_DCCP_PORT		(1024)
+#define MAX_DCCP_PORT		(65535)
+#define DEFAULT_DCCP_PORT	(10000)
+
+#define DCCP_OPT_SEND		(0x01)
+#define DCCP_OPT_SENDMSG	(0x02)
+#define DCCP_OPT_SENDMMSG	(0x03)
+
+#define DCCP_BUF		(1024)	/* DCCP I/O buffer size */
 
 #define MSGVEC_SIZE		(4)
 
 typedef struct {
 	const char *optname;
-	int	   opt;
-} dccp_opts_t;
+	const int   opt;
+} stress_dccp_opts_t;
 
-static const dccp_opts_t dccp_options[] = {
+static const stress_help_t help[] = {
+	{ NULL,	"dccp N",		"start N workers exercising network DCCP I/O" },
+	{ NULL,	"dccp-domain D",	"specify DCCP domain, default is ipv4" },
+	{ NULL,	"dccp-if I",		"use network interface I, e.g. lo, eth0, etc." },
+	{ NULL,	"dccp-ops N",		"stop after N DCCP  bogo operations" },
+	{ NULL,	"dccp-opts option",	"DCCP data send options [send|sendmsg|sendmmsg]" },
+	{ NULL,	"dccp-port P",		"use DCCP ports P to P + number of workers - 1" },
+	{ NULL,	NULL,			NULL }
+};
+
+static const stress_dccp_opts_t dccp_options[] = {
 	{ "send",	DCCP_OPT_SEND },
 	{ "sendmsg",	DCCP_OPT_SENDMSG },
 #if defined(HAVE_SENDMMSG)
@@ -61,14 +70,15 @@ static const dccp_opts_t dccp_options[] = {
  *  stress_set_dccp_opts()
  *	parse --dccp-opts
  */
-int stress_set_dccp_opts(const char *opt)
+static int stress_set_dccp_opts(const char *opt)
 {
-	int i;
+	size_t i;
 
 	for (i = 0; dccp_options[i].optname; i++) {
 		if (!strcmp(opt, dccp_options[i].optname)) {
 			int dccp_opt = dccp_options[i].opt;
-			set_setting("dccp-opts", TYPE_ID_INT, &dccp_opt);
+
+			stress_set_setting("dccp-opts", TYPE_ID_INT, &dccp_opt);
 			return 0;
 		}
 	}
@@ -85,41 +95,56 @@ int stress_set_dccp_opts(const char *opt)
  *  stress_set_dccp_port()
  *	set port to use
  */
-void stress_set_dccp_port(const char *opt)
+static int stress_set_dccp_port(const char *opt)
 {
 	int dccp_port;
 
 	stress_set_net_port("dccp-port", opt,
 		MIN_DCCP_PORT, MAX_DCCP_PORT - STRESS_PROCS_MAX,
 		&dccp_port);
-	set_setting("dccp-port", TYPE_ID_INT, &dccp_port);
+	return stress_set_setting("dccp-port", TYPE_ID_INT, &dccp_port);
 }
 
 /*
  *  stress_set_dccp_domain()
  *	set the socket domain option
  */
-int stress_set_dccp_domain(const char *name)
+static int stress_set_dccp_domain(const char *name)
 {
 	int ret, dccp_domain;
 
 	ret = stress_set_net_domain(DOMAIN_INET | DOMAIN_INET6,
 				"dccp-domain", name, &dccp_domain);
-	set_setting("dccp-domain", TYPE_ID_INT, &dccp_domain);
+	stress_set_setting("dccp-domain", TYPE_ID_INT, &dccp_domain);
 	return ret;
 }
 
-#if defined(SOCK_DCCP) && defined(IPPROTO_DCCP)
+static int stress_set_dccp_if(const char *name)
+{
+        return stress_set_setting("dccp-if", TYPE_ID_STR, name);
+}
+
+static const stress_opt_set_func_t opt_set_funcs[] = {
+	{ OPT_dccp_domain,	stress_set_dccp_domain },
+	{ OPT_dccp_if,		stress_set_dccp_if },
+	{ OPT_dccp_opts,	stress_set_dccp_opts },
+	{ OPT_dccp_port,	stress_set_dccp_port },
+	{ 0,			NULL },
+};
+
+#if defined(SOCK_DCCP) &&	\
+    defined(IPPROTO_DCCP)
 
 /*
  *  stress_dccp_client()
  *	client reader
  */
 static void stress_dccp_client(
-	const args_t *args,
+	const stress_args_t *args,
 	const pid_t ppid,
 	const int dccp_port,
-	const int dccp_domain)
+	const int dccp_domain,
+	const char *dccp_if)
 {
 	struct sockaddr *addr;
 
@@ -132,19 +157,28 @@ static void stress_dccp_client(
 		int retries = 0;
 		socklen_t addr_len = 0;
 retry:
-		if (!g_keep_stressing_flag) {
+		if (!keep_stressing_flag()) {
 			(void)kill(getppid(), SIGALRM);
-			exit(EXIT_FAILURE);
+			_exit(EXIT_FAILURE);
 		}
 		if ((fd = socket(dccp_domain, SOCK_DCCP, IPPROTO_DCCP)) < 0) {
-			pr_fail_dbg("socket");
+			if ((errno == ESOCKTNOSUPPORT) ||
+			    (errno == EPROTONOSUPPORT)) {
+				/*
+				 *  Protocol not supported - then return
+				 *  EXIT_NOT_IMPLEMENTED and skip the test
+				 */
+				_exit(EXIT_NOT_IMPLEMENTED);
+			}
+			pr_fail("%s: socket failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			/* failed, kick parent to finish */
 			(void)kill(getppid(), SIGALRM);
-			exit(EXIT_FAILURE);
+			_exit(EXIT_FAILURE);
 		}
 
-		stress_set_sockaddr(args->name, args->instance, ppid,
-			dccp_domain, dccp_port,
+		stress_set_sockaddr_if(args->name, args->instance, ppid,
+			dccp_domain, dccp_port, dccp_if,
 			&addr, &addr_len, NET_ADDR_ANY);
 		if (connect(fd, addr, addr_len) < 0) {
 			int err = errno;
@@ -155,9 +189,10 @@ retry:
 			if (retries > 100) {
 				/* Give up.. */
 				errno = err;
-				pr_fail_dbg("connect");
+				pr_fail("%s: connect failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				(void)kill(getppid(), SIGALRM);
-				exit(EXIT_FAILURE);
+				_exit(EXIT_FAILURE);
 			}
 			goto retry;
 		}
@@ -168,18 +203,21 @@ retry:
 				break;
 			if (n < 0) {
 				if (errno != EINTR)
-					pr_fail_dbg("recv");
+					pr_dbg("%s: recv failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
 				break;
 			}
-		} while (keep_stressing());
+		} while (keep_stressing(args));
 		(void)shutdown(fd, SHUT_RDWR);
 		(void)close(fd);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
-#if defined(AF_UNIX)
+#if defined(AF_UNIX) &&		\
+    defined(HAVE_SOCKADDR_UN)
 	if (dccp_domain == AF_UNIX) {
 		struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
-		(void)unlink(addr_un->sun_path);
+
+		(void)shim_unlink(addr_un->sun_path);
 	}
 #endif
 	/* Inform parent we're all done */
@@ -191,11 +229,12 @@ retry:
  *	server writer
  */
 static int stress_dccp_server(
-	const args_t *args,
+	const stress_args_t *args,
 	const pid_t pid,
 	const pid_t ppid,
 	const int dccp_port,
 	const int dccp_domain,
+	const char *dccp_if,
 	const int dccp_opts)
 {
 	char buf[DCCP_BUF];
@@ -205,6 +244,7 @@ static int stress_dccp_server(
 	struct sockaddr *addr = NULL;
 	uint64_t msgs = 0;
 	int rc = EXIT_SUCCESS;
+	double t1 = 0.0, t2 = 0.0, dt;
 
 	(void)setpgid(pid, g_pgrp);
 
@@ -213,33 +253,54 @@ static int stress_dccp_server(
 		goto die;
 	}
 	if ((fd = socket(dccp_domain, SOCK_DCCP, IPPROTO_DCCP)) < 0) {
+		if ((errno == ESOCKTNOSUPPORT) ||
+		    (errno == EPROTONOSUPPORT)) {
+			/*
+			 *  Protocol not supported - then return
+			 *  EXIT_NOT_IMPLEMENTED and skip the test
+			 */
+			if (args->instance == 0)
+				pr_inf_skip("%s: DCCP protocol not supported, "
+					"skipping stressor\n", args->name);
+			return EXIT_NOT_IMPLEMENTED;
+		}
 		rc = exit_status(errno);
-		pr_fail_dbg("socket");
+		pr_fail("%s: socket failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto die;
 	}
 	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
 		&so_reuseaddr, sizeof(so_reuseaddr)) < 0) {
-		pr_fail_dbg("setsockopt");
+		pr_fail("%s: setsockopt failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		rc = EXIT_FAILURE;
 		goto die_close;
 	}
 
-	stress_set_sockaddr(args->name, args->instance, ppid,
-		dccp_domain, dccp_port,
+	stress_set_sockaddr_if(args->name, args->instance, ppid,
+		dccp_domain, dccp_port, dccp_if,
 		&addr, &addr_len, NET_ADDR_ANY);
 	if (bind(fd, addr, addr_len) < 0) {
 		rc = exit_status(errno);
-		pr_fail_dbg("bind");
+		pr_fail("%s: bind failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto die_close;
 	}
 	if (listen(fd, 10) < 0) {
-		pr_fail_dbg("listen");
+		pr_fail("%s: listen failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		rc = EXIT_FAILURE;
 		goto die_close;
 	}
 
+	t1 = stress_time_now();
 	do {
-		int sfd = accept(fd, (struct sockaddr *)NULL, NULL);
+		int sfd;
+
+		if (!keep_stressing(args))
+			break;
+
+		sfd = accept(fd, (struct sockaddr *)NULL, NULL);
 		if (sfd >= 0) {
 			size_t i, j;
 			struct sockaddr saddr;
@@ -249,22 +310,23 @@ static int stress_dccp_server(
 			struct iovec vec[sizeof(buf)/16];
 #if defined(HAVE_SENDMMSG)
 			struct mmsghdr msgvec[MSGVEC_SIZE];
-			unsigned int msg_len = 0;
 #endif
 			len = sizeof(saddr);
 			if (getsockname(fd, &saddr, &len) < 0) {
-				pr_fail_dbg("getsockname");
+				pr_dbg("%s: getsockname failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				(void)close(sfd);
 				break;
 			}
 			len = sizeof(sndbuf);
 			if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sndbuf, &len) < 0) {
-				pr_fail_dbg("getsockopt");
+				pr_dbg("%s: getsockopt SO_SNDBUF failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				(void)close(sfd);
 				break;
 			}
 
-			(void)memset(buf, 'A' + (*args->counter % 26), sizeof(buf));
+			(void)memset(buf, 'A' + (get_counter(args) % 26), sizeof(buf));
 			switch (dccp_opts) {
 			case DCCP_OPT_SEND:
 				for (i = 16; i < sizeof(buf); i += 16) {
@@ -275,7 +337,8 @@ again:
 						if (errno == EAGAIN)
 							goto again;
 						if (errno != EINTR)
-							pr_fail_dbg("send");
+							pr_dbg("%s: send failed, errno=%d (%s)\n",
+								args->name, errno, strerror(errno));
 						break;
 					} else
 						msgs++;
@@ -291,7 +354,8 @@ again:
 				msg.msg_iovlen = j;
 				if (sendmsg(sfd, &msg, 0) < 0) {
 					if (errno != EINTR)
-						pr_fail_dbg("sendmsg");
+						pr_dbg("%s: sendmsg failed, errno=%d (%s)\n",
+							args->name, errno, strerror(errno));
 				} else
 					msgs += j;
 				break;
@@ -301,7 +365,6 @@ again:
 				for (j = 0, i = 16; i < sizeof(buf); i += 16, j++) {
 					vec[j].iov_base = buf;
 					vec[j].iov_len = i;
-					msg_len += i;
 				}
 				for (i = 0; i < MSGVEC_SIZE; i++) {
 					msgvec[i].msg_hdr.msg_iov = vec;
@@ -309,7 +372,8 @@ again:
 				}
 				if (sendmmsg(sfd, msgvec, MSGVEC_SIZE, 0) < 0) {
 					if (errno != EINTR)
-						pr_fail_dbg("sendmmsg");
+						pr_dbg("%s: sendmmsg failed, errno=%d (%s)\n",
+							args->name, errno, strerror(errno));
 				} else
 					msgs += (MSGVEC_SIZE * j);
 				break;
@@ -321,28 +385,44 @@ again:
 				goto die_close;
 			}
 			if (getpeername(sfd, &saddr, &len) < 0) {
-				pr_fail_dbg("getpeername");
+				pr_dbg("%s: getpeername failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 			}
+#if defined(SIOCOUTQ)
+			{
+				int pending;
+
+				(void)ioctl(sfd, SIOCOUTQ, &pending);
+			}
+#endif
 			(void)close(sfd);
 		}
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 die_close:
 	(void)close(fd);
 die:
-#if defined(AF_UNIX)
+	t2 = stress_time_now();
+#if defined(AF_UNIX) &&		\
+    defined(HAVE_SOCKADDR_UN)
 	if (addr && (dccp_domain == AF_UNIX)) {
 		struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
-		(void)unlink(addr_un->sun_path);
+
+		(void)shim_unlink(addr_un->sun_path);
 	}
 #endif
+
 	if (pid) {
 		(void)kill(pid, SIGKILL);
-		(void)waitpid(pid, &status, 0);
+		(void)shim_waitpid(pid, &status, 0);
 	}
 	pr_dbg("%s: %" PRIu64 " messages sent\n", args->name, msgs);
 
+	dt = t2 - t1;
+	if (dt > 0.0)
+		stress_misc_stats_set(args->misc_stats, 0, "messages per sec", (double)msgs / dt);
+
 	return rc;
 }
 
@@ -350,40 +430,71 @@ die:
  *  stress_dccp
  *	stress by heavy dccp  I/O
  */
-int stress_dccp(const args_t *args)
+static int stress_dccp(const stress_args_t *args)
 {
 	pid_t pid, ppid = getppid();
 	int dccp_port = DEFAULT_DCCP_PORT;
 	int dccp_domain = AF_INET;
 	int dccp_opts = DCCP_OPT_SEND;
+	int rc = EXIT_SUCCESS;
+	char *dccp_if = NULL;
 
-	(void)get_setting("dccp-port", &dccp_port);
-	(void)get_setting("dccp-domain", &dccp_domain);
-	(void)get_setting("dccp-opts", &dccp_opts);
+	(void)stress_get_setting("dcpp-if", &dccp_if);
+	(void)stress_get_setting("dccp-port", &dccp_port);
+	(void)stress_get_setting("dccp-domain", &dccp_domain);
+	(void)stress_get_setting("dccp-opts", &dccp_opts);
+
+	if (dccp_if) {
+		int ret;
+		struct sockaddr if_addr;
+
+		ret = stress_net_interface_exists(dccp_if, dccp_domain, &if_addr);
+		if (ret < 0) {
+			pr_inf("%s: interface '%s' is not enabled for domain '%s', defaulting to using loopback\n",
+				args->name, dccp_if, stress_net_domain(dccp_domain));
+			dccp_if = NULL;
+		}
+	}
 
 	pr_dbg("%s: process [%d] using socket port %d\n",
-		args->name, (int)args->pid, dccp_port + args->instance);
+		args->name, (int)args->pid, dccp_port + (int)args->instance);
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
 	pid = fork();
 	if (pid < 0) {
-		if (g_keep_stressing_flag && (errno == EAGAIN))
+		if (stress_redo_fork(errno))
 			goto again;
-		pr_fail_dbg("fork");
-		return EXIT_FAILURE;
+		if (!keep_stressing(args))
+			goto finish;
+		pr_dbg("%s: fork failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
+		return EXIT_NO_RESOURCE;
 	} else if (pid == 0) {
-		stress_dccp_client(args, ppid, dccp_port, dccp_domain);
-		exit(EXIT_SUCCESS);
+		(void)sched_settings_apply(true);
+		stress_dccp_client(args, ppid, dccp_port, dccp_domain, dccp_if);
+		_exit(EXIT_SUCCESS);
 	} else {
-		return stress_dccp_server(args, pid, ppid, dccp_port,
-			dccp_domain, dccp_opts);
+		rc = stress_dccp_server(args, pid, ppid, dccp_port,
+			dccp_domain, dccp_if, dccp_opts);
 	}
+finish:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+	return rc;
 }
 
+stressor_info_t stress_dccp_info = {
+	.stressor = stress_dccp,
+	.class = CLASS_NETWORK | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_ALWAYS,
+	.help = help
+};
 #else
-
-int stress_dccp(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_dccp_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_NETWORK | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
 #endif
