@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,17 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
+#define MIN_DIR_DIRS		(64)
+#define MAX_DIR_DIRS		(65536)
+#define DEFAULT_DIR_DIRS	(8192)
+
+#if defined(HAVE_LIBGEN_H)
+#include <libgen.h>
+#endif
+
 static const stress_help_t help[] = {
 	{ NULL,	"dir N",	"start N directory thrashing stressors" },
 	{ NULL,	"dir-ops N",	"stop after N directory bogo operations" },
@@ -99,17 +102,13 @@ static inline void stress_dir_flock(const int dir_fd)
  */
 static inline void stress_dir_truncate(const char *path, const int dir_fd)
 {
-	int ret;
-
 	if (dir_fd >= 0) {
 		/* Invalid ftruncate */
-		ret = ftruncate(dir_fd, 0);
-		(void)ret;
+		VOID_RET(int, ftruncate(dir_fd, 0));
 	}
 
 	/* Invalid truncate */
-	ret = truncate(path, 0);
-	(void)ret;
+	VOID_RET(int, truncate(path, 0));
 }
 
 /*
@@ -178,7 +177,7 @@ static void stress_dir_tidy(
 
 		(void)stress_temp_filename_args(args,
 			path, sizeof(path), gray_code);
-		(void)rmdir(path);
+		(void)shim_rmdir(path);
 	}
 }
 
@@ -228,7 +227,7 @@ static void stress_invalid_mkdir(const char *path)
 	(void)stress_strnrnd(filename + len, sizeof(filename) - len);
 	ret = mkdir(filename,  S_IRUSR | S_IWUSR);
 	if (ret == 0)
-		(void)rmdir(filename);
+		(void)shim_rmdir(filename);
 }
 
 /*
@@ -238,10 +237,7 @@ static void stress_invalid_mkdir(const char *path)
 static void stress_invalid_mkdirat(const int bad_fd)
 {
 #if defined(HAVE_MKDIRAT)
-	int ret;
-
-	ret = mkdirat(bad_fd, "bad", S_IRUSR | S_IWUSR);
-	(void)ret;
+	VOID_RET(int, mkdirat(bad_fd, "bad", S_IRUSR | S_IWUSR));
 #else
 	(void)bad_fd;
 #endif
@@ -254,23 +250,19 @@ static void stress_invalid_mkdirat(const int bad_fd)
  */
 static void stress_invalid_rmdir(const char *path)
 {
-	int ret;
 	char filename[PATH_MAX + 16];
 
 	(void)shim_strlcpy(filename, path, sizeof(filename));
 	/* remove . - exercise EINVAL error */
 	(void)shim_strlcat(filename, "/.", sizeof(filename));
-	ret = rmdir(filename);
-	(void)ret;
+	VOID_RET(int, shim_rmdir(filename));
 
 	/* remove /.. - exercise ENOTEMPTY error */
 	(void)shim_strlcat(filename, ".", sizeof(filename));
-	ret = rmdir(filename);
-	(void)ret;
+	VOID_RET(int, shim_rmdir(filename));
 
 	/* remove / - exercise EBUSY error */
-	ret = rmdir("/");
-	(void)ret;
+	VOID_RET(int, shim_rmdir("/"));
 }
 
 /*
@@ -310,6 +302,7 @@ static int stress_dir(const stress_args_t *args)
 
 			(void)stress_temp_filename_args(args,
 				path, sizeof(path), gray_code);
+			(void)shim_force_rmdir(path);
 			if (stress_mkdir(dir_fd, path, S_IRUSR | S_IWUSR) < 0) {
 				if ((errno != ENOSPC) &&
 				    (errno != ENOMEM) &&
@@ -344,7 +337,7 @@ static int stress_dir(const stress_args_t *args)
 	{
 		int rmret;
 
-		rmret = rmdir("");
+		rmret = shim_rmdir("");
 		(void)rmret;
 	}
 
@@ -368,5 +361,6 @@ stressor_info_t stress_dir_info = {
 	.stressor = stress_dir,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
