@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2012-2017 Canonical, Ltd.
+ * Copyright (C) 2012-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,25 +16,52 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
  */
 #include "stress-ng.h"
 
-#if defined(__linux__) && defined(F_NOTIFY)
-
+#if defined(HAVE_SYS_SELECT_H)
 #include <sys/select.h>
+#endif
+
+static const stress_help_t help[] = {
+	{ NULL,	"dnotify N",	 "start N workers exercising dnotify events" },
+	{ NULL,	"dnotify-ops N", "stop dnotify workers after N bogo operations" },
+	{ NULL,	NULL,		 NULL }
+};
+
+#if defined(F_NOTIFY) && \
+    defined(HAVE_SYS_SELECT_H)
 
-#define DIR_FLAGS	(S_IRWXU | S_IRWXG)
 #define FILE_FLAGS	(S_IRUSR | S_IWUSR)
 
 #define BUF_SIZE	(4096)
 
 static volatile int dnotify_fd;
 
+static int stress_dnotify_supported(const char *name)
+{
+	char buf[64];
+	static const char path[] = "/proc/sys/fs/dir-notify-enable";
+	int enabled;
+
+	if (system_read(path, buf, sizeof(buf)) < 0) {
+		pr_inf_skip("%s stressor will be skipped, cannot "
+			"open '%s', CONFIG_DNOTIFY is probably not set\n",
+			name, path);
+		return -1;
+	}
+	if (sscanf(buf, "%d", &enabled) != 1) {
+		pr_inf_skip("%s stressor will be skipped, cannot "
+			"parse '%s'\n", name, path);
+		return -1;
+	}
+	if (enabled != 1) {
+		pr_inf_skip("%s stressor will be skipped, dnotify is not enabled\n", name);
+		return -1;
+	}
+	return 0;
+}
+
 static void dnotify_handler(int sig, siginfo_t *si, void *data)
 {
 	(void)sig;
@@ -42,13 +70,13 @@ static void dnotify_handler(int sig, siginfo_t *si, void *data)
 	dnotify_fd = si->si_fd;
 }
 
-typedef int (*dnotify_helper)(const args_t *args, const char *path, const void *private);
-typedef void (*dnotify_func)(const args_t *args, const char *path);
+typedef int (*stress_dnotify_helper)(const stress_args_t *args, const char *path, const void *private);
+typedef void (*stress_dnotify_func)(const stress_args_t *args, const char *path);
 
 typedef struct {
-	const dnotify_func func;
+	const stress_dnotify_func func;
 	const char*	description;
-} dnotify_stress_t;
+} stress_dnotify_stress_t;
 
 /*
  *  dnotify_exercise()
@@ -56,25 +84,33 @@ typedef struct {
  *	required dnotify event flags 'flags'.
  */
 static void dnotify_exercise(
-	const args_t *args,	/* Stressor args */
-	const char *filename,	/* Filename in test */
-	const char *watchname,	/* File or directory to watch using dnotify */
-	const dnotify_helper func,	/* Helper func */
-	const int flags,	/* DN_* flags to watch for */
-	void *private)		/* Helper func private data */
+	const stress_args_t *args,	/* Stressor args */
+	const char *filename,		/* Filename in test */
+	const char *watchname,		/* File or directory to watch using dnotify */
+	const stress_dnotify_helper func,/* Helper func */
+	const unsigned long flags,	/* DN_* flags to watch for */
+	void *private)			/* Helper func private data */
 {
 	int fd, i = 0;
+#if defined(DN_MULTISHOT)
+	unsigned long flags_ms = flags | DN_MULTISHOT;
+#else
+	unsigned long flags_ms = flags;
+#endif
 
 	if ((fd = open(watchname, O_RDONLY)) < 0) {
-		pr_fail_err("open");
+		pr_fail("%s: open %s failed, errno=%d (%s)\n",
+			args->name, watchname, errno, strerror(errno));
 		return;
 	}
 	if (fcntl(fd, F_SETSIG, SIGRTMIN + 1) < 0) {
-		pr_fail_err("fcntl F_SETSIG");
+		pr_fail("%s: fcntl F_SETSIG failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto cleanup;
 	}
-	if (fcntl(fd, F_NOTIFY, flags) < 0) {
-		pr_fail_err("fcntl F_NOTIFY");
+	if (fcntl(fd, F_NOTIFY, flags_ms) < 0) {
+		pr_fail("%s: fcntl F_NOTIFY failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto cleanup;
 	}
 
@@ -84,13 +120,15 @@ static void dnotify_exercise(
 
 	/* Wait for up to 1 second for event */
 	while ((i < 1000) && (dnotify_fd == -1)) {
+		if (!keep_stressing(args))
+			goto cleanup;
 		i++;
 		(void)shim_usleep(1000);
 	}
 
 	if (dnotify_fd != fd) {
-		pr_fail("%s: did not get expected dnotify "
-			"file descriptor\n", args->name);
+		pr_fail("%s: did not get expected dnotify file descriptor\n",
+			args->name);
 	}
 
 cleanup:
@@ -102,9 +140,9 @@ cleanup:
  *  rm_file()
  *	remove a file
  */
-static int rm_file(const args_t *args, const char *path)
+static int rm_file(const stress_args_t *args, const char *path)
 {
-	if ((unlink(path) < 0) && errno != ENOENT) {
+	if ((shim_force_unlink(path) < 0) && errno != ENOENT) {
 		pr_err("%s: cannot remove file %s: errno=%d (%s)\n",
 			args->name, path, errno, strerror(errno));
 		return -1;
@@ -112,24 +150,11 @@ static int rm_file(const args_t *args, const char *path)
 	return 0;
 }
 
-/*
- *  mk_filename()
- *	simple helper to create a filename
- */
-static inline void mk_filename(
-	char *filename,
-	const size_t len,
-	const char *path,
-	const char *name)
-{
-	(void)snprintf(filename, len, "%s/%s", path, name);
-}
-
 /*
  *  mk_file()
  *	create file of length len bytes
  */
-static int mk_file(const args_t *args, const char *filename, const size_t len)
+static int mk_file(const stress_args_t *args, const char *filename, const size_t len)
 {
 	int fd;
 	size_t sz = len;
@@ -137,23 +162,28 @@ static int mk_file(const args_t *args, const char *filename, const size_t len)
 	char buffer[BUF_SIZE];
 
 	if ((fd = open(filename, O_CREAT | O_RDWR, FILE_FLAGS)) < 0) {
+		if ((errno == ENFILE) || (errno == ENOMEM) || (errno == ENOSPC))
+			return -1;
 		pr_err("%s: cannot create file %s: errno=%d (%s)\n",
 			args->name, filename, errno, strerror(errno));
 		return -1;
 	}
 
-	memset(buffer, 'x', BUF_SIZE);
-	while (sz > 0) {
+	(void)memset(buffer, 'x', BUF_SIZE);
+	while (keep_stressing(args) && (sz > 0)) {
 		size_t n = (sz > BUF_SIZE) ? BUF_SIZE : sz;
-		int ret;
+		ssize_t ret;
 
-		if ((ret = write(fd, buffer, n)) < 0) {
+		ret = write(fd, buffer, n);
+		if (ret < 0) {
+			if (errno == ENOSPC)
+				break;
 			pr_err("%s: error writing to file %s: errno=%d (%s)\n",
 				args->name, filename, errno, strerror(errno));
 			(void)close(fd);
 			return -1;
 		}
-		sz -= ret;
+		sz -= (size_t)ret;
 	}
 
 	if (close(fd) < 0) {
@@ -165,11 +195,11 @@ static int mk_file(const args_t *args, const char *filename, const size_t len)
 }
 
 static int dnotify_attrib_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *path,
-	const void *dummy)
+	const void *signum)
 {
-	(void)dummy;
+	(void)signum;
 	if (chmod(path, S_IRUSR | S_IWUSR) < 0) {
 		pr_err("%s: cannot chmod file %s: errno=%d (%s)\n",
 			args->name, path, errno, strerror(errno));
@@ -178,11 +208,11 @@ static int dnotify_attrib_helper(
 	return 0;
 }
 
-static void dnotify_attrib_file(const args_t *args, const char *path)
+static void dnotify_attrib_file(const stress_args_t *args, const char *path)
 {
 	char filepath[PATH_MAX];
 
-	mk_filename(filepath, PATH_MAX, path, "dnotify_file");
+	stress_mk_filename(filepath, sizeof(filepath), path, "dnotify_file");
 	if (mk_file(args, filepath, 4096) < 0)
 		return;
 
@@ -192,15 +222,15 @@ static void dnotify_attrib_file(const args_t *args, const char *path)
 }
 
 static int dnotify_access_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *path,
-	const void *dummy)
+	const void *signum)
 {
 	int fd;
 	char buffer[1];
 	int rc = 0;
 
-	(void)dummy;
+	(void)signum;
 	if ((fd = open(path, O_RDONLY)) < 0) {
 		pr_err("%s: cannot open file %s: errno=%d (%s)\n",
 			args->name, path, errno, strerror(errno));
@@ -209,7 +239,7 @@ static int dnotify_access_helper(
 
 	/* Just want to force an access */
 do_access:
-	if (g_keep_stressing_flag && (read(fd, buffer, 1) < 0)) {
+	if (keep_stressing(args) && (read(fd, buffer, 1) < 0)) {
 		if ((errno == EAGAIN) || (errno == EINTR))
 			goto do_access;
 		pr_err("%s: cannot read file %s: errno=%d (%s)\n",
@@ -220,11 +250,11 @@ do_access:
 	return rc;
 }
 
-static void dnotify_access_file(const args_t *args, const char *path)
+static void dnotify_access_file(const stress_args_t *args, const char *path)
 {
 	char filepath[PATH_MAX];
 
-	mk_filename(filepath, PATH_MAX, path, "dnotify_file");
+	stress_mk_filename(filepath, sizeof(filepath), path, "dnotify_file");
 	if (mk_file(args, filepath, 4096) < 0)
 		return;
 
@@ -234,14 +264,14 @@ static void dnotify_access_file(const args_t *args, const char *path)
 }
 
 static int dnotify_modify_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *path,
-	const void *dummy)
+	const void *signum)
 {
 	int fd, rc = 0;
 	char buffer[1] = { 0 };
 
-	(void)dummy;
+	(void)signum;
 	if (mk_file(args, path, 4096) < 0)
 		return -1;
 	if ((fd = open(path, O_RDWR)) < 0) {
@@ -251,7 +281,7 @@ static int dnotify_modify_helper(
 		goto remove;
 	}
 do_modify:
-	if (g_keep_stressing_flag && (write(fd, buffer, 1) < 0)) {
+	if (keep_stressing(args) && (write(fd, buffer, 1) < 0)) {
 		if ((errno == EAGAIN) || (errno == EINTR))
 			goto do_modify;
 		pr_err("%s: cannot write to file %s: errno=%d (%s)\n",
@@ -264,22 +294,23 @@ remove:
 	return rc;
 }
 
-static void dnotify_modify_file(const args_t *args, const char *path)
+static void dnotify_modify_file(const stress_args_t *args, const char *path)
 {
 	char filepath[PATH_MAX];
 
-	mk_filename(filepath, PATH_MAX, path, "dnotify_file");
+	stress_mk_filename(filepath, sizeof(filepath), path, "dnotify_file");
 	dnotify_exercise(args, filepath, path,
 		dnotify_modify_helper, DN_MODIFY, NULL);
 }
 
 static int dnotify_creat_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *path,
-	const void *dummy)
+	const void *signum)
 {
-	(void)dummy;
+	(void)signum;
 	int fd;
+
 	if ((fd = creat(path, FILE_FLAGS)) < 0) {
 		pr_err("%s: cannot create file %s: errno=%d (%s)\n",
 			args->name, path, errno, strerror(errno));
@@ -289,31 +320,31 @@ static int dnotify_creat_helper(
 	return 0;
 }
 
-static void dnotify_creat_file(const args_t *args, const char *path)
+static void dnotify_creat_file(const stress_args_t *args, const char *path)
 {
 	char filepath[PATH_MAX];
 
-	mk_filename(filepath, PATH_MAX, path, "dnotify_file");
+	stress_mk_filename(filepath, sizeof(filepath), path, "dnotify_file");
 	dnotify_exercise(args, filepath, path,
 		dnotify_creat_helper, DN_CREATE, NULL);
 	(void)rm_file(args, filepath);
 }
 
 static int dnotify_delete_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *path,
-	const void *dummy)
+	const void *signum)
 {
-	(void)dummy;
+	(void)signum;
 
 	return rm_file(args, path);
 }
 
-static void dnotify_delete_file(const args_t *args, const char *path)
+static void dnotify_delete_file(const stress_args_t *args, const char *path)
 {
 	char filepath[PATH_MAX];
 
-	mk_filename(filepath, PATH_MAX, path, "dnotify_file");
+	stress_mk_filename(filepath, sizeof(filepath), path, "dnotify_file");
 	if (mk_file(args, filepath, 4096) < 0)
 		return;
 	dnotify_exercise(args, filepath, path,
@@ -323,7 +354,7 @@ static void dnotify_delete_file(const args_t *args, const char *path)
 }
 
 static int dnotify_rename_helper(
-	const args_t *args,
+	const stress_args_t *args,
 	const char *oldpath,
 	const void *private)
 {
@@ -337,12 +368,12 @@ static int dnotify_rename_helper(
 	return 0;
 }
 
-static void dnotify_rename_file(const args_t *args, const char *path)
+static void dnotify_rename_file(const stress_args_t *args, const char *path)
 {
 	char oldfile[PATH_MAX], newfile[PATH_MAX];
 
-	mk_filename(oldfile, PATH_MAX, path, "dnotify_file");
-	mk_filename(newfile, PATH_MAX, path, "dnotify_file_renamed");
+	stress_mk_filename(oldfile, sizeof(oldfile), path, "dnotify_file");
+	stress_mk_filename(newfile, sizeof(newfile), path, "dnotify_file_renamed");
 
 	if (mk_file(args, oldfile, 4096) < 0)
 		return;
@@ -353,7 +384,7 @@ static void dnotify_rename_file(const args_t *args, const char *path)
 	(void)rm_file(args, newfile);
 }
 
-static const dnotify_stress_t dnotify_stressors[] = {
+static const stress_dnotify_stress_t dnotify_stressors[] = {
 	{ dnotify_access_file,		"DN_ACCESS" },
 	{ dnotify_modify_file,		"DN_MODIFY" },
 	{ dnotify_creat_file,		"DN_CREATE" },
@@ -367,9 +398,9 @@ static const dnotify_stress_t dnotify_stressors[] = {
  *  stress_dnotify()
  *	stress dnotify
  */
-int stress_dnotify(const args_t *args)
+static int stress_dnotify(const stress_args_t *args)
 {
-	char dirname[PATH_MAX];
+	char pathname[PATH_MAX];
 	int ret, i;
 	struct sigaction act;
 
@@ -382,22 +413,35 @@ int stress_dnotify(const args_t *args)
 		return EXIT_NO_RESOURCE;
 	}
 
-	stress_temp_dir_args(args, dirname, sizeof(dirname));
+	stress_temp_dir_args(args, pathname, sizeof(pathname));
 	ret = stress_temp_dir_mk_args(args);
 	if (ret < 0)
 		return exit_status(-ret);
+
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
 	do {
-		for (i = 0; g_keep_stressing_flag && dnotify_stressors[i].func; i++)
-			dnotify_stressors[i].func(args, dirname);
+		for (i = 0; keep_stressing(args) && dnotify_stressors[i].func; i++)
+			dnotify_stressors[i].func(args, pathname);
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
+
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	(void)stress_temp_dir_rm_args(args);
 
 	return EXIT_SUCCESS;
 }
+stressor_info_t stress_dnotify_info = {
+	.stressor = stress_dnotify,
+	.class = CLASS_FILESYSTEM | CLASS_SCHEDULER | CLASS_OS,
+	.supported = stress_dnotify_supported,
+	.verify = VERIFY_ALWAYS,
+	.help = help
+};
 #else
-int stress_dnotify(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_dnotify_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_FILESYSTEM | CLASS_SCHEDULER | CLASS_OS,
+	.help = help
+};
 #endif
