@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C) 2021-2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,13 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
+#define MIN_FALLOCATE_BYTES	(1 * MB)
+#define MAX_FALLOCATE_BYTES	(MAX_FILE_LIMIT)
+#define DEFAULT_FALLOCATE_BYTES	(1 * GB)
+
 static const stress_help_t help[] = {
 	{ NULL,	"fallocate N",		"start N workers fallocating 16MB files" },
 	{ NULL,	"fallocate-ops N",	"stop after N fallocate bogo operations" },
@@ -114,6 +113,13 @@ static int stress_fallocate(const stress_args_t *args)
 	char filename[PATH_MAX];
 	uint64_t ftrunc_errs = 0;
 	off_t fallocate_bytes = DEFAULT_FALLOCATE_BYTES;
+	int *mode_perms, all_modes;
+	size_t i, mode_count;
+	const char *fs_type;
+
+	for (all_modes = 0, i = 0; i < SIZEOF_ARRAY(modes); i++)
+		all_modes |= modes[i];
+	mode_count = stress_flag_permutation(all_modes, &mode_perms);
 
 	ret = sigsetjmp(jmp_env, 1);
 	if (ret) {
@@ -150,7 +156,8 @@ static int stress_fallocate(const stress_args_t *args)
 		(void)stress_temp_dir_rm_args(args);
 		return ret;
 	}
-	(void)unlink(filename);
+	fs_type = stress_fs_type(filename);
+	(void)shim_unlink(filename);
 
 	pipe_ret = pipe(pipe_fds);
 
@@ -169,8 +176,8 @@ static int stress_fallocate(const stress_args_t *args)
 			struct stat buf;
 
 			if (fstat(fd, &buf) < 0)
-				pr_fail("%s: fstat failed, errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
+				pr_fail("%s: fstat failed, errno=%d (%s)%s\n",
+					args->name, errno, strerror(errno), fs_type);
 			else if (buf.st_size != fallocate_bytes)
 				pr_fail("%s: file size %jd does not match size the expected file size of %jd\n",
 					args->name, (intmax_t)buf.st_size,
@@ -215,8 +222,6 @@ static int stress_fallocate(const stress_args_t *args)
 			/*
 			 *  non-portable Linux fallocate()
 			 */
-			int i;
-
 			(void)shim_fallocate(fd, 0, (off_t)0, fallocate_bytes);
 			if (!keep_stressing_flag())
 				break;
@@ -225,13 +230,22 @@ static int stress_fallocate(const stress_args_t *args)
 				break;
 
 			for (i = 0; i < 64; i++) {
-				size_t j = (stress_mwc32() >> 8) % SIZEOF_ARRAY(modes);	/* cppcheck-suppress moduloofone */
-				off_t offset = ((off_t)stress_mwc64() % fallocate_bytes) & ~0xfff;
+				const size_t j = (stress_mwc32() >> 8) % SIZEOF_ARRAY(modes);	/* cppcheck-suppress moduloofone */
+				const off_t offset = ((off_t)stress_mwc64() % fallocate_bytes) & ~0xfff;
 
-				(void)shim_fallocate(fd, modes[j], offset, 64 * KB);
+				if (shim_fallocate(fd, modes[j], offset, 64 * KB) == 0)
+					(void)shim_fsync(fd);
+				if (!keep_stressing_flag())
+					break;
+			}
+			/* Exercise all the mode permutations, most will fail */
+			for (i = 0; i < mode_count; i++) {
+				const off_t offset = ((off_t)stress_mwc64() % fallocate_bytes) & ~0xfff;
+
+				if (shim_fallocate(fd, mode_perms[i], offset, 64 * KB) == 0)
+					(void)shim_fsync(fd);
 				if (!keep_stressing_flag())
 					break;
-				(void)shim_fsync(fd);
 			}
 			if (ftruncate(fd, 0) < 0)
 				ftrunc_errs++;
@@ -259,8 +273,6 @@ static int stress_fallocate(const stress_args_t *args)
 		 *  Exercise with various illegal mode flags
 		 */
 		if (SIZEOF_ARRAY(illegal_modes) > 1) {
-			size_t i;
-
 			for (i = 0; i < SIZEOF_ARRAY(illegal_modes); i++) {
 				ret = shim_fallocate(fd, illegal_modes[i], (off_t)0, fallocate_bytes);
 				(void)ret;
@@ -315,7 +327,10 @@ done:
 	 *  exist.
 	 */
 	(void)stress_temp_dir_args(args, filename, sizeof(filename));
-	(void)rmdir(filename);
+	(void)shim_rmdir(filename);
+
+	if (mode_perms)
+		free(mode_perms);
 
 	return EXIT_SUCCESS;
 }
@@ -324,6 +339,7 @@ stressor_info_t stress_fallocate_info = {
 	.stressor = stress_fallocate,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_OPTIONAL,
 	.help = help
 };
 #else
