@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,16 +16,18 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 #include "io-uring.h"
 
+#if defined(HAVE_LINUX_IO_URING_H)
+#include <linux/io_uring.h>
+#endif
+
+#if !defined(O_DSYNC)
+#define O_DSYNC		(0)
+#endif
+
 static const stress_help_t help[] = {
 	{ NULL,	"io-uring N",		"start N workers that issue io-uring I/O requests" },
 	{ NULL,	"io-uring-ops N",	"stop after N bogo io-uring I/O requests" },
@@ -32,6 +35,7 @@ static const stress_help_t help[] = {
 };
 
 #if defined(HAVE_LINUX_IO_URING_H) &&	\
+    defined(HAVE_SYSCALL) &&		\
     defined(__NR_io_uring_enter) &&	\
     defined(__NR_io_uring_setup) &&	\
     defined(IORING_OFF_SQ_RING) &&	\
@@ -49,6 +53,7 @@ static const stress_help_t help[] = {
      defined(HAVE_IORING_OP_CLOSE) ||	\
      defined(HAVE_IORING_OP_MADVISE) ||	\
      defined(HAVE_IORING_OP_STATX) || 	\
+     defined(HAVE_IORING_OP_GETXATTR) || \
      defined(HAVE_IORING_OP_SYNC_FILE_RANGE))
 
 
@@ -58,6 +63,7 @@ static const stress_help_t help[] = {
  */
 typedef struct {
 	int fd;			/* file descriptor */
+	char *filename;		/* filename */
 	struct iovec *iovecs;	/* iovecs array 1 per block to submit */
 	size_t iovecs_sz;	/* size of iovecs allocation */
 	off_t file_size;	/* size of the file (bytes) */
@@ -179,14 +185,19 @@ static int stress_setup_io_uring(
 	struct io_uring_params p;
 
 	(void)memset(&p, 0, sizeof(p));
-	submit->io_uring_fd = shim_io_uring_setup(1, &p);
+	submit->io_uring_fd = shim_io_uring_setup(256, &p);
 	if (submit->io_uring_fd < 0) {
 		if (errno == ENOSYS) {
 			pr_inf_skip("%s: io-uring not supported by the kernel, skipping stressor\n",
 				args->name);
 			return EXIT_NOT_IMPLEMENTED;
 		}
-		pr_err("%s: io_uring_setup failed, errno=%d (%s)\n",
+		if (errno == ENOMEM) {
+			pr_inf_skip("%s: io-uring setup failed, out of memory, skipping stressor\n",
+				args->name);
+			return EXIT_NO_RESOURCE;
+		}
+		pr_fail("%s: io-uring setup failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 		return EXIT_FAILURE;
 	}
@@ -307,7 +318,7 @@ static inline int stress_io_uring_complete(
 			if (errno == EOPNOTSUPP) {
 				*supported = false;
 			} else  {
-				pr_err("%s: completion opcode=%d (%s), error=%d (%s)\n",
+				pr_fail("%s: completion opcode=%d (%s), error=%d (%s)\n",
 					args->name, opcode,
 					stress_io_uring_opcode_name(opcode),
 					err, strerror(err));
@@ -355,6 +366,7 @@ static int stress_io_uring_submit(
 	sring->array[index] = index;
 	tail = next_tail;
 	if (*sring->tail != tail) {
+		shim_mb();
 		*sring->tail = tail;
 		shim_mb();
 	}
@@ -464,7 +476,7 @@ static void stress_io_uring_fsync_setup(
 {
 	sqe->fd = io_uring_file->fd;
 	sqe->opcode = IORING_OP_FSYNC;
-	sqe->len = 512;
+	sqe->len = 0;
 	sqe->off = 0;
 	sqe->user_data = (uintptr_t)io_uring_file;
 	sqe->ioprio = 0;
@@ -588,12 +600,11 @@ static void stress_io_uring_statx_setup(
 	stress_io_uring_file_t *io_uring_file,
 	struct io_uring_sqe *sqe)
 {
-	static const char *pathname = "";
-	struct shim_statx statxbuf;
+	shim_statx_t statxbuf;
 
 	sqe->opcode = IORING_OP_STATX;
 	sqe->fd = io_uring_file->fd;
-	sqe->addr = (uintptr_t)pathname;
+	sqe->addr = (uintptr_t)io_uring_file->filename;
 	sqe->addr2 = (uintptr_t)&statxbuf;
 	sqe->statx_flags = AT_EMPTY_PATH;
 	sqe->ioprio = 0;
@@ -622,6 +633,30 @@ static void stress_io_uring_sync_file_range_setup(
 }
 #endif
 
+#if defined(HAVE_IORING_OP_GETXATTR)
+/*
+ *  stress_io_uring_getxattr_setup()
+ *	setup getxattr submit over io_uring
+ */
+static void stress_io_uring_getxattr_setup(
+	stress_io_uring_file_t *io_uring_file,
+	struct io_uring_sqe *sqe)
+{
+	char value[1024];
+
+	sqe->opcode = IORING_OP_STATX;
+	sqe->fd = io_uring_file->fd;
+	sqe->addr = (uintptr_t)io_uring_file->filename;
+	sqe->addr2 = (uintptr_t)value;
+	sqe->len = sizeof(value);
+	sqe->xattr_flags = 0;
+	sqe->ioprio = 0;
+	sqe->buf_index = 0;
+	sqe->flags = 0;
+}
+#endif
+
+
 static const stress_io_uring_setup_info_t stress_io_uring_setups[] = {
 #if defined(HAVE_IORING_OP_READV)
 	{ IORING_OP_READV,	"IORING_OP_READV", 	stress_io_uring_readv_setup },
@@ -659,6 +694,9 @@ static const stress_io_uring_setup_info_t stress_io_uring_setups[] = {
 #if defined(HAVE_IORING_OP_SYNC_FILE_RANGE)
 	{ IORING_OP_SYNC_FILE_RANGE, "IORING_OP_SYNC_FILE_RANGE", stress_io_uring_sync_file_range_setup },
 #endif
+#if defined(HAVE_IORING_OP_GETXATTR)
+	{ IORING_OP_GETXATTR, "IORING_OP_GETXATTR",	stress_io_uring_getxattr_setup },
+#endif
 };
 
 /*
@@ -724,6 +762,7 @@ static int stress_io_uring(const stress_args_t *args)
 			stress_io_uring_unmap_iovecs(&io_uring_file);
 			return EXIT_NO_RESOURCE;
 		}
+		(void)memset(io_uring_file.iovecs[i].iov_base, stress_mwc8(), block_size);
 		file_size -= iov_length;
 	}
 
@@ -736,17 +775,19 @@ static int stress_io_uring(const stress_args_t *args)
 	(void)stress_temp_filename_args(args,
 		filename, sizeof(filename), stress_mwc32());
 
+	io_uring_file.filename = filename;
+
 	rc = stress_setup_io_uring(args, &submit);
 	if (rc != EXIT_SUCCESS)
 		goto clean;
 
-	if ((io_uring_file.fd = open(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR)) < 0) {
+	if ((io_uring_file.fd = open(filename, O_CREAT | O_RDWR | O_DSYNC, S_IRUSR | S_IWUSR)) < 0) {
 		rc = exit_status(errno);
 		pr_fail("%s: open on %s failed, errno=%d (%s)\n",
 			args->name, filename, errno, strerror(errno));
 		goto clean;
 	}
-	(void)unlink(filename);
+	(void)shim_unlink(filename);
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 
@@ -789,6 +830,7 @@ clean:
 stressor_info_t stress_io_uring_info = {
 	.stressor = stress_io_uring,
 	.class = CLASS_IO | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
 #else
