@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,13 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
+#if defined(HAVE_LIBGEN_H)
+#include <libgen.h>
+#endif
+
 static const stress_help_t hardlink_help[] = {
 	{ NULL,	"link N",	 "start N workers creating hard links" },
 	{ NULL,	"link-ops N",	 "stop after N link bogo operations" },
@@ -53,14 +52,14 @@ static void stress_link_unlink(
 
 		(void)stress_temp_filename_args(args,
 			path, sizeof(path), i);
-		(void)unlink(path);
+		(void)shim_force_unlink(path);
 	}
 	(void)sync();
 }
 
-static inline int random_mount(const int mounts_max)
+static inline size_t random_mount(const int mounts_max)
 {
-	return (int)stress_mwc32() % mounts_max;
+	return (size_t)stress_mwc32() % (size_t)mounts_max;
 }
 
 /*
@@ -78,6 +77,7 @@ static int stress_link_generic(
 	bool symlink_func = (linkfunc == symlink);
 	char *mnts[MOUNTS_MAX];
 
+	(void)memset(tmp_newpath, 0, sizeof(tmp_newpath));
 	(void)snprintf(tmp_newpath, sizeof(tmp_newpath),
 		"/tmp/stress-ng-%s-%d-%" PRIu64 "-link",
 		args->name, (int)getpid(), stress_mwc64());
@@ -97,7 +97,6 @@ static int stress_link_generic(
 	(void)close(fd);
 
 	mounts_max = stress_mount_get(mnts, MOUNTS_MAX);
-
 	oldpathlen = strlen(oldpath);
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
@@ -130,8 +129,9 @@ static int stress_link_generic(
 					goto err_unlink;
 				}
 				rc = exit_status(errno);
-				pr_fail("%s: %s failed, errno=%d (%s)\n",
-					args->name, funcname, errno, strerror(errno));
+				pr_fail("%s: %s failed, errno=%d (%s)%s\n",
+					args->name, funcname, errno, strerror(errno),
+					stress_fs_type(oldpath));
 				n = i;
 				break;
 			}
@@ -157,8 +157,9 @@ static int stress_link_generic(
 					if (dir_fd >= 0) {
 						rret = readlinkat(dir_fd, filename, buf, sizeof(buf) - 1);
 						if ((rret < 0) && (errno != ENOSYS)) {
-							pr_fail("%s: readlinkat failed, errno=%d (%s)\n",
-							args->name, errno, strerror(errno));
+							pr_fail("%s: readlinkat failed, errno=%d (%s)%s\n",
+							args->name, errno, strerror(errno),
+							stress_fs_type(filename));
 						}
 						(void)close(dir_fd);
 					}
@@ -168,8 +169,9 @@ static int stress_link_generic(
 				rret = shim_readlink(newpath, buf, sizeof(buf) - 1);
 				if (rret < 0) {
 					rc = exit_status(errno);
-					pr_fail("%s: readlink failed, errno=%d (%s)\n",
-						args->name, errno, strerror(errno));
+					pr_fail("%s: readlink failed, errno=%d (%s)%s\n",
+						args->name, errno, strerror(errno),
+						stress_fs_type(newpath));
 				} else {
 					buf[rret] = '\0';
 					if ((size_t)rret != oldpathlen)
@@ -184,46 +186,44 @@ static int stress_link_generic(
 				/* Hard link, exercise illegal cross device link, EXDEV error */
 				if (mounts_max > 0) {
 					/* Try hard link on differet random mount point */
-					ret = linkfunc(mnts[random_mount(mounts_max)], tmp_newpath);
+					const size_t idx = random_mount(mounts_max);
+					
+					ret = linkfunc(mnts[idx], tmp_newpath);
 					if (ret == 0)
-						(void)unlink(tmp_newpath);
+						(void)shim_unlink(tmp_newpath);
 				}
 			}
 			if (lstat(newpath, &stbuf) < 0) {
 				rc = exit_status(errno);
-				pr_fail("%s: lstat failed, errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
+				pr_fail("%s: lstat failed, errno=%d (%s)%s\n",
+					args->name, errno, strerror(errno),
+					stress_fs_type(newpath));
 			}
 		}
 
 		/* exercise invalid newpath size, EINVAL */
-		rret = readlink(oldpath, testpath, 0);
-		(void)rret;
+		VOID_RET(ssize_t, readlink(oldpath, testpath, 0));
 
 		/* exercise empty oldpath, ENOENT */
-		rret = readlink("", testpath, sizeof(testpath));
-		(void)rret;
+		VOID_RET(ssize_t, readlink("", testpath, sizeof(testpath)));
 
 		/* exercise non-link, EINVAL */
-		rret = readlink("/", testpath, sizeof(testpath));
-		(void)rret;
+		VOID_RET(ssize_t, readlink("/", testpath, sizeof(testpath)));
 
 #if defined(HAVE_READLINKAT) && 	\
     defined(AT_FDCWD)
 		/* exercise invalid newpath size, EINVAL */
-		rret = readlinkat(AT_FDCWD, ".", testpath, 0);
-		(void)rret;
+		VOID_RET(ssize_t, readlinkat(AT_FDCWD, ".", testpath, 0));
 
 		/* exercise invalid newpath size, EINVAL */
-		rret = readlinkat(AT_FDCWD, "", testpath, sizeof(testpath));
-		(void)rret;
+		VOID_RET(ssize_t, readlinkat(AT_FDCWD, "", testpath, sizeof(testpath)));
 
 		/* exercise non-link, EINVAL */
-		rret = readlinkat(AT_FDCWD, "/", testpath, sizeof(testpath));
-		(void)rret;
+		VOID_RET(ssize_t, readlinkat(AT_FDCWD, "/", testpath, sizeof(testpath)));
 #endif
 
 err_unlink:
+		sync();
 		stress_link_unlink(args, n);
 
 		inc_counter(args);
@@ -231,7 +231,7 @@ err_unlink:
 
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
-	(void)unlink(oldpath);
+	(void)shim_unlink(oldpath);
 	(void)stress_temp_dir_rm_args(args);
 
 	stress_mount_free(mnts, mounts_max);
@@ -263,6 +263,7 @@ static int stress_symlink(const stress_args_t *args)
 stressor_info_t stress_link_info = {
 	.stressor = stress_link,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
 	.help = hardlink_help
 };
 #else
@@ -276,5 +277,6 @@ stressor_info_t stress_link_info = {
 stressor_info_t stress_symlink_info = {
 	.stressor = stress_symlink,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
 	.help = symlink_help
 };
