@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,45 +16,47 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
-#if defined(F_GETLK) && defined(F_SETLK) && defined(F_SETLKW) && \
-    defined(F_WRLCK) && defined(F_UNLCK)
+static const stress_help_t help[] = {
+	{ NULL,	"locka N",	"start N workers locking a file via advisory locks" },
+	{ NULL,	"locka-ops N",	"stop after N locka bogo operations" },
+	{ NULL, NULL,		NULL }
+};
+
+#if defined(F_GETLK) && 	\
+    defined(F_SETLK) &&		\
+    defined(F_SETLKW) && 	\
+    defined(F_WRLCK) &&		\
+    defined(F_UNLCK)
 
 #define LOCK_FILE_SIZE	(1024 * 1024)
-#define LOCK_SIZE	(8)
 #define LOCK_MAX	(1024)
 
 typedef struct locka_info {
+	struct locka_info *next;
 	off_t	offset;
 	off_t	len;
 	pid_t	pid;
-	struct locka_info *next;
-} locka_info_t;
+} stress_locka_info_t;
 
 typedef struct {
-	locka_info_t *head;		/* Head of locka_info procs list */
-	locka_info_t *tail;		/* Tail of locka_info procs list */
-	locka_info_t *free;		/* List of free'd locka_infos */
+	stress_locka_info_t *head;	/* Head of locka_info procs list */
+	stress_locka_info_t *tail;	/* Tail of locka_info procs list */
+	stress_locka_info_t *free;	/* List of free'd locka_infos */
 	uint64_t length;		/* Length of list */
-} locka_info_list_t;
+} stress_locka_info_list_t;
 
-static locka_info_list_t locka_infos;
+static stress_locka_info_list_t locka_infos;
 
 /*
  *  stress_locka_info_new()
  *	allocate a new locka_info, add to end of list
  */
-static locka_info_t *stress_locka_info_new(void)
+static stress_locka_info_t *stress_locka_info_new(void)
 {
-	locka_info_t *new;
+	stress_locka_info_t *new;
 
 	if (locka_infos.free) {
 		/* Pop an old one off the free list */
@@ -85,7 +88,7 @@ static locka_info_t *stress_locka_info_new(void)
 static void stress_locka_info_head_remove(void)
 {
 	if (locka_infos.head) {
-		locka_info_t *head = locka_infos.head;
+		stress_locka_info_t *head = locka_infos.head;
 
 		if (locka_infos.tail == locka_infos.head) {
 			locka_infos.tail = NULL;
@@ -109,14 +112,14 @@ static void stress_locka_info_head_remove(void)
 static void stress_locka_info_free(void)
 {
 	while (locka_infos.head) {
-		locka_info_t *next = locka_infos.head->next;
+		stress_locka_info_t *next = locka_infos.head->next;
 
 		free(locka_infos.head);
 		locka_infos.head = next;
 	}
 
 	while (locka_infos.free) {
-		locka_info_t *next = locka_infos.free->next;
+		stress_locka_info_t *next = locka_infos.free->next;
 
 		free(locka_infos.free);
 		locka_infos.free = next;
@@ -127,7 +130,7 @@ static void stress_locka_info_free(void)
  *  stress_locka_unlock()
  *	pop oldest lock record off list and unlock it
  */
-static int stress_locka_unlock(const args_t *args, const int fd)
+static int stress_locka_unlock(const stress_args_t *args, const int fd)
 {
 	struct flock f;
 
@@ -144,7 +147,8 @@ static int stress_locka_unlock(const args_t *args, const int fd)
 	stress_locka_info_head_remove();
 
 	if (fcntl(fd, F_SETLK, &f) < 0) {
-		pr_fail_err("F_SETLK");
+		pr_fail("%s: fcntl F_SETLK failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		return -1;
 	}
 	return 0;
@@ -155,24 +159,24 @@ static int stress_locka_unlock(const args_t *args, const int fd)
  *	hammer advisory lock/unlock to create some file lock contention
  */
 static int stress_locka_contention(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd)
 {
-	mwc_reseed();
+	stress_mwc_reseed();
 
 	do {
 		off_t offset;
 		off_t len;
 		int rc;
-		locka_info_t *locka_info;
+		stress_locka_info_t *locka_info;
 		struct flock f;
 
 		if (locka_infos.length >= LOCK_MAX)
 			if (stress_locka_unlock(args, fd) < 0)
 				return -1;
 
-		len  = (mwc16() + 1) & 0xfff;
-		offset = mwc64() % (LOCK_FILE_SIZE - len);
+		len = (stress_mwc16() + 1) & 0xfff;
+		offset = (off_t)stress_mwc64() % (LOCK_FILE_SIZE - len);
 
 		f.l_type = F_WRLCK;
 		f.l_whence = SEEK_SET;
@@ -188,7 +192,7 @@ static int stress_locka_contention(
 
 		locka_info = stress_locka_info_new();
 		if (!locka_info) {
-			pr_fail_err("calloc");
+			pr_err("%s: calloc failed, out of memory\n", args->name);
 			return -1;
 		}
 		locka_info->offset = offset;
@@ -196,7 +200,7 @@ static int stress_locka_contention(
 		locka_info->pid = args->pid;
 
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	return 0;
 }
@@ -205,27 +209,28 @@ static int stress_locka_contention(
  *  stress_locka
  *	stress file locking via advisory locking
  */
-int stress_locka(const args_t *args)
+static int stress_locka(const stress_args_t *args)
 {
 	int fd, ret = EXIT_FAILURE;
 	pid_t cpid = -1;
 	char filename[PATH_MAX];
-	char dirname[PATH_MAX];
+	char pathname[PATH_MAX];
 	char buffer[4096];
 	off_t offset;
 	ssize_t rc;
 
-	memset(buffer, 0, sizeof(buffer));
+	(void)memset(buffer, 0, sizeof(buffer));
 
 	/*
 	 *  There will be a race to create the directory
 	 *  so EEXIST is expected on all but one instance
 	 */
-	(void)stress_temp_dir_args(args, dirname, sizeof(dirname));
-	if (mkdir(dirname, S_IRWXU) < 0) {
+	(void)stress_temp_dir_args(args, pathname, sizeof(pathname));
+	if (mkdir(pathname, S_IRWXU) < 0) {
 		if (errno != EEXIST) {
 			ret = exit_status(errno);
-			pr_fail_err("mkdir");
+			pr_err("%s: mkdir %s failed, errno=%d (%s)\n",
+				args->name, pathname, errno, strerror(errno));
 			return ret;
 		}
 	}
@@ -236,22 +241,24 @@ int stress_locka(const args_t *args)
 	 *  stress flock processes
 	 */
 	(void)stress_temp_filename_args(args,
-		filename, sizeof(filename), mwc32());
+		filename, sizeof(filename), stress_mwc32());
 
 	if ((fd = open(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR)) < 0) {
 		ret = exit_status(errno);
-		pr_fail_err("open");
-		(void)rmdir(dirname);
+		pr_err("%s: open %s failed, errno=%d (%s)\n",
+			args->name, filename, errno, strerror(errno));
+		(void)shim_rmdir(pathname);
 		return ret;
 	}
 
 	if (lseek(fd, 0, SEEK_SET) < 0) {
-		pr_fail_err("lseek");
+		pr_err("%s: lseek failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto tidy;
 	}
 	for (offset = 0; offset < LOCK_FILE_SIZE; offset += sizeof(buffer)) {
 redo:
-		if (!g_keep_stressing_flag) {
+		if (!keep_stressing_flag()) {
 			ret = EXIT_SUCCESS;
 			goto tidy;
 		}
@@ -260,54 +267,65 @@ redo:
 			if ((errno == EAGAIN) || (errno == EINTR))
 				goto redo;
 			ret = exit_status(errno);
-			pr_fail_err("write");
+			pr_err("%s: write failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			goto tidy;
 		}
 	}
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
 	cpid = fork();
 	if (cpid < 0) {
-		if (!g_keep_stressing_flag) {
-			ret = EXIT_SUCCESS;
-			goto tidy;
-		}
-		if (errno == EAGAIN)
+		if (stress_redo_fork(errno))
 			goto again;
-		pr_fail_err("fork");
+		if (!keep_stressing(args))
+			goto tidy;
+		pr_err("%s: fork failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto tidy;
 	}
 	if (cpid == 0) {
 		(void)setpgid(0, g_pgrp);
 		stress_parent_died_alarm();
+		(void)sched_settings_apply(true);
 
 		if (stress_locka_contention(args, fd) < 0)
-			exit(EXIT_FAILURE);
+			_exit(EXIT_FAILURE);
 		stress_locka_info_free();
-		exit(EXIT_SUCCESS);
+		_exit(EXIT_SUCCESS);
 	}
 	(void)setpgid(cpid, g_pgrp);
 
 	if (stress_locka_contention(args, fd) == 0)
 		ret = EXIT_SUCCESS;
 tidy:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
 	if (cpid > 0) {
 		int status;
 
 		(void)kill(cpid, SIGKILL);
-		(void)waitpid(cpid, &status, 0);
+		(void)shim_waitpid(cpid, &status, 0);
 	}
 	stress_locka_info_free();
 
 	(void)close(fd);
-	(void)unlink(filename);
-	(void)rmdir(dirname);
+	(void)shim_unlink(filename);
+	(void)shim_rmdir(pathname);
 
 	return ret;
 }
+stressor_info_t stress_locka_info = {
+	.stressor = stress_locka,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
+	.help = help
+};
 #else
-int stress_locka(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_locka_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.help = help
+};
 #endif
