@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,12 +16,6 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
@@ -118,32 +113,22 @@ static size_t stress_mlock_max_lockable(void)
  */
 static void stress_mlock_misc(const stress_args_t *args, const size_t page_size)
 {
-	int ret;
-
 	(void)args;
 
 	/*
 	 *  mlock/munlock with invalid or unusual arguments
 	 */
-	ret = shim_mlock((void *)~0, page_size);
-	(void)ret;
-	ret = shim_munlock((void *)~0, page_size);
-	(void)ret;
+	VOID_RET(int, shim_mlock((void *)~0, page_size));
+	VOID_RET(int, shim_munlock((void *)~0, page_size));
 
-	ret = shim_mlock((void *)(~(uintptr_t)0 & ~(page_size - 1)), page_size << 1);
-	(void)ret;
-	ret = shim_munlock((void *)(~(uintptr_t)0 & ~(page_size - 1)), page_size << 1);
-	(void)ret;
+	VOID_RET(int, shim_mlock((void *)(~(uintptr_t)0 & ~(page_size - 1)), page_size << 1));
+	VOID_RET(int, shim_munlock((void *)(~(uintptr_t)0 & ~(page_size - 1)), page_size << 1));
 
-	ret = shim_mlock((void *)0, ~(size_t)0);
-	(void)ret;
-	ret = munlock((void *)0, ~(size_t)0);
-	(void)ret;
+	VOID_RET(int, shim_mlock((void *)0, ~(size_t)0));
+	VOID_RET(int, munlock((void *)0, ~(size_t)0));
 
-	ret = shim_mlock((void *)0, 0);
-	(void)ret;
-	ret = munlock((void *)0, 0);
-	(void)ret;
+	VOID_RET(int, shim_mlock((void *)0, 0));
+	VOID_RET(int, munlock((void *)0, 0));
 
 #if defined(HAVE_MLOCKALL)
 	{
@@ -260,7 +245,7 @@ static int stress_mlock_child(const stress_args_t *args, void *context)
 			if (ret < 0) {
 				if (errno == EAGAIN)
 					continue;
-				if (errno == ENOMEM)
+				if ((errno == ENOMEM) || (errno == EPERM))
 					break;
 				pr_fail("%s: mlock failed, errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
@@ -333,6 +318,7 @@ static int stress_mlock(const stress_args_t *args)
 stressor_info_t stress_mlock_info = {
 	.stressor = stress_mlock,
 	.class = CLASS_VM | CLASS_OS,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
 #else
