@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,21 +16,21 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
-#if !defined(__minix__) && !defined(__OpenBSD__) && !defined(__sun__)
-
 #define RMAP_CHILD_MAX		(16)
 #define MAPPINGS_MAX		(64)
 #define MAPPING_PAGES		(16)
 
+static const stress_help_t help[] = {
+	{ NULL,	"rmap N",	"start N workers that stress reverse mappings" },
+	{ NULL,	"rmap-ops N",	"stop after N rmap bogo operations" },
+	{ NULL,	NULL,		NULL }
+};
+
+static void *counter_lock;
+
 /*
  *  [ MAPPING 0 ]
  *  [ page ][ MAPPING 1 ]
@@ -43,90 +44,111 @@
  *  stress_rmap_handler()
  *      rmap signal handler
  */
-static void MLOCKED stress_rmap_handler(int dummy)
+static void MLOCKED_TEXT NORETURN stress_rmap_handler(int signum)
 {
-	(void)dummy;
+	(void)signum;
 
 	(void)kill(getppid(), SIGALRM);
-	exit(0);
+	_exit(0);
 }
 
-static void stress_rmap_child(
-	uint64_t *const counter,
-	const uint64_t max_ops,
+static void stress_rmap_touch(uint32_t *ptr, const size_t sz)
+{
+	const uint32_t *end = (uint32_t *)((uintptr_t)ptr + sz);
+	static uint32_t val = 0;
+	register const size_t inc = 64 >> 2; /* sizeof(*ptr) */
+
+	while (ptr < end) {
+		/* Bump val, never fill memory with non-zero val */
+		val++;
+		val = val ? val : 1;
+
+		*ptr = val;
+		ptr += inc;
+	}
+}
+
+static void NORETURN stress_rmap_child(
+	const stress_args_t *args,
 	const size_t page_size,
-	uint8_t *mappings[MAPPINGS_MAX])
+	uint32_t *mappings[MAPPINGS_MAX])
 {
 	const size_t sz = MAPPING_PAGES * page_size;
 
 	do {
 		ssize_t i;
-		const int sync_flag = mwc8() ? MS_ASYNC : MS_SYNC;
+		const uint8_t rnd8 = stress_mwc8();
+		const int sync_flag = (rnd8 & 0x80) ? MS_ASYNC : MS_SYNC;
 
-		switch (mwc32() & 3) {
-		case 0: for (i = 0; g_keep_stressing_flag && (i < MAPPINGS_MAX); i++) {
+		switch (rnd8 & 3) {
+		case 0: for (i = 0; i < MAPPINGS_MAX; i++) {
 				if (mappings[i] != MAP_FAILED) {
-					(void)memset(mappings[i], mwc8(), sz);
-					shim_msync(mappings[i], sz, sync_flag);
+					if (!inc_counter_lock(args, counter_lock, false))
+						break;
+					stress_rmap_touch(mappings[i], sz);
+					(void)shim_msync(mappings[i], sz, sync_flag);
 				}
 			}
 			break;
-		case 1: for (i = MAPPINGS_MAX - 1; g_keep_stressing_flag && (i >= 0); i--) {
+		case 1: for (i = MAPPINGS_MAX - 1; i >= 0; i--) {
 				if (mappings[i] != MAP_FAILED) {
-					(void)memset(mappings[i], mwc8(), sz);
-					shim_msync(mappings[i], sz, sync_flag);
+					if (!inc_counter_lock(args, counter_lock, false))
+						break;
+					stress_rmap_touch(mappings[i], sz);
+					(void)shim_msync(mappings[i], sz, sync_flag);
 				}
 			}
 			break;
-		case 2: for (i = 0; g_keep_stressing_flag && (i < MAPPINGS_MAX); i++) {
-				size_t j = mwc32() % MAPPINGS_MAX;
+		case 2: for (i = 0; i < MAPPINGS_MAX; i++) {
+				size_t j = stress_mwc32() % MAPPINGS_MAX;
+
 				if (mappings[j] != MAP_FAILED) {
-					(void)memset(mappings[j], mwc8(), sz);
-					shim_msync(mappings[j], sz, sync_flag);
+					if (!inc_counter_lock(args, counter_lock, false))
+						break;
+					stress_rmap_touch(mappings[j], sz);
+					(void)shim_msync(mappings[j], sz, sync_flag);
 				}
 			}
 			break;
-		case 3: for (i = 0; g_keep_stressing_flag && (i < MAPPINGS_MAX - 1); i++) {
+		case 3: for (i = 0; i < MAPPINGS_MAX - 1; i++) {
 				if (mappings[i] != MAP_FAILED) {
-					(void)memcpy(mappings[i], mappings[i + 1], sz);
-					shim_msync(mappings[i], sz, sync_flag);
+					if (!inc_counter_lock(args, counter_lock, false))
+						break;
+					stress_rmap_touch(mappings[i], sz);
+					(void)shim_msync(mappings[i], sz, sync_flag);
 				}
 			}
 			break;
 		}
-		(*counter)++;
-	} while (g_keep_stressing_flag && (!max_ops || *counter < max_ops));
+	} while (inc_counter_lock(args, counter_lock, true));
 
-	exit(0);
+	(void)kill(getppid(), SIGALRM);
+
+	_exit(0);
 }
 
 /*
  *  stress_rmap()
  *	stress mmap
  */
-int stress_rmap(const args_t *args)
+static int stress_rmap(const stress_args_t *args)
 {
 	const size_t page_size = args->page_size;
 	const size_t sz = ((MAPPINGS_MAX - 1) + MAPPING_PAGES) * page_size;
-	const size_t counters_sz =
-		(page_size + sizeof(uint64_t) * RMAP_CHILD_MAX) & ~(page_size - 1);
 	int fd = -1;
 	size_t i;
 	ssize_t rc;
 	pid_t pids[RMAP_CHILD_MAX];
-	uint8_t *mappings[MAPPINGS_MAX];
-	uint8_t *paddings[MAPPINGS_MAX];
-	uint64_t *counters;
+	uint32_t *mappings[MAPPINGS_MAX];
+	uint32_t *paddings[MAPPINGS_MAX];
 	char filename[PATH_MAX];
 
-	counters = (uint64_t *)mmap(NULL, counters_sz, PROT_READ | PROT_WRITE,
-		MAP_SHARED | MAP_ANONYMOUS, -1, 0);
-	if (counters == MAP_FAILED) {
-		pr_err("%s: mmap failed: errno=%d (%s)\n",
-			args->name, errno, strerror(errno));
-		exit(EXIT_FAILURE);
+	counter_lock = stress_lock_create();
+	if (!counter_lock) {
+		pr_inf("%s: failed to create counter lock. skipping stressor\n", args->name);
+		return EXIT_NO_RESOURCE;
 	}
-	(void)memset(counters, 0, counters_sz);
+
 	(void)memset(pids, 0, sizeof(pids));
 
 	for (i = 0; i < MAPPINGS_MAX; i++) {
@@ -135,48 +157,49 @@ int stress_rmap(const args_t *args)
 	}
 
 	/* Make sure this is killable by OOM killer */
-	set_oom_adjustment(args->name, true);
+	stress_set_oom_adjustment(args->name, true);
 
 	rc = stress_temp_dir_mk_args(args);
 	if (rc < 0)
-		return exit_status(-rc);
+		return exit_status((int)-rc);
 
 	(void)stress_temp_filename_args(args,
-		filename, sizeof(filename), mwc32());
+		filename, sizeof(filename), stress_mwc32());
 
-	(void)umask(0077);
 	if ((fd = open(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR)) < 0) {
 		rc = exit_status(errno);
-		pr_fail_err("open");
-		(void)unlink(filename);
+		pr_err("%s: open %s failed, errno=%d (%s)\n",
+			args->name, filename, errno, strerror(errno));
+		(void)shim_unlink(filename);
 		(void)stress_temp_dir_rm_args(args);
-		(void)munmap((void *)counters, counters_sz);
+		(void)stress_lock_destroy(counter_lock);
 
-		return rc;
+		return (int)rc;
 	}
-	(void)unlink(filename);
+	(void)shim_unlink(filename);
 
-	if (shim_fallocate(fd, 0, 0, sz) < 0) {
-		pr_fail_err("fallocate");
+	if (shim_fallocate(fd, 0, 0, (off_t)sz) < 0) {
+		pr_err("%s: fallocate failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		(void)close(fd);
 		(void)stress_temp_dir_rm_args(args);
-		(void)munmap((void *)counters, counters_sz);
+		(void)stress_lock_destroy(counter_lock);
 
-		return EXIT_FAILURE;
+		return EXIT_NO_RESOURCE;
 	}
 
 	for (i = 0; i < MAPPINGS_MAX; i++) {
-		off_t offset = i * page_size;
+		off_t offset = (off_t)(i * page_size);
 
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			goto cleanup;
 
 		mappings[i] =
-			mmap(0, MAPPING_PAGES * page_size, PROT_READ | PROT_WRITE,
+			(uint32_t *)mmap(0, MAPPING_PAGES * page_size, PROT_READ | PROT_WRITE,
 				MAP_SHARED, fd, offset);
 		/* Squeeze at least a page in between each mapping */
 		paddings[i] =
-			mmap(0, page_size, PROT_READ | PROT_WRITE,
+			(uint32_t *)mmap(0, page_size, PROT_READ | PROT_WRITE,
 				MAP_SHARED | MAP_ANONYMOUS, -1, 0);
 	}
 
@@ -184,7 +207,7 @@ int stress_rmap(const args_t *args)
 	 *  Spawn children workers
 	 */
 	for (i = 0; i < RMAP_CHILD_MAX; i++) {
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			goto cleanup;
 
 		pids[i] = fork();
@@ -195,38 +218,39 @@ int stress_rmap(const args_t *args)
 		} else if (pids[i] == 0) {
 			if (stress_sighandler(args->name, SIGALRM,
 			    stress_rmap_handler, NULL) < 0)
-				exit(EXIT_FAILURE);
+				_exit(EXIT_FAILURE);
 
 			(void)setpgid(0, g_pgrp);
 			stress_parent_died_alarm();
+			(void)sched_settings_apply(true);
 
 			/* Make sure this is killable by OOM killer */
-			set_oom_adjustment(args->name, true);
-			stress_rmap_child(&counters[i], args->max_ops / RMAP_CHILD_MAX,
-				page_size, mappings);
+			stress_set_oom_adjustment(args->name, true);
+			stress_rmap_child(args, page_size, mappings);
 		} else {
 			(void)setpgid(pids[i], g_pgrp);
 		}
 	}
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
 	/*
 	 *  Wait for SIGINT or SIGALRM
 	 */
-	while (keep_stressing()) {
-		/* Twiddle fingers */
-		shim_usleep(100000);
+	while (inc_counter_lock(args, counter_lock, false)) {
+		pause();
 	}
 
-	for (i = 0; i < RMAP_CHILD_MAX; i++)
-		*args->counter += counters[i];
 cleanup:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
 	/*
 	 *  Kill and wait for children
 	 */
 	for (i = 0; i < RMAP_CHILD_MAX; i++) {
 		if (pids[i] <= 0)
 			continue;
-		(void)kill(pids[i], SIGKILL);
+		(void)stress_killpid(pids[i]);
 	}
 	for (i = 0; i < RMAP_CHILD_MAX; i++) {
 		int status, ret;
@@ -234,33 +258,33 @@ cleanup:
 		if (pids[i] <= 0)
 			continue;
 
-		ret = waitpid(pids[i], &status, 0);
+		ret = shim_waitpid(pids[i], &status, 0);
 		if (ret < 0) {
 			if (errno != EINTR)
 				pr_dbg("%s: waitpid(): errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
 			(void)kill(pids[i], SIGTERM);
-			(void)kill(pids[i], SIGKILL);
-			(void)waitpid(pids[i], &status, 0);
+			(void)stress_killpid(pids[i]);
+			(void)shim_waitpid(pids[i], &status, 0);
 		}
 	}
 
-	(void)munmap((void *)counters, counters_sz);
-	(void)close(fd);
-	(void)stress_temp_dir_rm_args(args);
-
 	for (i = 0; i < MAPPINGS_MAX; i++) {
 		if (mappings[i] != MAP_FAILED)
-			(void)munmap(mappings[i], MAPPING_PAGES * page_size);
+			(void)munmap((void *)mappings[i], MAPPING_PAGES * page_size);
 		if (paddings[i] != MAP_FAILED)
-			(void)munmap(paddings[i], page_size);
+			(void)munmap((void *)paddings[i], page_size);
 	}
 
+	(void)close(fd);
+	(void)stress_temp_dir_rm_args(args);
+	(void)stress_lock_destroy(counter_lock);
+
 	return EXIT_SUCCESS;
 }
-#else
-int stress_rmap(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
-#endif
+
+stressor_info_t stress_rmap_info = {
+	.stressor = stress_rmap,
+	.class = CLASS_OS | CLASS_MEMORY,
+	.help = help
+};
