@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,16 +16,17 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
-#if defined(__linux__) && defined(__NR_memfd_create)
+static const stress_help_t help[] = {
+	{ NULL,	"seal N",	"start N workers performing fcntl SEAL commands" },
+	{ NULL,	"seal-ops N",	"stop after N SEAL bogo operations" },
+	{ NULL,	NULL,		NULL }
+};
+
+#if defined(__linux__) &&	\
+    defined(HAVE_MEMFD_CREATE)
 
 #ifndef F_ADD_SEALS
 #define F_ADD_SEALS		(1024 + 9)
@@ -44,27 +46,36 @@
 #ifndef F_SEAL_WRITE
 #define F_SEAL_WRITE		0x0008
 #endif
+#ifndef F_SEAL_FUTURE_WRITE
+#define F_SEAL_FUTURE_WRITE	0x0010
+#endif
 
+#ifndef MFD_ALLOW_SEALING
 #define MFD_ALLOW_SEALING	0x0002
+#endif
 
 /*
  *  stress_seal
  *	stress file sealing
  */
-int stress_seal(const args_t *args)
+static int stress_seal(const stress_args_t *args)
 {
 	int fd, ret;
 	int rc = EXIT_FAILURE;
 	const size_t page_size = args->page_size;
 	char filename[PATH_MAX];
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
 	do {
-		const off_t sz = page_size;
+		const off_t sz = (off_t)page_size;
 		uint8_t *ptr;
 		char buf[page_size];
+		ssize_t wret;
 
-		(void)snprintf(filename, sizeof(filename), "%s-%d-%" PRIu32 "-%" PRIu32,
-			args->name, args->pid, args->instance, mwc32());
+		(void)snprintf(filename, sizeof(filename), "%s-%" PRIdMAX "-%" PRIu32 "-%" PRIu32,
+			args->name, (intmax_t)args->pid,
+			args->instance, stress_mwc32());
 
 		fd = shim_memfd_create(filename, MFD_ALLOW_SEALING);
 		if (fd < 0) {
@@ -73,18 +84,21 @@ int stress_seal(const args_t *args)
 					"system call memfd_created\n", args->name);
 				return EXIT_NO_RESOURCE;
 			}
-			pr_fail_err("memfd_create");
+			pr_fail("%s: memfd_create %s failed, errno=%d (%s)\n",
+				args->name, filename, errno, strerror(errno));
 			return EXIT_FAILURE;
 		}
 
 		if (ftruncate(fd, sz) < 0) {
-			pr_fail_err("ftruncate");
+			pr_fail("%s: ftruncate failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
 
 		if (fcntl(fd, F_GET_SEALS) < 0) {
-			pr_fail_err("fcntl F_GET_SEALS");
+			pr_fail("%s: fcntl F_GET_SEALS failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
@@ -93,13 +107,15 @@ int stress_seal(const args_t *args)
 		 *  Add shrink SEAL, file cannot be make smaller
 		 */
 		if (fcntl(fd, F_ADD_SEALS, F_SEAL_SHRINK) < 0) {
-			pr_fail_err("fcntl F_ADD_SEALS F_SEAL_SHRINK");
+			pr_fail("%s: fcntl F_ADD_SEALS F_SEAL_SHRINK failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
 		ret = ftruncate(fd, 0);
 		if ((ret == 0) || ((ret < 0) && (errno != EPERM))) {
-			pr_fail_err("ftruncate did not fail with EPERM");
+			pr_fail("%s: ftruncate did not fail with EPERM as expected, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
@@ -108,13 +124,15 @@ int stress_seal(const args_t *args)
 		 *  Add grow SEAL, file cannot be made larger
 		 */
 		if (fcntl(fd, F_ADD_SEALS, F_SEAL_GROW) < 0) {
-			pr_fail_err("fcntl F_ADD_SEALS F_SEAL_GROW");
+			pr_fail("%s: fcntl F_ADD_SEALS F_SEAL_GROW failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
 		ret = ftruncate(fd, sz + 1);
 		if ((ret == 0) || ((ret < 0) && (errno != EPERM))) {
-			pr_fail_err("ftruncate did not fail with EPERM");
+			pr_fail("%s: ftruncate did not fail with EPERM as expected, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
@@ -123,55 +141,75 @@ int stress_seal(const args_t *args)
 		 *  mmap file, sealing it will return EBUSY until
 		 *  the mapping is removed
 		 */
-		ptr = mmap(NULL, sz, PROT_WRITE, MAP_SHARED,
+		ptr = mmap(NULL, (size_t)sz, PROT_WRITE, MAP_SHARED,
 			fd, 0);
 		if (ptr == MAP_FAILED) {
 			if (errno == ENOMEM)
 				goto next;
-			pr_fail_err("mmap");
+			pr_fail("%s: mmap failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
 		(void)memset(ptr, 0xea, page_size);
 		ret = fcntl(fd, F_ADD_SEALS, F_SEAL_WRITE);
 		if ((ret == 0) || ((ret < 0) && (errno != EBUSY))) {
-			pr_fail_err("fcntl F_ADD_SEALS F_SEAL_WRITE did not fail with EBUSY");
-			(void)munmap(ptr, sz);
+			pr_fail("%s: fcntl F_ADD_SEALS F_SEAL_WRITE did not fail with EBUSY as expected, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
+			(void)munmap(ptr, (size_t)sz);
 			(void)close(fd);
 			goto err;
 		}
 		(void)shim_msync(ptr, page_size, MS_SYNC);
-		(void)munmap(ptr, sz);
+		(void)munmap(ptr, (size_t)sz);
 
 		/*
 		 *  Now write seal the file, no more writes allowed
 		 */
 		if (fcntl(fd, F_ADD_SEALS, F_SEAL_WRITE) < 0) {
-			pr_fail_err("fcntl F_ADD_SEALS F_SEAL_WRITE");
+			if (errno == EBUSY)
+				goto next;
+			pr_fail("%s: fcntl F_ADD_SEALS F_SEAL_WRITE failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
 		(void)memset(buf, 0xff, sizeof(buf));
-		ret = write(fd, buf, sizeof(buf));
-		if ((ret == 0) || ((ret < 0) && (errno != EPERM))) {
-			pr_fail_err("write on sealed file did not fail with EPERM");
+		wret = write(fd, buf, sizeof(buf));
+		if ((wret == 0) || ((wret < 0) && (errno != EPERM))) {
+			pr_fail("%s: write on sealed file did not fail with EPERM as expected, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)close(fd);
 			goto err;
 		}
+
+		/*
+		 *  And try (and ignore error) from a F_SEAL_FUTURE_WRITE
+		 */
+		VOID_RET(int, fcntl(fd, F_ADD_SEALS, F_SEAL_FUTURE_WRITE));
 next:
 		(void)close(fd);
 
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	rc = EXIT_SUCCESS;
 err:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
 	return rc;
 }
+
+stressor_info_t stress_seal_info = {
+	.stressor = stress_seal,
+	.class = CLASS_OS,
+	.verify = VERIFY_ALWAYS,
+	.help = help
+};
 #else
-int stress_seal(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_seal_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_OS,
+	.help = help
+};
 #endif
