@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,15 +16,19 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
 
+#if defined(HAVE_SEM_SYSV)
+#include <sys/sem.h>
+#else
+UNEXPECTED
+#endif
+
+#define MIN_SEM_SYSV_PROCS     (2)
+#define MAX_SEM_SYSV_PROCS     (64)
+#define DEFAULT_SEM_SYSV_PROCS (2)
+
 #define STRESS_MAX_SEMS		(100)
 
 static const stress_help_t help[] = {
@@ -48,7 +53,7 @@ static int stress_set_semaphore_sysv_procs(const char *opt)
 
 	semaphore_sysv_procs = stress_get_uint64(opt);
 	stress_check_range("sem-sysv-procs", semaphore_sysv_procs,
-		MIN_SEMAPHORE_PROCS, MAX_SEMAPHORE_PROCS);
+		MIN_SEM_SYSV_PROCS, MAX_SEM_SYSV_PROCS);
 	return stress_set_setting("sem-sysv-procs", TYPE_ID_UINT64, &semaphore_sysv_procs);
 }
 
@@ -114,7 +119,7 @@ static void stress_semaphore_sysv_init(void)
 			"(%s), skipping semaphore stressor\n",
 			errno, strerror(errno));
 	} else {
-		pr_err("semaphore init (System V) failed: errno=%d: "
+		pr_fail("semaphore init (System V) failed: errno=%d: "
 			"(%s)\n", errno, strerror(errno));
 		_exit(EXIT_FAILURE);
 	}
@@ -200,9 +205,10 @@ static int stress_semaphore_sysv_thrash(const stress_args_t *args)
 			}
 			timeout.tv_sec = 0;
 			timeout.tv_nsec = 100000;
-			ret = semtimedop(sem_id, sems, STRESS_MAX_SEMS * 3, &timeout);
-			(void)ret;
+			VOID_RET(int, semtimedop(sem_id, sems, STRESS_MAX_SEMS * 3, &timeout));
 		}
+#else
+		UNEXPECTED
 #endif
 
 		for (i = 0; i < 1000; i++) {
@@ -272,8 +278,9 @@ timed_out:
 			} else {
 #if defined(IPC_SET)
 				s.buf = &ds;
-				ret = semctl(sem_id, 2, IPC_SET, &s);
-				(void)ret;
+				VOID_RET(int, semctl(sem_id, 2, IPC_SET, &s));
+#else
+				UNEXPECTED
 #endif
 			}
 
@@ -284,13 +291,12 @@ timed_out:
 				nsems = 64;
 			s.array = calloc(nsems, sizeof(*s.array));
 			if (s.array) {
-				ret = semctl(sem_id, 2, GETALL, s);
-				(void)ret;
+				VOID_RET(int, semctl(sem_id, 2, GETALL, s));
 #if defined(SETALL) && 0
 				/*
 				 *  SETALL across the semaphores will clobber the state
 				 *  and cause waits on semaphores because of the unlocked
-				 *  state change. Currenltly this is disabled.
+				 *  state change. Currently this is disabled.
 				 */
 				if (ret == 0)
 					ret = semctl(sem_id, 2, SETALL, s);
@@ -299,6 +305,8 @@ timed_out:
 			}
 #endif
 		}
+#else
+		UNEXPECTED
 #endif
 #if defined(SEM_STAT) &&	\
     defined(__linux__)
@@ -308,13 +316,11 @@ timed_out:
 
 			/* Exercise with a 0 index into kernel array */
 			s.buf = &ds;
-			ret = semctl(0, 0, SEM_STAT, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0, 0, SEM_STAT, &s));
 
 			/* Exercise with a probably illegal index into kernel array */
 			s.buf = &ds;
-			ret = semctl(0x1fffffff, 0, SEM_STAT, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0x1fffffff, 0, SEM_STAT, &s));
 		}
 #endif
 #if defined(SEM_STAT_ANY) &&	\
@@ -325,13 +331,11 @@ timed_out:
 
 			/* Exercise with a 0 index into kernel array */
 			s.__buf = &si;
-			ret = semctl(0, 0, SEM_STAT_ANY, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0, 0, SEM_STAT_ANY, &s));
 
 			/* Exercise with a probably illegal index into kernel array */
 			s.__buf = &si;
-			ret = semctl(0x1fffffff, 0, SEM_STAT_ANY, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0x1fffffff, 0, SEM_STAT_ANY, &s));
 		}
 #endif
 #if defined(IPC_INFO) &&	\
@@ -352,16 +356,14 @@ timed_out:
 			 *  actually ignored, so should be OK
 			 */
 			s.__buf = &si;
-			ret = semctl(0x1fffffff, 0, IPC_INFO, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0x1fffffff, 0, IPC_INFO, &s));
 
 			/*
 			 *  Exercise with an illegal sem number, this arg is
 			 *  also ignored, so should be OK
 			 */
 			s.__buf = &si;
-			ret = semctl(sem_id, ~0, IPC_INFO, &s);
-			(void)ret;
+			VOID_RET(int, semctl(sem_id, ~0, IPC_INFO, &s));
 		}
 #endif
 #if defined(SEM_INFO) &&	\
@@ -379,68 +381,61 @@ timed_out:
 
 			/* Exercise with probably an illegal sem id */
 			s.__buf = &si;
-			ret = semctl(0x1fffffff, 0, IPC_INFO, &s);
-			(void)ret;
+			VOID_RET(int, semctl(0x1fffffff, 0, IPC_INFO, &s));
 
 			/* Exercise with an illegal sem number */
 			s.__buf = &si;
-			ret = semctl(sem_id, ~0, IPC_INFO, &s);
-			(void)ret;
+			VOID_RET(int, semctl(sem_id, ~0, IPC_INFO, &s));
 		}
 #endif
 #if defined(GETVAL)
-		ret = semctl(sem_id, 0, GETVAL, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, 0, GETVAL, 0));
 
 		/* Exercise with probably an illegal sem id */
-		ret = semctl(0x1fffffff, 0, GETVAL, 0);
-		(void)ret;
+		VOID_RET(int, semctl(0x1fffffff, 0, GETVAL, 0));
 
 		/* Exercise with an illegal sem number */
-		ret = semctl(sem_id, ~0, GETVAL, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, ~0, GETVAL, 0));
+#else
+		UNEXPECTED
 #endif
 #if defined(GETPID)
-		ret = semctl(sem_id, 0, GETPID, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, 0, GETPID, 0));
 
 		/* Exercise with probably an illegal sem id */
-		ret = semctl(0x1fffffff, 0, GETPID, 0);
-		(void)ret;
+		VOID_RET(int, semctl(0x1fffffff, 0, GETPID, 0));
 
 		/* Exercise with an illegal sem number */
-		ret = semctl(sem_id, ~0, GETPID, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, ~0, GETPID, 0));
+#else
+		UNEXPECTED
 #endif
 #if defined(GETNCNT)
-		ret = semctl(sem_id, 0, GETNCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, 0, GETNCNT, 0));
 
 		/* Exercise with probably an illegal sem id */
-		ret = semctl(0x1fffffff, 0, GETNCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(0x1fffffff, 0, GETNCNT, 0));
 
 		/* Exercise with an illegal sem number */
-		ret = semctl(sem_id, ~0, GETNCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, ~0, GETNCNT, 0));
+#else
+		UNEXPECTED
 #endif
 #if defined(GETZCNT)
-		ret = semctl(sem_id, 0, GETZCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, 0, GETZCNT, 0));
 
 		/* Exercise with probably an illegal sem id */
-		ret = semctl(0x1fffffff, 0, GETZCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(0x1fffffff, 0, GETZCNT, 0));
 
 		/* Exercise with an illegal sem number */
-		ret = semctl(sem_id, ~0, GETZCNT, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, ~0, GETZCNT, 0));
+#else
+		UNEXPECTED
 #endif
 		/*
 		 * Now exercise invalid options and arguments
 		 */
-		ret = semctl(sem_id, -1, SETVAL, 0);
-		(void)ret;
+		VOID_RET(int, semctl(sem_id, -1, SETVAL, 0));
 #if defined(HAVE_SEMTIMEDOP)
 		if (got_semtimedop) {
 			/*
@@ -454,39 +449,36 @@ timed_out:
 			semwait.sem_num = 0;
 			semwait.sem_op = -1;
 			semwait.sem_flg = SEM_UNDO;
-			ret = semtimedop(sem_id, &semwait, 1, &timeout);
-			(void)ret;
+			VOID_RET(int, semtimedop(sem_id, &semwait, 1, &timeout));
 
 			/*
 			 *  Exercise invalid semid, EINVAL
 			 */
 			timeout.tv_sec = 0;
 			timeout.tv_nsec = 10000;
-			ret = semtimedop(-1, &semwait, 1, &timeout);
-			(void)ret;
+			VOID_RET(int, semtimedop(-1, &semwait, 1, &timeout));
 
 			/*
 			 *  Exercise invalid nsops, E2BIG
 			 */
 			timeout.tv_sec = 0;
 			timeout.tv_nsec = 10000;
-			ret = semtimedop(sem_id, &semwait, -1, &timeout);
-			(void)ret;
+			VOID_RET(int, semtimedop(sem_id, &semwait, (size_t)-1, &timeout));
 		}
+#else
+		UNEXPECTED
 #endif
 		{
 			struct sembuf semwait;
 			/*
 			 *  Exercise invalid semid, EINVAL
 			 */
-			ret = semop(-1, &semwait, 1);
-			(void)ret;
+			VOID_RET(int, semop(-1, &semwait, 1));
 
 			/*
 			 *  Exercise invalid nsops, E2BIG
 			 */
-			ret = semop(sem_id, &semwait, -1);
-			(void)ret;
+			VOID_RET(int, semop(sem_id, &semwait, -1));
 		}
 
 		/*
@@ -495,12 +487,11 @@ timed_out:
 		{
 			struct sembuf semwait;
 
-			semwait.sem_num = -1;
+			semwait.sem_num = (unsigned short)-1;
 			semwait.sem_op = -1;
 			semwait.sem_flg = SEM_UNDO;
 
-			ret = semop(sem_id, &semwait, 1);
-			(void)ret;
+			VOID_RET(int, semop(sem_id, &semwait, 1));
 		}
 
 		/*
@@ -509,7 +500,8 @@ timed_out:
 		 *  is invalid without doing the actual system call
 		 */
 		(void)semctl(sem_id, 0, INT_MAX, 0);
-#if defined(__NR_semctl)
+#if defined(__NR_semctl) &&	\
+    defined(HAVE_SYSCALL)
 		(void)syscall(__NR_semctl, sem_id, 0, INT_MAX, 0);
 #endif
 	} while ((rc == EXIT_SUCCESS)  && keep_stressing(args));
@@ -557,16 +549,16 @@ static void stress_sem_sysv_sigchild(int sig)
  */
 static int stress_sem_sysv(const stress_args_t *args)
 {
-	pid_t pids[MAX_SEMAPHORE_PROCS];
+	pid_t pids[MAX_SEM_SYSV_PROCS];
 	uint64_t i;
-	uint64_t semaphore_sysv_procs = DEFAULT_SEMAPHORE_PROCS;
+	uint64_t semaphore_sysv_procs = DEFAULT_SEM_SYSV_PROCS;
 	int rc = EXIT_SUCCESS;
 
 	if (!stress_get_setting("sem-sysv-procs", &semaphore_sysv_procs)) {
 		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
-			semaphore_sysv_procs = MAX_SEMAPHORE_PROCS;
+			semaphore_sysv_procs = MAX_SEM_SYSV_PROCS;
 		if (g_opt_flags & OPT_FLAGS_MINIMIZE)
-			semaphore_sysv_procs = MIN_SEMAPHORE_PROCS;
+			semaphore_sysv_procs = MIN_SEM_SYSV_PROCS;
 	}
 
 	if (!g_shared->sem_sysv.init) {
@@ -594,7 +586,7 @@ reap:
 
 	for (i = 0; i < semaphore_sysv_procs; i++) {
 		if (pids[i] > 0)
-			(void)kill(pids[i], SIGKILL);
+			(void)stress_killpid(pids[i]);
 	}
 	for (i = 0; i < semaphore_sysv_procs; i++) {
 		if (pids[i] > 0) {
@@ -617,6 +609,7 @@ stressor_info_t stress_sem_sysv_info = {
 	.deinit = stress_semaphore_sysv_deinit,
 	.class = CLASS_OS | CLASS_SCHEDULER,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
 #else
