@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,14 +16,15 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
+#include "core-capabilities.h"
+
+#if defined(HAVE_LINUX_RANDOM_H)
+#include <linux/random.h>
+#else
+UNEXPECTED
+#endif
 
 static const stress_help_t help[] = {
 	{ "u N","urandom N",	 "start N workers reading /dev/urandom" },
@@ -114,8 +116,10 @@ static int stress_urandom(const stress_args_t *args)
 	do {
 		uint8_t buffer[8192];
 		ssize_t ret;
+#if defined(HAVE_SELECT)
 		struct timeval timeout;
 		fd_set rdfds;
+#endif
 
 		if (fd_urnd >= 0) {
 			ret = read(fd_urnd, buffer, sizeof(buffer));
@@ -133,7 +137,6 @@ static int stress_urandom(const stress_args_t *args)
 		 * this fails, just skip this part of the stressor
 		 */
 		if (fd_rnd >= 0) {
-			off_t offset;
 #if defined(RNDGETENTCNT)
 			unsigned long val = 0;
 
@@ -142,6 +145,8 @@ static int stress_urandom(const stress_args_t *args)
 			/* Try to avoid emptying entropy pool */
 			if (val < 128)
 				goto next;
+#else
+			UNEXPECTED
 #endif
 			ret = read(fd_rnd, buffer, 1);
 			if (ret < 0) {
@@ -155,8 +160,7 @@ static int stress_urandom(const stress_args_t *args)
 #if defined(RNDGETENTCNT)
 next:
 #endif
-			offset = lseek(fd_rnd, (off_t)0, SEEK_SET);
-			(void)offset;
+			VOID_RET(off_t, lseek(fd_rnd, (off_t)0, SEEK_SET));
 
 			if (!sys_admin) {
 				/*
@@ -170,10 +174,14 @@ next:
 #if defined(RNDCLEARPOOL)
 				ret = ioctl(fd_rnd, RNDCLEARPOOL, NULL);
 				check_eperm(args, ret, errno);
+#else
+				UNEXPECTED
 #endif
 #if defined(RNDZAPENTCNT)
 				ret = ioctl(fd_rnd, RNDZAPENTCNT, NULL);
 				check_eperm(args, ret, errno);
+#else
+				UNEXPECTED
 #endif
 #if defined(RNDADDTOENTCNT)
 				{
@@ -186,16 +194,19 @@ next:
 					ret = ioctl(fd_rnd, RNDADDTOENTCNT, &count);
 					check_eperm(args, ret, errno);
 				}
+#else
+				UNEXPECTED
 #endif
 #if defined(RNDRESEEDCRNG)
 				ret = ioctl(fd_rnd, RNDRESEEDCRNG, stress_mwc32());
 				check_eperm(args, ret, errno);
+#else
+				UNEXPECTED
 #endif
 				/*
 				 *  Exercise invalid ioctl command
 				 */
-				ret = ioctl(fd_rnd, 0xffff, NULL);
-				(void)ret;
+				VOID_RET(int, ioctl(fd_rnd, 0xffff, NULL));
 #if defined(__linux__)
 				if (fd_rnd_wr >= 0) {
 					buffer[0] = stress_mwc8();
@@ -218,6 +229,7 @@ next:
 				(void)munmap(ptr, args->page_size);
 		}
 
+#if defined(HAVE_SELECT)
 		/*
 		 *  Peek if data is available on blockable /dev/random and
 		 *  try to read it.
@@ -231,17 +243,39 @@ next:
 			ret = select(fd_rnd_blk + 1, &rdfds, NULL, NULL, &timeout);
 			if (ret > 0) {
 				if (FD_ISSET(fd_rnd_blk, &rdfds)) {
-					ret = read(fd_rnd, buffer, 1);
+					ret = -1;
+#if defined(__linux__)
+					char *ptr;
+
+					/*
+					 *  Older kernels will EFAULT on reads of data off the end of
+					 *  a page, where as newer kernels 5.18-rc2+ will return a
+					 *  single byte in the same way as reading /dev/zero does,
+					 *  as fixed in Linux kernel commit:
+					 *  "random: allow partial reads if later user copies fail"
+					 */
+					ptr = (char *)mmap(NULL, args->page_size, PROT_WRITE,
+						MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
+					if (ptr != MAP_FAILED) {
+						/* Exercise 2 byte read on last 1 byte of page */
+						ret = read(fd_rnd, ptr + args->page_size - 1, 2);
+						(void)munmap((void *)ptr, args->page_size);
+					}
+#endif
 					if (ret < 0) {
-						if ((errno != EAGAIN) && (errno != EINTR)) {
-							pr_fail("%s: read of /dev/random failed, errno=%d (%s)\n",
-								args->name, errno, strerror(errno));
-							goto err;
+						ret = read(fd_rnd, buffer, 1);
+						if (ret < 0) {
+							if ((errno != EAGAIN) && (errno != EINTR)) {
+								pr_fail("%s: read of /dev/random failed, errno=%d (%s)\n",
+									args->name, errno, strerror(errno));
+								goto err;
+							}
 						}
 					}
 				}
 			}
 		}
+#endif
 
 		inc_counter(args);
 	} while (keep_stressing(args));
@@ -266,5 +300,6 @@ err:
 stressor_info_t stress_urandom_info = {
 	.stressor = stress_urandom,
 	.class = CLASS_DEV | CLASS_OS,
+	.verify = VERIFY_OPTIONAL,
 	.help = help
 };
