@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
+ * Copyright (C)      2022 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,14 +16,15 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * This code is a complete clean re-write of the stress tool by
- * Colin Ian King <colin.king@canonical.com> and attempts to be
- * backwardly compatible with the stress tool by Amos Waterland
- * <apw@rossby.metr.ou.edu> but has more stress tests and more
- * functionality.
- *
  */
 #include "stress-ng.h"
+#include "core-pragma.h"
+
+#if defined(HAVE_UTIME_H)
+#include <utime.h>
+#else
+UNEXPECTED
+#endif
 
 static const stress_help_t help[] = {
 	{ NULL,	"utime N",	"start N workers updating file timestamps" },
@@ -39,6 +41,26 @@ static int stress_set_utime_fsync(const char *opt)
 	return stress_set_setting("utime-fsync", TYPE_ID_BOOL, &utime_fsync);
 }
 
+static const stress_opt_set_func_t opt_set_funcs[] = {
+	{ OPT_utime_fsync,	stress_set_utime_fsync },
+	{ 0,			NULL }
+};
+
+#if defined(HAVE_UTIME_H)
+
+#if defined(HAVE_UTIME) &&	\
+    defined(HAVE_UTIMBUF)
+static int shim_utime(const char *filename, const struct utimbuf *times)
+{
+#if defined(__NR_utime) &&	\
+    defined(HAVE_SYSCALL)
+	return (int)syscall(__NR_utime, filename, times);
+#else
+	return utime(filename, times);
+#endif
+}
+#endif
+
 /*
  *  stress_utime()
  *	stress system by setting file utime
@@ -55,7 +77,11 @@ static int stress_utime(const stress_args_t *args)
 	char hugename[PATH_MAX + 16];
 	int ret, fd;
 	bool utime_fsync = false;
-
+#if defined(HAVE_UTIME_H) &&	\
+    defined(HAVE_UTIME) &&	\
+    defined(HAVE_UTIMBUF)
+	const bool verify = !!(g_opt_flags & OPT_FLAGS_VERIFY);
+#endif
 	(void)stress_get_setting("utime-fsync", &utime_fsync);
 
 	ret = stress_temp_dir_mk_args(args);
@@ -67,6 +93,8 @@ static int stress_utime(const stress_args_t *args)
     defined(UTIME_NOW)
 	(void)stress_temp_dir_args(args, path, sizeof(path));
 	dir_fd = open(path, O_DIRECTORY | O_PATH);
+#else
+	UNEXPECTED
 #endif
 
 	(void)stress_temp_filename_args(args,
@@ -92,32 +120,32 @@ static int stress_utime(const stress_args_t *args)
 		struct timespec ts[2];
 #endif
 
-		if (gettimeofday(&timevals[0], NULL) == 0) {
-			timevals[1] = timevals[0];
-			if (utimes(filename, timevals) < 0) {
-				pr_dbg("%s: utimes failed: errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
-				break;
-			}
+		(void)gettimeofday(&timevals[0], NULL);
+		timevals[1] = timevals[0];
+		if (utimes(filename, timevals) < 0) {
+			pr_dbg("%s: utimes failed: errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
+			break;
 		}
 		if (utimes(filename, NULL) < 0) {
-			pr_dbg("%s: utimes failed: errno=%d (%s)\n",
-				args->name, errno, strerror(errno));
+			pr_dbg("%s: utimes failed: errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
 			break;
 		}
 
 		/* Exercise with invalid filename, ENOENT */
-		ret = utimes("", timevals);
-		(void)ret;
+		VOID_RET(int, utimes("", timevals));
 
 		/* Exercise huge filename, ENAMETOOLONG */
-		ret = utimes(hugename, timevals);
-		(void)ret;
+		VOID_RET(int, utimes(hugename, timevals));
 
 #if defined(HAVE_FUTIMENS)
 		if (futimens(fd, NULL) < 0) {
-			pr_dbg("%s: futimens failed: errno=%d (%s)\n",
-				args->name, errno, strerror(errno));
+			pr_dbg("%s: futimens failed: errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
 			break;
 		}
 
@@ -126,8 +154,7 @@ static int stress_utime(const stress_args_t *args)
 		ts[0].tv_nsec = UTIME_NOW;
 		ts[1].tv_sec = UTIME_NOW;
 		ts[1].tv_nsec = UTIME_NOW;
-		ret = futimens(-1, ts);
-		(void)ret;
+		VOID_RET(int, futimens(-1, ts));
 
 #if defined(UTIME_NOW)
 		ts[0].tv_sec = UTIME_NOW;
@@ -135,20 +162,26 @@ static int stress_utime(const stress_args_t *args)
 		ts[1].tv_sec = UTIME_NOW;
 		ts[1].tv_nsec = UTIME_NOW;
 		if (futimens(fd, ts) < 0) {
-			pr_dbg("%s: futimens failed: errno=%d (%s)\n",
-				args->name, errno, strerror(errno));
+			pr_dbg("%s: futimens failed: errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
 			break;
 		}
+#else
+		UNEXPECTED
 #endif
 
 #if defined(UTIME_OMIT)
 		ts[0].tv_sec = UTIME_OMIT;
 		ts[0].tv_nsec = UTIME_OMIT;
 		if (futimens(fd, ts) < 0) {
-			pr_dbg("%s: futimens failed: errno=%d (%s)\n",
-				args->name, errno, strerror(errno));
+			pr_dbg("%s: futimens failed: errno=%d (%s)%s\n",
+				args->name, errno, strerror(errno),
+				stress_fs_type(filename));
 			break;
 		}
+#else
+		UNEXPECTED
 #endif
 #endif
 
@@ -161,25 +194,24 @@ static int stress_utime(const stress_args_t *args)
 		ts[1].tv_sec = UTIME_NOW;
 		ts[1].tv_nsec = UTIME_NOW;
 
-		ret = utimensat(AT_FDCWD, filename, ts, 0);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, filename, ts, 0));
 
 		/* Exercise invalid filename, ENOENT */
-		ret = utimensat(AT_FDCWD, "", ts, 0);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, "", ts, 0));
 
 		/* Exercise huge filename, ENAMETOOLONG */
-		ret = utimensat(AT_FDCWD, hugename, ts, 0);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, hugename, ts, 0));
 
 		/* Exercise invalid flags */
-		ret = utimensat(AT_FDCWD, filename, ts, ~0);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, filename, ts, ~0));
+#else
+		UNEXPECTED
 #endif
 
 #if defined(O_DIRECTORY) &&	\
     defined(O_PATH) &&		\
-    defined(UTIME_NOW)
+    defined(UTIME_NOW) &&	\
+    defined(HAVE_PRAGMA_INSIDE)
 STRESS_PRAGMA_PUSH
 STRESS_PRAGMA_WARN_OFF
 		if (dir_fd >= 0) {
@@ -189,73 +221,88 @@ STRESS_PRAGMA_WARN_OFF
 			ts[1].tv_sec = UTIME_NOW;
 			ts[1].tv_nsec = UTIME_NOW;
 
-			ret = utimensat(dir_fd, "", ts, 0);
-			(void)ret;
+			VOID_RET(int, utimensat(dir_fd, "", ts, 0));
 		}
 STRESS_PRAGMA_POP
 #endif
 
 #if defined(UTIME_OMIT)
 		ts[1].tv_nsec = UTIME_OMIT;
-		ret = utimensat(AT_FDCWD, filename, ts, 0);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, filename, ts, 0));
 #endif
 
 
 #if defined(AT_SYMLINK_NOFOLLOW)
 #if defined(UTIME_NOW)
 		ts[1].tv_nsec = UTIME_NOW;
-		ret = utimensat(AT_FDCWD, filename, ts, AT_SYMLINK_NOFOLLOW);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, filename, ts, AT_SYMLINK_NOFOLLOW));
 		if (utime_fsync) {
-			ret = shim_fsync(fd);
-			(void)ret;
+			VOID_RET(int, shim_fsync(fd));
 		}
 #endif
 #if defined(UTIME_OMIT)
 		ts[1].tv_nsec = UTIME_OMIT;
-		ret = utimensat(AT_FDCWD, filename, ts, AT_SYMLINK_NOFOLLOW);
-		(void)ret;
+		VOID_RET(int, utimensat(AT_FDCWD, filename, ts, AT_SYMLINK_NOFOLLOW));
 		if (utime_fsync) {
-			ret = shim_fsync(fd);
-			(void)ret;
+			VOID_RET(int, shim_fsync(fd));
 		}
 #endif
 #endif
+#else
+		UNEXPECTED
 #endif
 
-#if defined(HAVE_UTIME_H)
+#if defined(HAVE_UTIME_H) &&	\
+    defined(HAVE_UTIME) &&	\
+    defined(HAVE_UTIMBUF)
 		{
 			struct utimbuf utbuf;
-			const double actime = stress_time_now();
+			struct timeval tv;
 
-			utbuf.actime = (time_t)actime;
+			(void)gettimeofday(&tv, NULL);
+			utbuf.actime = (time_t)tv.tv_sec;
 			utbuf.modtime = utbuf.actime;
 
-			if (utime(filename, &utbuf) < 0) {
-				pr_dbg("%s: utime failed: errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
+			if (shim_utime(filename, &utbuf) < 0) {
+				pr_fail("%s: utime failed: errno=%d (%s)%s\n",
+					args->name, errno, strerror(errno),
+					stress_fs_type(filename));
 				break;
 			}
-			if (utime(filename, NULL) < 0) {
-				pr_dbg("%s: utime failed: errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
+
+			if (verify) {
+				struct stat statbuf;
+
+				if (stat(filename, &statbuf) == 0) {
+					if (statbuf.st_atime < tv.tv_sec) {
+						pr_fail("%s: utime failed: access time is less than expected time\n",
+							args->name);
+					}
+					if (statbuf.st_mtime < tv.tv_sec) {
+						pr_fail("%s: utime failed: modified time is less than expected time\n",
+							args->name);
+					}
+				}
+			}
+			if (shim_utime(filename, NULL) < 0) {
+				pr_fail("%s: utime failed: errno=%d (%s)%s\n",
+					args->name, errno, strerror(errno),
+					stress_fs_type(filename));
 				break;
 			}
 
 			/* Exercise invalid timename, ENOENT */
-			ret = utime("", &utbuf);
-			(void)ret;
+			VOID_RET(int, shim_utime("", &utbuf));
 
 			/* Exercise huge filename, ENAMETOOLONG */
-			ret = utime(hugename, &utbuf);
-			(void)ret;
+			VOID_RET(int, shim_utime(hugename, &utbuf));
 		}
+#else
+		UNEXPECTED
 #endif
 		/* forces metadata writeback */
 		if (utime_fsync) {
-			ret = shim_fsync(fd);
-			(void)ret;
+			VOID_RET(int, shim_fsync(fd));
 		}
 		inc_counter(args);
 	} while (keep_stressing(args));
@@ -269,20 +316,28 @@ STRESS_PRAGMA_POP
 		(void)close(dir_fd);
 #endif
 	(void)close(fd);
-	(void)unlink(filename);
+	(void)shim_unlink(filename);
 	(void)stress_temp_dir_rm_args(args);
 
 	return EXIT_SUCCESS;
 }
 
-static const stress_opt_set_func_t opt_set_funcs[] = {
-	{ OPT_utime_fsync,	stress_set_utime_fsync },
-	{ 0,			NULL }
-};
-
 stressor_info_t stress_utime_info = {
 	.stressor = stress_utime,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_OPTIONAL,
 	.help = help
 };
+
+#else
+
+stressor_info_t stress_utime_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_OPTIONAL,
+	.help = help
+};
+
+#endif
